SystemOrganization addCategory: #'SRP-StateReplicationProtocol'!
SystemOrganization addCategory: #'SRP-StateReplicationProtocol-SrpPortableBehaviors'!
SystemOrganization addCategory: #'SRP-PortableObjectRepresentations'!
SystemOrganization addCategory: #'SRP-Tests'!

!Character class methodsFor: '*SRP-Char' stamp: ''!
srpDataTypeMap

	^super srpDataTypeMap, #(
		(self Character)
	)! !

!Character class methodsFor: '*SRP-Char' stamp: ''!
srpPrimaryStoreType
	"Answer the type of object that should be stored in place of
	 the receiver or for all indexable slots of the receiver. Careful,
	 subclasses get this too. The valid types are: 
			#(State Unsigned Character Signed Byte String)."

	^#Character

! !

!Character class methodsFor: '*SRP-Char' stamp: ''!
srpStateStructureType
	"Answer the type of storage format that should be
	 used, instead of the default, for all instances of the
	 receiver. Careful, subclasses are affected also."

	^#Typed! !

!Behavior methodsFor: '*SRP-BaseExt' stamp: ''!
srpDataTypeMap
	"This can be overriden in subclasses to specify how SRP will save the state instances
	 and instance variables.  Answer a collection of two element arrays as shown below.
	 The first element of an array is either #self, #IndexedSlots, or the name of the instance
	 variable to be mapped. The second element of an array is the SRP data type to save
	 and load as (refer to #storeTypeArrays).

		#(	(self State)			- The instance itself will be saved as State (the default).
			(IndexableType Unsigned)	- Each indexable slot will be saved as an Unsigned integer.
			(instVarA ByteArray)	- <instVarA> will be saved as a ByteArray.
			(instVarB nil)				- <instVarB> will not be saved.
			(instVarC Proxy)			- <instVarC> can be saved as a proxy.
		)"

	^#()
! !

!Behavior methodsFor: '*SRP-BaseExt' stamp: ''!
srpHasIndexableValues

	^SrpPortal behaviorHasIndexableValues: self! !

!Behavior methodsFor: '*SRP-BaseExt' stamp: ''!
srpNew
	"Answer a fixed size instance of the receiver."

	^self basicNew
! !

!Behavior methodsFor: '*SRP-BaseExt' stamp: ''!
srpNew: anInteger
	"Answer an instance of the receiver with anInteger number
	 of indexable slots (either variables or bytes depending on
	 the receiver's type). Error if the receiver does not allow
	 variable length instances."

	^SrpPortal behavior: self basicNew: anInteger! !

!Behavior methodsFor: '*SRP-BaseExt' stamp: ''!
srpSlotNamed: stringOrSymbol
	"Private. Answer an integer value. Answer zero if not found."

	| myVarNames |
	myVarNames := self allInstVarNames.
	1 to: myVarNames size do: [:i |
		(SrpPortal sequenceableCollection: (myVarNames at: i) equals: stringOrSymbol)
			ifTrue: [^i].
	].
	^0! !

!Behavior methodsFor: '*SRP-BaseExt' stamp: ''!
srpVersionNumberArray
	"Answer nil or an array of unsigned integers that represent the class version number of the receiver."

	^nil! !

!UndefinedObject class methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpNew
	"You never want to create a new instance of 
	 this class so just answer the image singleton."

	^nil
! !

!UndefinedObject methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpLoadPreMap: aContext
	"This ensures there is only one instance in the image."

	^nil
! !

!Integer class methodsFor: '*SRP-Integer' stamp: ''!
srpDataTypeMap

	^super srpDataTypeMap, #(
		(self Signed)
	)! !

!Integer class methodsFor: '*SRP-Integer' stamp: ''!
srpPrimaryStoreType
	"Answer the type of object that should be stored in place of
	 the receiver or for all indexable slots of the receiver. Careful,
	 subclasses get this too. The valid types are: 
			#(State Unsigned Character Signed Byte String)."

	^#Signed

! !

!Integer class methodsFor: '*SRP-Integer' stamp: ''!
srpStateStructureType
	"Answer the type of storage format that should be
	 used, instead of the default, for all instances of the
	 receiver. Careful, subclasses are affected also."

	^#Typed
! !

!String class methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpDataTypeMap

	^super srpDataTypeMap, #(
		(IndexableType Character)
	)

	"Note: Strings are normally mapped to PmrString. This is
				required if normal mapping is not in effect. The type
				#String could not be used because subclasses would
				inadvertently become Strings."
! !

!String class methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpPrimaryStoreType
	"Answer the type of object that should be stored in place of
	 the receiver or for all indexable slots of the receiver. Careful,
	 subclasses get this too. The valid types are: 
			#(State Unsigned Character Signed Byte String)."

	^#Character

	"Note: Strings are normally mapped to PmrString. This is
				required if normal mapping is not in effect. The type
				#String could not be used because subclasses would
				inadvertently become Strings."
! !

!String methodsFor: '*SRP-MinimalMappingRulesVW' stamp: ''!
srpBasicAt: index put: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer, but that would be less portable with other
	 Smalltalk dialects."

	self basicAt: index put: aCharacter asInteger.
	^aCharacter! !

!String methodsFor: '*SRP-pMinimalMappingRules' stamp: ''!
srpIndexableSize
	"This is implemented only for the benefit of ObjectStudio
	 because String>>basicSize returns 0 or 1 in ObjectStudio.
	 This override is likely OK for other dialects."

	^self size! !

!String methodsFor: '*SRP-pMinimalMappingRules' stamp: ''!
srpIndexableValueAt: index
	"This is implemented only for the benefit of ObjectStudio
	 because string in ObjectStudio appear to be OS pointers.
	 This override is likely OK for other dialects."

	^self at: index
! !

!String methodsFor: '*SRP-pMinimalMappingRules' stamp: ''!
srpIndexableValueAt: index put: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer, but that would be less portable with other
	 Smalltalk dialects."

	^SrpPortal characterCollection: self at: index putCharacter: aCharacter
! !

!Collection class methodsFor: '*SRP-Collection' stamp: ''!
srpNew: anInteger
	"Answer an instance of the receiver with anInteger number
	 of indexable slots (either variables or bytes depending on
	 the receiver's type). Error if the receiver does not allow
	 variable length instances."

	^self new: anInteger! !

!Boolean class methodsFor: '*SRP-dataTypeMap' stamp: ''!
srpDataTypeMap
	"Some dialects don't have subclasses of Boolean."

	^self class == Boolean
		ifTrue: [super srpDataTypeMap]
		ifFalse: [super srpDataTypeMap, #((self Boolean))]! !

!Boolean class methodsFor: '*SRP-singleton' stamp: 'cwp 12/25/2006 13:18'!
srpNew
	^ self initializedInstance! !

!Symbol class methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpDataTypeMap

	^super srpDataTypeMap, #(
		(IndexableType Character)
	)

	"Note: Symbols are normally mapped to PmrSymbol. This is
				required if normal mapping is not in effect."
! !

!Symbol class methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpNew: anInteger
	"Answer an instance of the receiver with anInteger number
	 of indexable slots (either variables or bytes depending on
	 the receiver's type). Error if the receiver does not allow
	 variable length instances."

	^SrpPortal behavior: self basicNew: anInteger! !

!Symbol class methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpPrimaryStoreType
	"Answer the type of object that should be stored in place of
	 the receiver or for all indexable slots of the receiver. Careful,
	 subclasses get this too. The valid types are: 
			#(State Unsigned Character Signed Byte String)."

	^#Character

	"Note: Symbols are normally mapped to PmrSymbol. This is
				required if normal mapping is not in effect."
! !

!Symbol methodsFor: '*SRP-MinimalMappingRules' stamp: ''!
srpLoadPreMap: aContext
	"Symbols must be canonical in the Smalltalk image.
	 This double conversion is a portable way to ensure
	 the receiver is interned. At this point, the receiver is
	 not part of the symbol table."

	^self asString asSymbol
! !

!False methodsFor: '*SRP-False' stamp: ''!
srpLoadPreMap: aContext
	"This ensures there is only one instance in the image."

	^false
! !

!True methodsFor: '*SRP-loading' stamp: ''!
srpLoadPreMap: aContext
	"This ensures there is only one instance in the image."

	^true
! !

TestCase subclass: #SrtSrpTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

SrtSrpTest subclass: #SrtBehaviorResolution
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtBehaviorResolution methodsFor: 'SUnit Tests' stamp: 'sd 7/19/2006 16:50'!
testBehaviorNameAccess
	"	self new testBehaviorNameAccess	"
	
	self assert: (self portal shortNameForBehavior: Object) == #Object.
	(#(VisualSmalltalk ObjectStudio) includes: self portal vmTypeName)
		ifTrue: [self assert: #MetaClass == (self portal shortNameForBehavior: Object class)]
		ifFalse: [self assert: #Metaclass == (self portal shortNameForBehavior: Object class)].
	self assert: (self portal shortNameForClass: Object) == #Object.
	self assert: (self portal shortNameForClass: Object class) == #Object.
! !

!SrtBehaviorResolution methodsFor: 'SUnit Tests' stamp: ''!
testNameConversions
	"	self new testNameConversions	"
	"Note: ST/X doesn't support #withAllSubclassesDo:."
	| namePath resolved |
	self portal behaviorWithAllSubclasses: Collection do: [:aClass |
		namePath := self portal namespacePathForBehavior: aClass.
		resolved := self portal behaviorForNamespacePath: namePath.
		self assert: resolved == aClass.
	].! !

SrtSrpTest subclass: #SrtDialectPortability
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtDialectPortability methodsFor: 'SUnit Tests' stamp: ''!
testFileReads
	"	self new testFileReads.	"

	"Note: VW VariableBinding currently isn't being remapped.
			   This will produce multiple class doesn't exist warnings."
"	self timedReadFilename: 'vwTestSrpDefaultsComplexObject.srp'.
	self timedReadFilename: 'vwTestRtcDefaultsComplexObject.srp'.
"
	self timedReadFilename: 'vaTestSrpDefaultsComplexObject.srp'.
	self timedReadFilename: 'vaTestRtcDefaultsComplexObject.srp'.! !

!SrtDialectPortability methodsFor: 'Test support code' stamp: ''!
timedReadFilename: filename
	| ostream t1 obj |
	(self portal stringExistsAsFile: filename) ifFalse: [
		self reportComment: 'File ''', filename, ''' didn''t exist for test load.'.
		^nil
	].
	ostream := SrpConfiguration default newStreamOpenBinaryFilename: filename.
	[
	t1 := self portal millisecondsToRun: [ obj := ostream next ].
	] ensure: [ostream close].
	self reportComment: 'File ''', filename, ''' loaded in ', t1 printString, 'ms.'.
	^obj! !

!SrtDialectPortability methodsFor: 'Test support code' stamp: ''!
timedWriteRead: obj on: ostream filename: filename
	| t1 t2 startPos |
	t1 := self portal millisecondsToRun: [ ostream nextPut: obj ].
	ostream reset.
	startPos := ostream position.
	t2 := self portal millisecondsToRun: [ ostream next ].
	self reportComment: 
		'File ', filename, 
		' save ', t1 printString, 
		'ms, load ', t2 printString, 
		'ms, size ', (ostream position - startPos) printString, ' bytes.'.
! !

!SrtDialectPortability methodsFor: 'SUnit Tests' stamp: ''!
vaTestRtcDefaultsComplexObjectWrite
	"	self new vaTestRtcDefaultsComplexObjectWrite.	"

	| filename ostream obj |
	filename := 'vaTestRtcDefaultsComplexObject.srp'.
	obj := self complexObject.
	ostream := (SrpConfiguration default newStreamOpenBinaryFilename: filename)
		metastateTableName: 'RtcDefaults1';
		yourself.
	self
		timedWriteRead: obj
		on: ostream
		filename: filename.
	ostream close.! !

!SrtDialectPortability methodsFor: 'SUnit Tests' stamp: ''!
vaTestSrpDefaultsComplexObjectWrite
	"	self new vaTestSrpDefaultsComplexObjectWrite.	"

	| filename ostream obj |
	filename := 'vaTestSrpDefaultsComplexObject.srp'.
	obj := self complexObject.
	ostream := SrpConfiguration default newStreamOpenBinaryFilename: filename.
	self
		timedWriteRead: obj
		on: ostream
		filename: filename.
	ostream close.! !

!SrtDialectPortability methodsFor: 'SUnit Tests' stamp: ''!
vwTestRtcDefaultsComplexObjectWrite
	"	self new vwTestRtcDefaultsComplexObjectWrite.	"

	| filename ostream obj |
	filename := 'vwTestRtcDefaultsComplexObject.srp'.
	obj := self complexObject.
	ostream := (SrpConfiguration default newStreamOpenBinaryFilename: filename)
		metastateTableName: 'RtcDefaults1';
		yourself.
	self
		timedWriteRead: obj
		on: ostream
		filename: filename.
	ostream close.! !

!SrtDialectPortability methodsFor: 'SUnit Tests' stamp: ''!
vwTestSrpDefaultsComplexObjectWrite
	"	self new vwTestSrpDefaultsComplexObjectWrite.	"

	| filename ostream obj |
	filename := 'vwTestSrpDefaultsComplexObject.srp'.
	obj := self complexObject.
	ostream := SrpConfiguration default newStreamOpenBinaryFilename: filename.
	self
		timedWriteRead: obj
		on: ostream
		filename: filename.
	ostream close.! !

SrtSrpTest subclass: #SrtDistinctTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtDistinctTest methodsFor: 'test running' stamp: ''!
betatestSimpleDistinct
	"	self new betatestSimpleDistinct.	"

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: obj asSymbol ];
		fromProxyBlock: [:obj | loadedProxies add: obj asString ].
	object := Array with: (SrtDistinctTestObject key: 'key' value: 'value').
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 1.
	self assert: savedProxies first = #value.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 1.
	self assert: loadedProxies first = 'value'.
	self assert: loaded value == loadedProxies first.
	^loaded
	
		
	
! !

SrtSrpTest subclass: #SrtExtendedAttributesTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtExtendedAttributesTest classSide methodsFor: 'Not categorized' stamp: ''!
srpExtendedAttributes

	^'myExtendedAttributeClassObject'

! !

!SrtExtendedAttributesTest classSide methodsFor: 'Not categorized' stamp: ''!
srpExtendedAttributes: eaObject

	self new assert: eaObject = self srpExtendedAttributes.


! !

!SrtExtendedAttributesTest methodsFor: 'Test Support Code' stamp: ''!
srpExtendedAttributes

	^'myExtendedAttributeInstanceObject'

! !

!SrtExtendedAttributesTest methodsFor: 'Test Support Code' stamp: ''!
srpExtendedAttributes: eaObject

	self assert: eaObject = self srpExtendedAttributes.


! !

!SrtExtendedAttributesTest methodsFor: 'SUnit Tests' stamp: ''!
testExtendedAttributes
	"	self new testExtendedAttributes.	"

	"The assertions tests triggered in #srpExtendedAttributes: do the actual testing."
	^SrpConfiguration new replicate: self.


! !

SrtSrpTest subclass: #SrtLiteralReplication
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtLiteralReplication methodsFor: 'Test Support Code' stamp: ''!
equalityLiterals
	"Literals that should compare = when replicated."

	^(OrderedCollection new)
		add: 'ABC';
		add: (self portal vmTypeName == #ObjectStudio ifTrue: [999999] ifFalse: [999999999999999]);
		add: (self portal vmTypeName == #ObjectStudio ifTrue: [-999999] ifFalse: [-999999999999999]);
		add: 1.5;
		add: -1.5;
		add: 1 / 3;
		add: -1 / 3;
		add: #(name date);
		add: (self portal byteArrayClassWithAll: #(255 65 66 67 0));
		asArray

	"Note: ObjectStudio is unable to parse #(#name #date)."
	"Note: ObjectStudio is treats 999999999999999 as a Decimal."! !

!SrtLiteralReplication methodsFor: 'Test Support Code' stamp: ''!
identityLiterals
	"Literals that should compare == when replicated."

	SrpPortal vmTypeName == #ObjectStudio ifTrue: [
		"ObjectStudio has a non-standard parser for array literals."
		^(OrderedCollection new)
			add: true;
			add: false;
			add: nil;
			add: #name;
			add: $A;
			add: 7;
			add: -7;
			asArray
	].
	^#(true false nil name $A 7 -7)

	! !

!SrtLiteralReplication methodsFor: 'SUnit Tests' stamp: ''!
testLiteralDefaultReplications
	"	self new testLiteralDefaultReplications	"

	self equalityLiterals do: [:ea |
		self assert: (SrpConfiguration new replicate: ea) = ea.
		self assert: (SrpNonMappingConfiguration new replicate: ea) = ea.
		self assert: (SrpBase64Configuration new replicate: ea) = ea.
		self assert: (SrtSequenceCheckConfiguration new replicate: ea) = ea.
	].
	self identityLiterals do: [:ea |
		self assert: (SrpConfiguration new replicate: ea) == ea.
		self assert: (SrpNonMappingConfiguration new replicate: ea) == ea.
		self assert: (SrpBase64Configuration new replicate: ea) == ea.
		self assert: (SrtSequenceCheckConfiguration new replicate: ea) == ea.
	].! !

!SrtLiteralReplication methodsFor: 'SUnit Tests' stamp: ''!
testSingletonCreation
	"Because #srpLoaded also returns the image singleton,
	 implementors of #srpNew that return the image
	 singleton are redundant. Failing this test may not
	 matter so long as replication works properly."

	SrpPortal vmTypeName == #'GemStone/S' ifFalse: [
		"GemStone doesn't have subclasses of Boolean."
		self assert: (true class srpNew) == true.
		self assert: (false class srpNew) == false.
	].
	self assert: (UndefinedObject srpNew) == nil.
! !

!SrtLiteralReplication methodsFor: 'SUnit Tests' stamp: ''!
testStringReplications
	"	self new testStringReplications	"

	| stream |
	stream := WriteStream on: String new.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream cr.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream nextPut: $A.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream cr.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream nextPutAll: 'ABCD'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream cr; nextPutAll: 'EFGH'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream cr; cr; nextPutAll: 'IJKL'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream nextPut: (self portal unsignedToCharacter: 10); cr; nextPutAll: 'MNOP'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream nextPut: (self portal unsignedToCharacter: 13); cr; nextPutAll: 'QRST'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream nextPut: (self portal unsignedToCharacter: 13); nextPut: (self portal unsignedToCharacter: 13); cr; nextPutAll: 'UVW'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream nextPut: (self portal unsignedToCharacter: 10); nextPut: (self portal unsignedToCharacter: 10); cr; nextPutAll: 'XYZ'.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.
	stream cr; cr; tab; cr; cr.
	self assert: (SrpConfiguration default replicate: stream contents) = stream contents.! !

SrtSrpTest subclass: #SrtMetastateTableReplication
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtMetastateTableReplication methodsFor: 'SUnit Tests' stamp: ''!
testSrpDefaultsComplexObject100xWriteRead
	"	self new testSrpDefaultsComplexObject100xWriteRead	"

	| ostream method t1 t2 |
	method := self complexObject.
	ostream := SrpConfiguration new newStream.
	t1 := t2 := 0.
	100 timesRepeat: [
		t1 := t1 + (self portal millisecondsToRun: [ ostream nextPut: method ]).
		ostream reset.
		t2 := t2 + (self portal millisecondsToRun: [ ostream next ]).
		ostream reset.
	].
	self reportComment: #testSrpDefaultsComplexObject100xWriteRead, ' save ', t1 printString, 'ms, load ', t2 printString, 'ms.'.
! !

!SrtMetastateTableReplication methodsFor: 'SUnit Tests' stamp: ''!
testSrpDefaultsComplexObjectWriteRead
	"	self new testSrpDefaultsComplexObjectWriteRead	"

	| ostream method t1 t2 bytes |
	method := self complexObject.
	ostream := SrpNonMappingConfiguration new
		addExchangeRule: PmrClassReference;
		addExchangeRule: PmrMetaclassReference;
		newStream.
	t1 := self portal millisecondsToRun: [ ostream nextPut: method ].
	ostream reset.
	t2 := self portal millisecondsToRun: [ ostream next ].
	bytes := ostream streamContents.
	self reportComment: 
		#testSrpDefaultsComplexObjectWriteRead, ' save ', 
		t1 printString, 'ms, load ', 
		t2 printString, 'ms, size ', 
		bytes size printString, ' bytes.'.
! !

!SrtMetastateTableReplication methodsFor: 'SUnit Tests' stamp: ''!
vaTestSwapperComplexObjectWriteRead
	"	self new vaTestSwapperComplexObjectWriteRead.	"

	| method stream bytes t1 t2 |
	method := self complexObject.
	stream := ReadWriteStream on: ByteArray new.
	t1 := self portal millisecondsToRun: [
			(SrpPortal behaviorForName: #ObjectDumper) new
				unload: method intoStream: stream.
	].
	stream reset.
	t2 := self portal millisecondsToRun: [
		(SrpPortal behaviorForName: #ObjectLoader) new
			loadFromStream: stream.
	].
	bytes := stream contents.
	self reportComment: 
		#vaTestSwapperComplexObjectWriteRead, ' save ', 
		t1 printString, 'ms, load ', 
		t2 printString, 'ms, size ', 
		bytes size printString, ' bytes.'.
! !

!SrtMetastateTableReplication methodsFor: 'SUnit Tests' stamp: ''!
vwTestBossComplexObjectWriteRead
	"	self new vwTestBossComplexObjectWriteRead	"

	| method stream boss bytes t1 t2 |
	(SrpPortal behaviorForName: #BinaryObjectStorage) isNil ifTrue: [^self].
	method := self complexObject.
	stream := ReadWriteStream on: ByteArray new.
	boss := (SrpPortal behaviorForName: #BinaryObjectStorage) onNew: stream.
	boss expectCycles: true.
	t1 := self portal millisecondsToRun: [
			boss nextPut: method.
	].
	boss reset.
	t2 := self portal millisecondsToRun: [
		boss next.
	].
	bytes := stream contents.
	self reportComment: 
		#vwTestBossComplexObjectWriteRead, ' save ', 
		t1 printString, 'ms, load ', 
		t2 printString, 'ms, size ', 
		bytes size printString, ' bytes.'.
! !

SrtSrpTest subclass: #SrtMultipleObjectTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtMultipleObjectTest methodsFor: 'SUnit Tests' stamp: ''!
testWriteReadMultipleObjects
	"	self new testWriteReadMultipleObjects	"

	| objs |
	objs := (0 to: 99) asArray.
	self assert: (
		SrpConfiguration default newStream
			nextPutAll: objs;
			contents
	) = objs.
	
! !

!SrtMultipleObjectTest methodsFor: 'SUnit Tests' stamp: ''!
testWriteReadMultipleObjectsBase64
	"	self new testWriteReadMultipleObjectsBase64	"

	| objs |
	objs := (0 to: 99) asArray.
	self assert: (
		SrpConfiguration default newStream text
			nextPutAll: objs;
			contents
	) = objs.
	
! !

SrtSrpTest subclass: #SrtPortableMappingTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtPortableMappingTest methodsFor: 'Test Support Code' stamp: ''!
collectionInstances
	" self runTests	"
	^OrderedCollection new
	"	add: (Bag with: 'you' with: 'me' with: 'others');
		add: (Set with: 'you' with: 'me' with: 'others');
		add: (Dictionary new
			at: 'you' put: 'Sam';
			at: 'me' put: 'Paul';
			at: 'others' put: 'World';
			yourself);
		add: (IdentityDictionary new
			at: #'you' put:  'Sam';
			at: #'me' put: 'Paul';
			at: #'others' put: 'World';
			yourself);	"
		add: (50 to: 75);
		add: ((-999 to: 3010 by: 11) asArray);
		add: (self portal byteArrayClassWithAll: (0 to: 255));
		add: (self portal byteArrayAsString: (self portal byteArrayClassWithAll: (0 to: 255)));
		add: (self newDoubleByteString);
		yourself
! !

!SrtPortableMappingTest methodsFor: 'Test Support Code' stamp: ''!
newDoubleByteString

	"Couldn't create DBString in VA with (String with: (Character value: 16rFFF).
	 Couldn't create TwoByteString in VW with ((Character value: 16rFFF) asString).
	 Had to rely on my own portable method #asTwoByteString added to VA."

	self portal twoByteStringClass isNil ifTrue: [^String new].
	self portal vmTypeName == #'GemStone/S' ifTrue: [
		^self portal byteArrayAsString: (0 to: 16rFFFF by: 50).
	].
	^(WriteStream on: (self portal stringAsTwoByteString: String new))
		nextPutAll: ((0 to: 16rFFFF by: 50) collect: [:ea | self portal unsignedToCharacter: ea ]);
		contents.
! !

!SrtPortableMappingTest methodsFor: 'SUnit Tests' stamp: ''!
testCommonCollectionReplications
	"	self new testCommonCollectionReplications	"

	| all |
	(all := self collectionInstances) do: [:ea |
		self assert: (SrpConfiguration default replicate: ea) = ea.
	].
	self assert: (SrpConfiguration default replicate: all) = all.
	self assert: (SrtSequenceCheckConfiguration new replicate: all) = all.! !

!SrtPortableMappingTest methodsFor: 'SUnit Tests' stamp: ''!
testDateReplications
	"self new testDateReplications	"

	| d |
	d := Date today.
	self assert: (SrpConfiguration default replicate: d) = d.
	self assert: (SrpNonMappingConfiguration new replicate: d) = d.
	self assert: (SrtSequenceCheckConfiguration new replicate: d) = d.! !

!SrtPortableMappingTest methodsFor: 'SUnit Tests' stamp: ''!
testTimeReplications
	"self new testTimeReplications	"

	| t portableTimeClass |
	t := Time now.
	self assert: (SrpConfiguration default replicate: t) = t.
	self assert: (SrpNonMappingConfiguration new replicate: t) = t.
	self assert: (SrtSequenceCheckConfiguration new replicate: t) = t.

	"This tests the ability to convert a loaded time with fractional
	 seconds. If this fails then check #timeClassFromSeconds: for your
	 configuration."
	(portableTimeClass := self portal behaviorForName: #PmrTime) isNil ifFalse: [
		self assert: (portableTimeClass fromSeconds: (10/3)) asNative asSeconds >= 3.
		"A failure in this test may indicate a loss of precision in the time conversion."
		self assert: (portableTimeClass newFromTime: t) asNative = t.
	].! !

SrtSrpTest subclass: #SrtProxyTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtProxyTest methodsFor: 'SUnit Tests' stamp: ''!
testNestedObjectContentsNotSaved
	"	self new testNestedObjectContentsNotSaved.	"

	"This ensures contents of the object made into a proxy
	 is not saved."	

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: obj key ];
		fromProxyBlock: [:obj | loadedProxies add: obj ].
	object := 	(1 to: 5) 
		inject: 'lastValue' 
		into: [:ans :ea | SrtProxyTestObject key: ea value: ans ].
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 1.
	self assert: savedProxies first = 4.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 1.
	self assert: loadedProxies first = 4.
	self assert: loaded value == loadedProxies first.
	^loaded
	
		
	
! !

!SrtProxyTest methodsFor: 'SUnit Tests' stamp: ''!
testNestedProxyContentsIsSaved
	"	self new testNestedProxyContentsIsSaved.	"

	"This ensures contents of the object made into a proxy
	 is not saved."	

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: (Array with: obj) ];
		fromProxyBlock: [:obj | loadedProxies add: obj first ].
	object := 	(1 to: 5) 
		inject: 'lastValue' 
		into: [:ans :ea | SrtProxyTestObject key: ea value: ans ].
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 5.
	self assert: savedProxies last first = 'lastValue'.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 5.
	self assert: loadedProxies first = 'lastValue'.
	^loaded
	
		
	
! !

!SrtProxyTest methodsFor: 'SUnit Tests' stamp: ''!
testProxyIdentityNotPreserved
	"	self new testProxyIdentityNotPreserved.	"

	"This verifies that SRP isn't maintaining proxy identity.
	 It is up to the application to maintain proxy identity if that
	 is desired."

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: obj asSymbol ];
		fromProxyBlock: [:obj | loadedProxies add: obj asString ].
	object := SrtProxyTestObject key: 'key' value: 'value'.
	object := Array 
		with: object 
		with: object
		with: (SrtProxyTestObject key: object key value: object value).
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 2.
	self assert: savedProxies first = #value.
	self assert: savedProxies last = #value.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 2.
	self assert: loadedProxies first = 'value'.
	self assert: loadedProxies last = 'value'.
	self assert: (loaded at: 1) value == loadedProxies first.
	self assert: (loaded at: 3) value == loadedProxies last.
	self assert: (loaded at: 1) == (loaded at: 2).
	self assert: (loaded at: 1) ~~ (loaded at: 3).
	^loaded
	
		
	
! !

!SrtProxyTest methodsFor: 'SUnit Tests' stamp: ''!
testSharedValueProxyFirst
	"	self new testSharedValueProxyFirst.	"

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: obj asSymbol ];
		fromProxyBlock: [:obj | loadedProxies add: obj asString ].
	object := Association key: 'key' value: 'value'.
	object := SrtProxyTestObject key: object value: object value.
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 1.
	self assert: savedProxies first = #value.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 1.
	self assert: loadedProxies first = 'value'.
	self assert: loaded value == loadedProxies first.
	"This test has no reference index so the value loaded shouldn't be identical."
	self assert: loaded key value = loaded value.
	self assert: loaded key value ~~ loaded value.
	^loaded
	
		
	
! !

!SrtProxyTest methodsFor: 'SUnit Tests' stamp: ''!
testSharedValueProxyLast
	"	self new testSharedValueProxyLast.	"

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: obj asSymbol ];
		fromProxyBlock: [:obj | loadedProxies add: obj asString ].
	object := SrtProxyTestObject key: 'key' value: 'value'.
	object := Association key: object value: object value.
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 1.
	self assert: savedProxies first = #value.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 1.
	self assert: loadedProxies first = 'value'.
	self assert: loaded key value == loadedProxies first.
	"This test has no reference index so the value loaded shouldn't be identical."
	self assert: loaded key value = loaded value.
	self assert: loaded key value ~~ loaded value.
	^loaded
	
		
	
! !

!SrtProxyTest methodsFor: 'SUnit Tests' stamp: ''!
testSimpleProxy
	"	self new testSimpleProxy.	"

	| object config data savedProxies loaded loadedProxies |
	savedProxies := OrderedCollection new.
	loadedProxies := OrderedCollection new.
	config := SrtProxyTestConfiguration new
		toProxyBlock: [:obj | savedProxies add: obj asSymbol ];
		fromProxyBlock: [:obj | loadedProxies add: obj asString ].
	object := SrtProxyTestObject key: 'key' value: 'value'.
	data := config saveObjectAsText: object.
	self assert: savedProxies size = 1.
	self assert: savedProxies first = #value.
	self assert: loadedProxies isEmpty.
	loaded := config loadObjectFromText: data.
	self assert: loadedProxies size = 1.
	self assert: loadedProxies first = 'value'.
	self assert: loaded value == loadedProxies first.
	^loaded
	
		
	
! !

SrtSrpTest subclass: #SrtReadWriteTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtReadWriteTest methodsFor: 'SUnit Tests' stamp: ''!
testConsistentWrite
	"	self new testConsistentWrite.	"

	| obj pByteArray byteArray |
	obj := self complexObject.
	pByteArray := nil.
	1 to: 5 do: [:i |
		byteArray := SrpConfiguration default saveObjectAsBinary: obj.
		pByteArray isNil ifFalse: [
			self assert: pByteArray = byteArray.
		].
		pByteArray := byteArray.
	].
	^pByteArray! !

!SrtReadWriteTest methodsFor: 'SUnit Tests' stamp: ''!
testConsistentWriteRead
	"	self new testConsistentWriteRead.	"

	| obj pByteArray byteArray |
	obj := self complexObject.
	pByteArray := nil.
	1 to: 5 do: [:i |
		byteArray := SrpConfiguration default saveObjectAsBinary: obj.
		pByteArray isNil ifFalse: [
			self assert: pByteArray = byteArray.
		].
		obj := SrpConfiguration default loadObjectFromBinary: byteArray.
		pByteArray := byteArray.
	].
	^pByteArray! !

!SrtReadWriteTest methodsFor: 'SUnit Tests' stamp: ''!
testSaveLoadSequences
	"	self new testSaveLoadSequences.	"

	SrtSequenceCheckConfiguration new
		replicate: self complexObject.
! !

SrtSrpTest subclass: #SrtRecursionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtRecursionTest methodsFor: 'Test Support Code' stamp: ''!
arrayToOcRecursive
	"Warning: Some dialects (like VW) have trouble printing recursive structures."

	| nastyObject |
	nastyObject := Array new: 3.
	1 to: 3 do: [:index |
		nastyObject
			at: index
			put: OrderedCollection new.
		(nastyObject at: index) add: nastyObject.
	].
	^nastyObject
! !

!SrtRecursionTest methodsFor: 'Test Support Code' stamp: ''!
isValidFirstLevelRecursive: anObject

	anObject detect: [:ea | ea first ~~ anObject ] ifNone: [^true].
	^false.
! !

!SrtRecursionTest methodsFor: 'Test Support Code' stamp: ''!
isValidShortRecursive: anObject
	"	self new isValidShortRecursive: self new shortRecursiveArray.	"

	anObject detect: [:ea | ea ~~ anObject ] ifNone: [^true].
	^false
! !

!SrtRecursionTest methodsFor: 'Test Support Code' stamp: ''!
ocToArrayRecursive
	"Warning: Some dialects (like VW) have trouble printing recursive structures."

	| nastyObject |
	nastyObject := OrderedCollection new: 30.
	1 to: nastyObject size do: [:index |
		nastyObject
			at: index
			put: (Array with: nastyObject).
	].
	^nastyObject
! !

!SrtRecursionTest methodsFor: 'Test Support Code' stamp: ''!
shortArrayRecursive
	"Warning: Some dialects (like VW) have trouble printing recursive structures."

	| nastyObject |
	nastyObject := Array new: 30.
	1 to: nastyObject size do: [:index |
		nastyObject
			at: index
			put: nastyObject.
	].
	^nastyObject
! !

!SrtRecursionTest methodsFor: 'Test Support Code' stamp: ''!
shortOcRecursive
	"Warning: Some dialects (like VW) have trouble printing recursive structures."

	| nastyObject |
	nastyObject := OrderedCollection new.
	30 timesRepeat: [nastyObject add: nastyObject].
	^nastyObject
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testArrayToOcRecursive
	"	 self new testArrayToOcRecursive.	"

	| testObject |
	testObject := Association key: #key value: self arrayToOcRecursive.
	self assert: (self isValidFirstLevelRecursive: testObject value).
	self assert: (self isValidFirstLevelRecursive: (SrpConfiguration new  replicate: testObject) value).
	self assert: (self isValidFirstLevelRecursive: (SrpBase64Configuration new replicate: testObject) value).
	self assert: (self isValidFirstLevelRecursive: (SrpNonMappingConfiguration new replicate: testObject) value).
	self assert: (self isValidFirstLevelRecursive: (SrtSequenceCheckConfiguration new replicate: testObject) value).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testArrayToOcRootRecursive
	"	 self new testArrayToOcRootRecursive.	"

	| testObject |
	testObject := self arrayToOcRecursive.
	self assert: (self isValidFirstLevelRecursive: testObject).
	self assert: (self isValidFirstLevelRecursive: (SrpConfiguration new  replicate: testObject)).
	self assert: (self isValidFirstLevelRecursive: (SrpBase64Configuration new replicate: testObject)).
	self assert: (self isValidFirstLevelRecursive: (SrpNonMappingConfiguration new replicate: testObject)).
	self assert: (self isValidFirstLevelRecursive: (SrtSequenceCheckConfiguration new replicate: testObject)).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testOcToArrayRecursive
	"	 self new testOcToArrayRecursive.	"

	| testObject |
	testObject := Association key: #key value: self ocToArrayRecursive.
	self assert: (self isValidFirstLevelRecursive: testObject value).
	self assert: (self isValidFirstLevelRecursive: (SrpConfiguration new  replicate: testObject) value).
	self assert: (self isValidFirstLevelRecursive: (SrpBase64Configuration new replicate: testObject) value).
	self assert: (self isValidFirstLevelRecursive: (SrpNonMappingConfiguration new replicate: testObject) value).
	self assert: (self isValidFirstLevelRecursive: (SrtSequenceCheckConfiguration new replicate: testObject) value).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testOcToArrayRootRecursive
	"	 self new testOcToArrayRootRecursive.	"

	| testObject |
	testObject := self ocToArrayRecursive.
	self assert: (self isValidFirstLevelRecursive: testObject).
	self assert: (self isValidFirstLevelRecursive: (SrpConfiguration new  replicate: testObject)).
	self assert: (self isValidFirstLevelRecursive: (SrpBase64Configuration new replicate: testObject)).
	self assert: (self isValidFirstLevelRecursive: (SrpNonMappingConfiguration new replicate: testObject)).
	self assert: (self isValidFirstLevelRecursive: (SrtSequenceCheckConfiguration new replicate: testObject)).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testShortArrayRecursive
	"	 self new testShortArrayRecursive.	"

	| testObject |
	testObject := Association key: #key value: self shortArrayRecursive.
	self assert: (self isValidShortRecursive: testObject value).
	self assert: (self isValidShortRecursive: (SrpConfiguration new  replicate: testObject) value).
	self assert: (self isValidShortRecursive: (SrpBase64Configuration new replicate: testObject) value).
	self assert: (self isValidShortRecursive: (SrpNonMappingConfiguration new replicate: testObject) value).
	self assert: (self isValidShortRecursive: (SrtSequenceCheckConfiguration new replicate: testObject) value).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testShortArrayRootRecursive
	"	 self new testShortArrayRootRecursive.	"

	| testObject |
	testObject := self shortArrayRecursive.
	self assert: (self isValidShortRecursive: testObject).
	self assert: (self isValidShortRecursive: (SrpConfiguration new  replicate: testObject)).
	self assert: (self isValidShortRecursive: (SrpBase64Configuration new replicate: testObject)).
	self assert: (self isValidShortRecursive: (SrpNonMappingConfiguration new replicate: testObject)).
	self assert: (self isValidShortRecursive: (SrtSequenceCheckConfiguration new replicate: testObject)).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testShortOcRecursive
	"	 self new testShortOcRecursive.	"

	| testObject |
	testObject := Association key: #key value: self shortOcRecursive.
	self assert: (self isValidShortRecursive: testObject value).
	self assert: (self isValidShortRecursive: (SrpConfiguration new  replicate: testObject) value).
	self assert: (self isValidShortRecursive: (SrpBase64Configuration new replicate: testObject) value).
	self assert: (self isValidShortRecursive: (SrpNonMappingConfiguration new replicate: testObject) value).
	self assert: (self isValidShortRecursive: (SrtSequenceCheckConfiguration new replicate: testObject) value).
! !

!SrtRecursionTest methodsFor: 'SUnit Tests' stamp: ''!
testShortOcRootRecursive
	"	 self new testShortOcRootRecursive.	"

	| testObject |
	testObject := self shortOcRecursive.
	self assert: (self isValidShortRecursive: testObject).
	self assert: (self isValidShortRecursive: (SrpConfiguration new  replicate: testObject)).
	self assert: (self isValidShortRecursive: (SrpBase64Configuration new replicate: testObject)).
	self assert: (self isValidShortRecursive: (SrpNonMappingConfiguration new replicate: testObject)).
	self assert: (self isValidShortRecursive: (SrtSequenceCheckConfiguration new replicate: testObject)).
! !

!SrtSrpTest classSide methodsFor: 'Test Support Code' stamp: ''!
complexObject
	"Answer something big and complex for testing. A good test
	 object would have lots of recursion and references to identical
	 objects. Use the parse nodes produced by the RtcCompiler
	 if that is loaded in the image."

	| aClass |
	"This gives a huge and complex object, but it uses a parser that isn't available for every dialect."
	aClass := SrpPortal behaviorForName: #RtcCompiler.
	aClass isNil ifFalse: [
		^(aClass class >> #parseSTElementsFromString:) asRtcCompiledMethod.
	].

^(SrpConfiguration new loadObjectFromText: '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==')! !

!SrtSrpTest classSide methodsFor: 'Test Support Code' stamp: ''!
lessComplexObject
	"Need a big, portable object here. This is just a start."

	^OrderedCollection new
		add: #(1 abc (4 5 6) ghi);
		add: (Array with: true with: nil with: Object with: 1.0);
		add: (Association new key: 1; value:'one'; yourself);
		add: (Dictionary new at: #a put: $A;at: #b put: $B; yourself);
		add: Date today;
		yourself
! !

!SrtSrpTest classSide methodsFor: 'Test Support Code' stamp: ''!
portal
	"This accessor is defined so that the portal can be accessed by subclasses
	 without introducing namespace visibility problems in VW."

	^SrpPortal! !

!SrtSrpTest classSide methodsFor: 'Test running' stamp: ''!
smalltalkTestPrefix

	| type |
	type := self portal vmTypeName.
	type = #VisualAge ifTrue: [^'vaTest'].
	type = #VisualWorks ifTrue: [^'vwTest'].
	type = #VisualSmalltalk ifTrue: [^'vsTest'].
	type = #Dolphin ifTrue: [^'dolTest'].
	type = #'GemStone/S' ifTrue: [^'gssTest'].
	type = #Squeak ifTrue: [^'sqkTest'].
	type = #'Smalltalk/X' ifTrue: [^'stxTest'].
	type = #SmalltalkAgents ifTrue: [^'qksTest'].
	type = #SmalltalkMT ifTrue: [^'smtTest'].
	type = #SmalltalkExpress ifTrue: [^'stvTest'].
	type = #ObjectStudio ifTrue: [^'ostTest'].
	^'Unknown'! !

!SrtSrpTest classSide methodsFor: 'Test running' stamp: ''!
suite
	| testSelectors result configMatch |
	configMatch := self smalltalkTestPrefix, '*'.
	testSelectors := self sunitSelectors select: [:each |
		('test*' sunitMatch: each) or: [configMatch sunitMatch: each]
	].
	result := (self portal behaviorForNamespacePath: #(XProgramming SUnit TestSuite)) new.
	testSelectors do: [:each | result addTest: (self selector: each)].
	^result! !

!SrtSrpTest methodsFor: 'Test Support Code' stamp: ''!
complexObject
	"Answer something big and complex for testing. A good test
	 object would have lots of recursion and references to identical
	 objects."

	^self class complexObject! !

!SrtSrpTest methodsFor: 'Test Support Code' stamp: ''!
lessComplexObject
	"Need a big, portable object here."

	^self class lessComplexObject! !

!SrtSrpTest methodsFor: 'Test Support Code' stamp: ''!
portal
	"This accessor is defined so that the portal can be accessed by subclasses
	 without introducing namespace visibility problems in VW."

	^self class portal! !

!SrtSrpTest methodsFor: 'Test Support Code' stamp: ''!
reportComment: commentString
	"Messages normally written to the transcript are routed through this method."

	| transcript |
	(self respondsTo: #cstMessenger)
		ifTrue: [^self cstMessenger comment: commentString].
	(transcript := SrpPortal behaviorForName: #Transcript) isNil
		ifFalse: [transcript cr; show: commentString].


	! !

SrtSrpTest subclass: #SrtStreamEncoding
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64AnswersString
	"Ensures that SrpBase64Configuration answers a string
	 instead of a byte array."

	| obj result1 result2 |
	obj := Object new.
	result1 := SrpConfiguration default newStream text
		nextPut: obj; 
		streamContents.
	result2 := SrpBase64Configuration default newStream 
		nextPut: obj; 
		streamContents.
	self assert: (result1 class = result2 class).
	self assert: (result1 = result2).
	self assert: (result1 isKindOf: String).
	self assert: (result2 isKindOf: String).
	self assert: (result1 class ~= (self portal byteArrayClassWithAll: #()) class).
	self assert: (result2 class ~= (self portal byteArrayClassWithAll: #()) class).
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64ConversionTables
	"	 self new testBase64ConversionTables 	"

	| inst |
	inst := SrpBase64 new.
	0 to: 2r111111 do: [:int |
		self assert: int = (inst base64ForBase256: (inst base256ForBase64: int)).
	].! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64DecodeOfMixedData
	"	 self new testBase64DecodeOfMixedData 	"

	| inst stream bytes validBytes |
	validBytes := (SrpBase64 validCharacters, '=') asByteArray.
	bytes := (self portal byteArrayClassWithAll: (0 to: 255))
		reject: [:ea | validBytes includes: ea ].
	stream := ReadWriteStream on: ByteArray new.
	inst := SrpBase64 new srpStream: stream.
	bytes do: [:byte |
		inst nextPut: byte.
		stream nextPut: byte.
	].
	self assert: inst contents = bytes.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64EncodeAndDecodeAllBytes
	"	 self new testBase64EncodeAndDecodeAllBytes 	"

	| inst stream bytes |
	bytes := self portal byteArrayClassWithAll: (0 to: 255).
	stream := ReadWriteStream on: ByteArray new.
	inst := SrpBase64 new srpStream: stream.
	inst nextPutAll: bytes.
	self assert: inst contents = bytes.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64EncodeAndDecodeValidCharacters
	"	 self new testBase64EncodeAndDecodeValidCharacters 	"

	| inst stream |
	stream := ReadWriteStream on: ByteArray new.
	inst := SrpBase64 new srpStream: stream.
	inst nextPutAll: (self portal byteArrayClassWithAll: SrpBase64 validCharacters).
	self assert: inst contents = (self portal byteArrayClassWithAll: SrpBase64 validCharacters).! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64FinishByteGroup
	"	 self new testBase64FinishByteGroup 	"

	| inst stream strings |
	strings := #('ABC' 'ABC' 'AB' 'A' 'ABCD').
	stream := self portal newReadOverwriteStreamWith: 'abcdefghijklmnopqrstuvwxyz' asByteArray.
	inst := SrpBase64 new srpStream: stream.
	strings do: [:ea | inst nextPutAll: (ea asByteArray); srpFinishByteGroupWrite ].
	self assert: (self portal byteArrayAsString: stream contents) = 'QUJDQUJDQUI=QQ==QUJDRA==yz'.
	^inst contents asString.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64Flush
	"	 self new testBase64Flush 	"

	| inst stream strings |
	strings := #('ABC' 'ABC' 'AB' 'A' 'ABCD').
	stream := self portal newReadOverwriteStreamWith: 'abcdefghijklmnopqrstuvwxyz' asByteArray.
	inst := SrpBase64 new srpStream: stream.
	strings do: [:ea | inst nextPutAll: (ea asByteArray); flush ].
	self assert: (self portal byteArrayAsString: stream contents) = 'QUJDQUJDQUJBQUJDRLstuvwxyz'.
	^inst contents asString.
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64FlushInPlace
	"	 self new testBase64FlushInPlace 	"

	| inst stream strings |
	strings := #('ABC' 'ABC' 'AB' 'A' 'ABCD').
	stream := self portal newReadOverwriteStreamWith: 'abcdefghijklmnopqrstuvwxyz' asByteArray.
	inst := SrpBase64 new srpStream: stream.
	strings do: [:ea | inst nextPutAll: (ea asByteArray); srpFlushInPlace ].
	self assert: (self portal byteArrayAsString: stream contents) = 'QUJDQUJDQUJBQUJDRLstuvwxyz'.
	^inst contents asString.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBase64FlushInPlacePreservesExistingData
	"	 self new testBase64FlushInPlacePreservesExistingData 	"

	| inst stream strings |
	strings := #('ABC' 'ABC' 'AB' 'A' 'ABCD').
	stream := self portal newReadOverwriteStreamWith: 'abcdefghijklmnopqrstuvwxyz' asByteArray.
	inst := SrpBase64 new srpStream: stream.
	strings do: [:ea | inst nextPutAll: (ea asByteArray); skip: 1 ].
	self assert: (self portal byteArrayAsString: stream contents) = 'QUJDeQUJDjQUKJBpQUJDRPwxyz'.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testBasicReadWriteStream
	"This tests to ensure ReadWriteStream instances behave as SRP expects them to."

	"	self new testBasicReadWriteStream.	"

	| s temp |
	s := SrpPortal newReadOverwriteStreamWith: String new.
	self assert: s isEmpty.
	self assert: s atEnd.
	s nextPut: $A.
	self assert: s atEnd.
	s nextPutAll: 'BC'.
	self assert: s atEnd.
	s reset.
	self assert: s atEnd not.
	s next; next.
	self assert: s atEnd not.
	s next.
	self assert: s atEnd.
	temp := s contents.
	self assert: temp = 'ABC'.
	temp := s reset; contents.
	self assert: temp = 'ABC'.

	s := SrpPortal newReadOverwriteStreamWith: 'abcdef' copy.
	"self assert: s isEmpty not."		"VW would fail this"
	self assert: s atEnd not.
	s nextPut: $A.
	self assert: s atEnd not.
	s nextPutAll: 'BC'.
	self assert: s atEnd not.
	s reset.
	self assert: s atEnd not.
	s next; next; next.
	self assert: s atEnd not.
	self assert: s next = $d.
	self assert: s atEnd not.
	self assert: s next = $e.
	self assert: s atEnd not.
	self assert: s next = $f.
	self assert: s atEnd.
	temp := s contents.
	self assert: temp = 'ABCdef'.
	self assert: s atEnd.
	temp := s reset; contents.
	self assert: temp = 'ABCdef'.
	self assert: s atEnd not.

	s := SrpPortal newReadOverwriteStreamWith: 'abcdef' copy.
	temp := s contents.
	self assert: temp = 'abcdef'.
	s reset.
	s position: (s position + temp size).
	self assert: s atEnd.
	s nextPut: $G.
	self assert: s atEnd.
	s nextPutAll: 'HI'.
	self assert: s atEnd.
	temp := s contents.
	self assert: temp = 'abcdefGHI'.
	self assert: s atEnd.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testFixedWidthUnsigned
	"	self new testFixedWidthUnsigned.	"

	| stream |
	stream := SrpConfiguration new newStream.
	stream writer nextPut: $A.
	stream reset.
	stream reader next.
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testNboEncodingDiffers
	"	self new testNboEncodingDiffers.	"

	| cont1 cont2 obj |
	obj := self complexObject.
	cont1 := SrtNboConfiguration new newStream 
		nextPut: obj;
		streamContents.
	cont2 := SrpConfiguration default newStream 
		nextPut: obj;
		streamContents.
	self deny: cont1 = cont2.
	^Array
		with: cont1
		with: cont2	! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testNboEncodingWorks
	"	self new testNboEncodingWorks.	"

	| cont1 obj |
	obj := self complexObject.
	"Should error if there is a problem."
	^SrtNboConfiguration new replicate: obj.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testNextPutInteger
	"	self new testNextPutInteger.	"

	| bytes expectedBytes |
	bytes := (SrpObjectStream new writer
		nextPut: 2r1010101010101010101010101010101;
		stream) contents.

	expectedBytes := self portal byteArrayClassWithAll: #(0 2 1 49 40 76 97 114 103 101 80 111 115 105 116 105 118 101 73 110 116 101 103 101 114 1 170 213 170 213 10).
	((#(VisualAge Dolphin) includes: self portal vmTypeName)
	or: [self portal vmTypeName == #'Smalltalk/X'])	"ObjectStudio can't parse # in a literal array."
		ifTrue: [expectedBytes := self portal byteArrayClassWithAll: #(0 2 1 49 24 76 97 114 103 101 73 110 116 101 103 101 114 1 170 213 170 213 10)].
	(self portal vmTypeName = #VisualWorks and: [self portal isNamespaceImage]) 
		ifTrue: [expectedBytes := self portal byteArrayClassWithAll: #(0 2 1 49 5 4 67 111 114 101 20 76 97 114 103 101 80 111 115 105 116 105 118 101 73 110 116 101 103 101 114 1 170 213 170 213 10)].

	self assert: bytes = expectedBytes.
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveCharacterArray
	"	self new testSaveCharacterArray.	"

	| string bytes expected |
	string := 'Hello', self portal platformLineDelimiter, 'Joe'.
	expected := (ByteArray with: string size * 2), string asByteArray.
	bytes := (SrpObjectStream new writer
		saveCharacterArray: string;
		stream) contents.
	"The saving of a characterArray instead of a string is inherently configuration
	 specific so we create a byte array we are expecting to find."
	self assert: bytes = expected.! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveCharacterArrayArray
	"	self new testSaveCharacterArrayArray.	"

	| bytes |
	bytes := (SrpObjectStream new writer 
		saveCharacterArrayArray: #('Cat');
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(6 67 97 116)).
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveSigned
	"	self new testSaveSigned.	"

	| bytes expectedBytes |
	bytes := (SrpObjectStream new writer
		saveSigned: 2r10101010101010;
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(212 170 1)).

	bytes := (SrpObjectStream new writer
		saveSigned: 5;
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(10)).

	expectedBytes := self portal byteArrayClassWithAll: #(0 2 1 49 24 83 109 97 108 108 73 110 116 101 103 101 114 1 10).	
	(self portal vmTypeName = #VisualWorks and: [self portal isNamespaceImage]) 
		ifTrue: [expectedBytes := self portal byteArrayClassWithAll: #(0 2 1 49 5 4 67 111 114 101 12 83 109 97 108 108 73 110 116 101 103 101 114 1 10)].

	bytes := (SrpNonMappingConfiguration new newStream binary writer
		saveState: 5;
		stream) contents.

	self assert: bytes = expectedBytes.

! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveString
	"	self new testSaveString.	"

	| bytes |
	bytes := (SrpObjectStream new writer
		saveString: 'Hello';
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(10 72 101 108 108 111)).

	bytes := (SrpObjectStream new writer
		saveString: ('Hello', self portal platformLineDelimiter, 'Joe');
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(5 5 3 72 101 108 108 111 13 10 74 111 101)).! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveStringArray
	"	self new testSaveStringArray.	"

	| bytes |
	bytes := (SrpObjectStream new writer 
		saveStringArray: #('Cat' 'Bat');
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(2 6 67 97 116 6 66 97 116)).
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveUnsigned
	"	self new testSaveUnsigned.	"

	| bytes |
	bytes := (SrpObjectStream new writer 
		saveUnsigned: 2r10101010101010;
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(170 85)).! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveUnsignedArray
	"	self new testSaveUnsignedArray.	"

	| bytes |
	bytes := (SrpObjectStream new writer 
		saveUnsignedArray: #(502 5024 50243);
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(3 246 3 160 39 195 136 3)).! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveUnsignedArrayWidth
	"	self new testSaveUnsignedArrayWidth.	"

	| bytes |
	bytes := (SrpObjectStream new writer 
		saveUnsignedArray: #(502 5024 50243) width: 3;
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(3 246 131 0 160 167 0 195 136 3)).
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testSaveUnsignedWidth
	"	self new testSaveUnsignedWidth.	"

	| bytes |
	bytes := (SrpObjectStream new writer 
		saveUnsigned: 2r10101010101010 width: 3;
		stream) contents.
	self assert: bytes = (self portal byteArrayClassWithAll: #(170 213 0)).! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testUnsigned
	"	self new testUnsigned.	"

	| stream marshal array loaded |
	array := #(0 16 126 127 128 132 256 1023 1024 1025 16382 16383).
	stream := SrpConfiguration new newStream.
	marshal := stream writer.
	array do: [:ea | marshal saveUnsigned: ea ].
	stream reset.
	marshal := stream reader.
	array do: [:ea | 
		loaded := marshal loadUnsigned.
		self assert: loaded = ea.
	].
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testUnsignedArray
	"	self new testUnsignedArray.	"

	| stream array loaded |
	array := #(0 16 126 127 128 132 256 1023 1024 1025 16382 16383).
	stream := SrpConfiguration new newStream.
	stream writer saveUnsignedArray: array.
	stream reset.
	loaded := stream reader loadUnsignedArray.
	self assert: loaded = array.
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testUnsignedArrayWidth
	"	self new testUnsignedArrayWidth.	"

	| stream array width loaded |
	array := #(0 16 126 127 128 132 256 1023 1024 1025 16382 16383).
	stream := SrpConfiguration new newStream.
	stream writer saveUnsignedArray: array width: 5.
	stream reset.
	loaded := stream reader loadUnsignedArray.
	self assert: loaded = array.
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testUnsignedWidth
	"	self new testUnsignedWidth.	"

	| stream marshal array width loaded |
	array := #(0 16 126 127 128 132 256 1023 1024 1025 16382 16383).
	stream := SrpConfiguration new newStream.
	marshal := stream writer.
	width := array inject: 0 into: [:ans :ea | ans max: (marshal widthOfUnsigned: ea) ].
	array do: [:ea | marshal saveUnsigned: ea width: width ].

	"Try sequential reading."
	stream reset.
	marshal := stream reader.
	array do: [:ea | 
		loaded := marshal loadUnsigned.
		self assert: loaded = ea.
	].

	"Try positional reading."
	marshal := stream reader.
	1 to: array size do: [:i | 
		stream position: ((i - 1) * width).
		loaded := marshal loadUnsigned.
		self assert: loaded = (array at: i).
	].
! !

!SrtStreamEncoding methodsFor: 'SUnit Tests' stamp: ''!
testWrappedBase64
	"	self new testWrappedBase64.	"

	| stream inst |
	stream := ReadWriteStream on: ByteArray new.
	inst := SrpWrappedBase64 new srpStream: stream.
	self assert: inst srpCurrentLineLength = 0.
	inst nextPut: $A asInteger.
	self assert: inst srpCurrentLineLength = 1.
	inst nextPutAll: 'bcdefg' asByteArray.
	self assert: inst srpCurrentLineLength = 7.
	inst nextPut: 10.
	self assert: inst srpCurrentLineLength = 0.
	inst nextPutAll: 'hijkl' asByteArray.
	self assert: inst srpCurrentLineLength = 5.
! !

SrtSrpTest subclass: #SrtSubstitutionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtSubstitutionTest methodsFor: 'SUnit Tests' stamp: ''!
testDynamicMappingRules

	| saveCount loadCount answer |
	saveCount := loadCount := 0.
	answer := SrpConfiguration new
		beforeSavingAnyNamed: #(Symbol ByteSymbol) doWithContext: [:ea :context | saveCount := saveCount + 1. ea  ];
		afterLoadingAnyNamed: #(PmrSymbol Symbol ByteSymbol) doWithContext: [:ea :context | loadCount := loadCount + 1. ea asString ];
		replicate: #(name date).
	self assert: saveCount = 2.
	self assert: loadCount = 2.
	self assert: (answer at: 1) = 'name'.
	self assert: (answer at: 2) = 'date'.! !

!SrtSubstitutionTest methodsFor: 'SUnit Tests' stamp: ''!
testSelfReferencingSubstitution
	"	self new testSelfReferencingSubstitution.	"

	"Uses a mapping rule that will exchange a date for
	 an object that refers to the same date. Tests to see
	 that the date is actually saved instead of just a relative
	 reference to the substitute (would loop when loading)."
	
	| origDate loaded |
	origDate := Date today.
	loaded := (SrpConfiguration new
			addExchangeRules: (Array with: SrtSaveExchangeRule);
			yourself) newStream
		nextPut: (Array with: origDate with: origDate);
		reset;
		next.
	self assert: loaded first = origDate.
	self assert: loaded first == loaded last.

! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpAttribute: varNameString
	"This can be sent by mapping rules. It is a way to access named
	 variable slots regardless of whether accessor methods are defined."

	| s |
	^(s := self class srpSlotNamed: varNameString) = 0
		ifTrue: [nil]
		ifFalse: [self instVarAt: s]! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpAttribute: varNameString put: anObject
	"This can be sent by mapping rules. It is a way to access named
	 variable slots regardless of whether accessor methods are defined."

	| s |
	(s := self class srpSlotNamed: varNameString) = 0
		ifFalse: [self instVarAt: s put: anObject].
	^anObject! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpBasicAt: index
	"Answer the object at the variable index position of the receiver.
	 This can be sent by mapping rules. It is preferable to #basicAt: because
	 it works with special loader constructs."

	^self basicAt: index
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpBasicAt: index put: value
	"Store the Object argument value at the variable index position 
	 in the receiver. This can be sent by mapping rules. It is preferable to
	 #basicAt:put: because it works with special loader constructs."

	^self basicAt: index put: value
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpBasicSize
	"Answer the basic size of the receiver. This can be sent by
	 mapping rules. It is preferable to #basicSize because it
	 works with special loader constructs."

	^self basicSize
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpClass
	"Answer the class of the receiver. This can be sent by
	 mapping rules. It is preferable to #class because it
	 works with special loader constructs."

	^self class
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpDirectSaveOn: aSaver as: aMetastate
	"The receiver wants to save itself directly to aSaver instead
	 of using standard persistence routines. This can be more
	 efficient but requires proper load rules to restore."

	"Do nothing by default."! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpExtendedAttributes
	"Answer an object that is to be saved along with the receiver.
	 The object will be sent as an argument to #srpExtendedAttributes:
	 when loading. For flexibility it is recommended that you answer
	 a dictionary with key value pairs."

	^nil
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpExtendedAttributes: anObject
	"anObject was saved along with the receiver as a holder of additional
	 information related to the receiver. This method is sent when loading to
	 reassociate the two objects. anObject is customarily a dictionary with
	 key value pairs."

	"Do nothing by default."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpIndexableSize
	"Answer the basic size of the receiver. This can be sent by
	 mapping rules. It is preferable to #basicSize because it
	 works with special loader constructs."

	^self basicSize

	"[GemStone] #_basicSize is sent in GemStone; refer to #gemstoneSpecificFixupTopazScript."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpIndexableValueAt: index
	"Answer the object at the variable index position of the receiver.
	 This can be sent by mapping rules. It is preferable to #basicAt: because
	 it works with special loader constructs."

	^self basicAt: index
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpIndexableValueAt: index put: value
	"Store the Object argument value at the variable index position 
	 in the receiver. This can be sent by mapping rules. It is preferable to
	 #basicAt:put: because it works with special loader constructs."

	^self basicAt: index put: value
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpInstVarAt: index
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt: because it
	 works with special loader constructs."

	^self instVarAt: index
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpInstVarAt: anInteger put: anObject
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt:put: because
	 it works with special loader constructs."

	^self instVarAt: anInteger put: anObject
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpIsLoadReceptacle
	"Objects are temporarily wrapped with a load receptacle while
	 loading. This can be sent in load mapping rules or #srpLoaded
	 to test if the receiver is a load receptacle."

	^false
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpIsPlaceholder
	"Objects are temporarily wrapped with a placeholder while
	 loading. This can be sent in load mapping rules or #srpLoaded
	 to test if the receiver is a placeholder."

	^false
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpLoadPostMap: aContext
	"Answer an object that is to represent the receiver when
	 loaded. Keep in mind that slots of the receiver could 
	 still contain placeholders around an object that may
	 yet be mapped. Also be aware that when an object
	 references the receiver, and the receiver also references
	 that object, that the slots for that object may not be fully
	 loaded yet. If unsure of placeholders then declare an
	 #srpPostLoadAction instead."

	"Do nothing by default."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpLoadPreMap: aContext
	"Answer an object that is to represent the receiver when
	 loaded. Keep in mind that slots of the receiver could 
	 still contain placeholders around an object that may
	 yet be mapped. Also be aware that when an object
	 references the receiver, and the receiver also references
	 that object, that the slots for that object may not be fully
	 loaded yet. If unsure of placeholders then declare an
	 #srpPostLoadAction instead."

	"Do nothing by default."! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpNamedValueAt: index
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt: because it
	 works with special loader constructs."

	^self instVarAt: index
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpNamedValueAt: anInteger put: anObject
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt:put: because
	 it works with special loader constructs."

	^self instVarAt: anInteger put: anObject
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpPostLoadAction
	"Answer a command (directed message or block) that
	 will answer an object to take the place of the
	 receiver when sent #value. The command will be
	 sent #value after the root object has been completely
	 loaded. Answer nil if no further processing is required."

	^nil
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpPostLoadActionForLoader: aMarshaler
	"Answer a command (directed message or block) that
	 will answer an object to take the place of the
	 receiver when sent #value. The command will be
	 sent #value after the root object has been completely
	 loaded. Answer nil if no further processing is required."

	^nil
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpRecursiveReferencesShouldReferenceMapped
	"An original object has been mapped to the receiver, and while saving the receiver there
	 was a reference to the original object. Answer true (the default) if a reference to the
	 mapped object should be saved, or false if the original object should be serialized
	 without remapping. This method should only be overridden to answer false for instances
	 that are only meant for mapping operations--not normal kernel instances."

	^true
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpSavePostMap: aContext
	"Answer an object that is to represent the receiver when
	 saved. This is sent after class-based save mapping rules
	 are applied."

	"Do nothing by default."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpSavePreMap: aContext
	"Answer an object that is to represent the receiver when
	 saved. This is sent before class-based save mapping rules
	 are applied."

	"Do nothing by default."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpSavedAs: savedObject
	"Notification that the receiver was saved as savedObject."

	"Do nothing by default."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpStripLoadReceptacles: recursionSet
	"Private protocol. This is used to cleanup receptacles used for correctly
	 resolving changing object pointers when loading objects.
	 THIS non-recursive method is only run because the receiver
	 has been mutated while loading. This was necessary to prevent
	 rare problems doing special mutations to recursive structures."

	| item |
	(recursionSet includes: self) ifTrue: [^self].
	recursionSet add: self.
	"Strip from any fixed variables -- if there are any."
	1  to: self srpClass instSize do: [:instVarIndex | 
		(item := self srpInstVarAt: instVarIndex) srpIsLoadReceptacle ifTrue: [
			self srpInstVarAt: instVarIndex put: (item srpStripLoadReceptacles: recursionSet)
		].
	].
	"Strip from any indexed (pointer only) variables -- if there are any."
	self srpClass isPointers ifTrue: [
		1 to: self srpBasicSize do: [:indexedVarIndex |
			(item := self srpBasicAt: indexedVarIndex) srpIsLoadReceptacle ifTrue: [
				self srpBasicAt: indexedVarIndex put: (item srpStripLoadReceptacles: recursionSet).
			].
		].
	].
	^self
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpStripPlaceholders: recursionSet
	"Private protocol. This is used to cleanup placeholders used for correctly
	 resolving changing object pointers when loading objects.
	 THIS non-recursive method is only run because the receiver
	 has been mutated while loading. This was necessary to prevent
	 rare problems doing special mutations to recursive structures."

	| item |
	(recursionSet includes: self) ifTrue: [^self].
	recursionSet add: self.
	"Strip from any fixed variables -- if there are any."
	1  to: self srpClass instSize do: [:instVarIndex | 
		(item := self srpNamedValueAt: instVarIndex) srpIsPlaceholder ifTrue: [
			self srpNamedValueAt: instVarIndex put: (item srpStripPlaceholders: recursionSet)
		].
	].
	"Strip from any indexed (pointer only) variables -- if there are any."
	(SrpPortal behaviorIsPointers: self srpClass) ifTrue: [
		1 to: self srpIndexableSize do: [:indexedVarIndex |
			(item := self srpIndexableValueAt: indexedVarIndex) srpIsPlaceholder ifTrue: [
				self srpIndexableValueAt: indexedVarIndex put: (item srpStripPlaceholders: recursionSet).
			].
		].
	].
	^self
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpWithoutLoadReceptacle
	"This can be sent by mapping rules and in #srpLoaded 
	 when it is necessary to ensure an object is not wrapped
	 by a load receptacle when loading. If you ever end up with
	 a load receptacle after loading and mapping is completely
	 finished then this wasn't sent when it should have been. "

	"Do nothing by default."
! !

!Object methodsFor: '*SRP-BaseExt' stamp: ''!
srpWithoutPlaceholder
	"This can be sent by mapping rules and in #srpLoaded 
	 when it is necessary to ensure an object is not wrapped
	 by a placeholder when loading. If you ever end up with
	 a placeholder after loading and mapping is completely
	 finished then this wasn't sent when it should have been. "

	"Do nothing by default."
! !

Object subclass: #SrpActionItem
	instanceVariableNames: 'placeholder baseObject action'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpActionItem classSide methodsFor: 'Not categorized' stamp: ''!
forPlaceholder: aPlaceholder action: valueThingy

	^self new
		placeholder: aPlaceholder;
		action: valueThingy

! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
action

	^action! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
action: valueThingy

	action := valueThingy.! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
baseObject

	^baseObject! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
baseObject: anObject

	baseObject := anObject.! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
placeholder

	^placeholder! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
placeholder: aPlaceholder

	placeholder := aPlaceholder.
	baseObject := aPlaceholder srpWithoutPlaceholder.
! !

!SrpActionItem methodsFor: 'Not categorized' stamp: ''!
value

	| answer |
	answer := action value.
	baseObject == placeholder srpWithoutPlaceholder
		ifTrue: [ placeholder srpBaseObject: answer ].
	" ^self "
! !

Object subclass: #SrpConfiguration
	instanceVariableNames: 'exchangeRules metastates'
	classVariableNames: 'DefaultInstances MetastateTables'
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

SrpConfiguration subclass: #SrpBase64Configuration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpBase64Configuration classSide methodsFor: 'documentation' stamp: ''!
documentation
"
	Write operations with this configuration produce a text data in
	wrapped base64 encoding. This encoding is useful when sending
	objects in the body of an email. 

"! !

!SrpBase64Configuration methodsFor: 'data encoding' stamp: ''!
useBinaryAccess
	"Answer boolean whether the data stream contains binary or text data.
	 This is really a default value because a stream can be treated on
	 demand as either binary or text--regardless of this setting."

	^false
! !

!SrpBase64Configuration methodsFor: 'data encoding' stamp: ''!
wrapAppendedText
	"Answer boolean whether text written to the end of a stream
	 should be wrapped."

	^false
! !

!SrpConfiguration classSide methodsFor: 'default configuration' stamp: ''!
default
	"Answer the default instance for the class. This used to be in a 
	 class instance variable but ObjectStudio doesn't support them."

	"	self default.	"

	| newInstance |
	DefaultInstances isNil ifTrue: [DefaultInstances := Dictionary new].
	^DefaultInstances 
		at: self 
		ifAbsent: [DefaultInstances at: self put: (newInstance := self new). newInstance].

	"Note: [ObjectStudio] Dictionary>>#at:put: returns the dictionary instead of the object put."! !

!SrpConfiguration classSide methodsFor: 'default configuration' stamp: ''!
default: aConfig
	"This allows you to change the default configuration that is
	 in effect for the image. This isn't the normal way to apply
	 configurations, but it can be useful."

	DefaultInstances isNil ifTrue: [DefaultInstances := Dictionary new].
	DefaultInstances at: self put: aConfig.
! !

!SrpConfiguration classSide methodsFor: 'default configuration' stamp: ''!
defaultMetastateTables

	^Dictionary new
! !

!SrpConfiguration classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 An instance of this class is usually held in the Default class
 variable and is used to specify SRP configuration options. 

 You can use a custom set of options by creating a subclass
 of SrpConfiguration, overriding methods as desired, and then
 using your configuration to create instances of SrpObjectStream.
 The configuration itself is usually used to create new instances
 of the appropriate object stream type. Here are some examples
 of methods you will want to use to apply a configuration:

 ^SrpConfiguration default newStream.

 ^SrpConfiguration default onStream: aStream.

 ^SrpConfiguration default replicate: #(one two three).

 ^SrpConfiguration default saveObject: #(one two three).

 ^SrpConfiguration default loadObjectFrom: aCharacterCollection.

 Configurations offer you the ultimate in flexibility. Almost every aspect
 of SRP can be customized for whatever special purposes you have in
 mind. You can simultaneously use multiple configurations if you desire.

"! !

!SrpConfiguration classSide methodsFor: 'accessing' stamp: ''!
metastateTables

	MetastateTables isNil ifTrue: [MetastateTables := Dictionary new].
	^MetastateTables! !

!SrpConfiguration classSide methodsFor: 'accessing' stamp: ''!
metastateTables: newValue

	MetastateTables := newValue
! !

!SrpConfiguration classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize! !

!SrpConfiguration classSide methodsFor: 'SrpPortableObjectRepresentations' stamp: ''!
pmrDefaultClassNames
	"This only contains the most commonly used PMR classes."

	^#(
		 PmrDictionary PmrIdentityDictionary PmrOrderedCollection
		 PmrArray PmrString PmrSymbol PmrSet PmrIdentitySet
		 PmrDecimal PmrFloat
		 PmrDate PmrTime PmrDateTime
		 PmrMessage PmrDirectedMessage
		 PmrNamespaceReference PmrClassReference
		 PmrMetaclassReference PmrGlobalReference PmrMethodBinding
	 ).! !

!SrpConfiguration classSide methodsFor: 'SrpPortableObjectRepresentations' stamp: ''!
pmrDefaults1
	"Treat these definitions as a historical record. Do not modify them or else
	 you risk making objects unreadable that are optimized by them. This is kind 
	 of like an XML DTD file--it avoids having to store metastate
	 information along with state."

	"	 self pmrDefaults1 	"
	"	self recordsForClasses: self pmrDefaultClassNames	"
	"	self smalltalkSourceForClasses: self pmrDefaultClassNames.	"

	^(SrpObjectStream loadObjectFromText: '
AAIBBBBQbXJBcnJheQEUAAICAhhTcnBNZXRhc3RhdGUPBGNvZGUNaW5zdGFuY2VT
aGFwZQxiZWhhdmlvck5hbWUIYmVoYXZpb3IPYWxsSW5zdFZhck5hbWVzE25hbWV
kU3RvcmVUeXBlQ29kZXMIbWFwVGFibGUCAAIDMRhTbWFsbEludGVnZXID2AEAAg
QCIlNycE1ldGFzdGF0ZVNoYXBlDmludGVnZXIEAwgBAQACBQISUG1yU3ltYm9sD
HN0cmluZwUAAgZRElBtclN0cmluZwYaUG1yRGljdGlvbmFyeQACBwIiUG1yQ2xh
c3NSZWZlcmVuY2UFCW5hbWVzcGFjZQtjbGFzc1N5bWJvbAcAAggAHlVuZGVmaW5
lZE9iamVjdAgAAQQBAAgIAgPaAQQDCAEBBQYqUG1ySWRlbnRpdHlEaWN0aW9uYX
J5BwgAAQQBAAgIAgPcAQQDCAEBBQYoUG1yT3JkZXJlZENvbGxlY3Rpb24HCAABB
AEACAgCA94BBAMIAQEFBhBQbXJBcnJheQcIAAEEAQAICAID4AEEA6IBAQEFBhJQ
bXJTdHJpbmcHCAABBAEBBgxzdHJpbmcICAID4gEEAwQBAQUGElBtclN5bWJvbAc
IAAEEAQEAAQ0ICAID5AEEAwgBAQUGDFBtclNldAcIAAEEAQAICAID5gEEAwgBAQ
UGHFBtcklkZW50aXR5U2V0BwgAAQQBAAgIAgPoAQQDBAEBBQYUUG1yRGVjaW1hb
AcIAAEEAQIGEGludGVncmFsBgpzY2FsZQgIAgPqAQQDBAEBBQYQUG1yRmxvYXQH
CAABBAECBhJudW1lcmF0b3IGFmRlbm9taW5hdG9yCAgCA+wBBAMEAQEFBg5QbXJ
EYXRlBwgAAQQBAwYIeWVhcgYKbW9udGgGBmRheQgIAgPuAQQDBAEBBQYOUG1yVG
ltZQcIAAEEAQEGHGVsYXBzZWRTZWNvbmRzCAgCA/ABBAMEAQEFBhZQbXJEYXRlV
GltZQcIAAEEAQIGCGRhdGUGCHRpbWUICAID8gEEAwQBAQUGFFBtck1lc3NhZ2UH
CAABBAECBhBzZWxlY3RvcgYSYXJndW1lbnRzCAgCA/QBBAMEAQEFBiRQbXJEaXJ
lY3RlZE1lc3NhZ2UHCAABBAEDAAEOAAENBhByZWNlaXZlcggIAgP2AQQDBAEBBQ
YqUG1yTmFtZXNwYWNlUmVmZXJlbmNlBwgAAQQBAQYSbmFtZXNwYWNlCAgCA/gBB
AMEAQEFBiJQbXJDbGFzc1JlZmVyZW5jZQcIAAEEAQIAAQ0GFmNsYXNzU3ltYm9s
CAgCA/oBBAMEAQEFBipQbXJNZXRhY2xhc3NSZWZlcmVuY2UHCAABBAECAAEaAAE
NCAgCA/wBBAMEAQEFBiRQbXJHbG9iYWxSZWZlcmVuY2UHCAABBAECAAEmBghuYW
1lCAgCA/4BBAMEAQEFBiBQbXJNZXRob2RCaW5kaW5nBwgAAQQBAwABMwYMbWV0a
G9kBgZrZXkICA==')! !

!SrpConfiguration classSide methodsFor: 'SrpPortableObjectRepresentations' stamp: ''!
pmrDefaults2
	"Treat these definitions as a historical record. Do not modify them or else
	 you risk making objects unreadable that are optimized by them. This is kind 
	 of like an XML DTD file--it avoids having to store metastate
	 information along with state."

	"	 self pmrDefaults2 	"
	"	self recodeMetastates: self pmrDefaultClassNames	"
	"	self smalltalkSourceForClasses: self pmrDefaultClassNames.	"

	^(SrpObjectStream loadObjectFromText: '
AAIBBBBQbXJBcnJheQEUAAICAhhTcnBNZXRhc3RhdGUTBGNvZGUNaW5zdGFuY2VT
aGFwZQxiZWhhdmlvck5hbWUSdmVyc2lvbk51bWJlckFycmF5CGJlaGF2aW9yD2F
sbEluc3RWYXJOYW1lcxNuYW1lZFN0b3JlVHlwZUNvZGVzCG1hcFRhYmxlEmV4dG
VuZGVkQXR0cmlidXRlcwIAAgMxGFNtYWxsSW50ZWdlcgPYAQACBAIiU3JwTWV0Y
XN0YXRlU2hhcGUOaW50ZWdlcgQDCAEBAAIFAxJQbXJTeW1ib2wMc3RyaW5nAQUF
GlBtckRpY3Rpb25hcnkAAgYAHlVuZGVmaW5lZE9iamVjdAYAAgcCIlBtckNsYXN
zUmVmZXJlbmNlBQluYW1lc3BhY2ULY2xhc3NTeW1ib2wHBgABBAEABgYGAgPaAQ
QDCAEBBSpQbXJJZGVudGl0eURpY3Rpb25hcnkGBwYAAQQBAAYGBgID3AEEAwgBA
QUoUG1yT3JkZXJlZENvbGxlY3Rpb24GBwYAAQQBAAYGBgID3gEEAwgBAQUQUG1y
QXJyYXkGBwYAAQQBAAYGBgID4AEEAwYBAQUSUG1yU3RyaW5nBgcGAAEEAQEAAgg
DElBtclN0cmluZwxzdHJpbmcBBQgMc3RyaW5nAQEDCgYGAgPiAQQDBgEBBRJQbX
JTeW1ib2wGBwYAAQQBAQABDwEBAwoGBgID5AEEAwgBAQUMUG1yU2V0BgcGAAEEA
QAGBgYCA+YBBAMIAQEFHFBtcklkZW50aXR5U2V0BgcGAAEEAQAGBgYCA+gBBAME
AQEFFFBtckRlY2ltYWwGBwYAAQQBAggQaW50ZWdyYWwICnNjYWxlBgYGAgPqAQQ
DBAEBBRBQbXJGbG9hdAYHBgABBAECCBJudW1lcmF0b3IIFmRlbm9taW5hdG9yBg
YGAgPsAQQDBAEBBQ5QbXJEYXRlBgcGAAEEAQMICHllYXIICm1vbnRoCAZkYXkGB
gYCA+4BBAMEAQEFDlBtclRpbWUGBwYAAQQBAQgcZWxhcHNlZFNlY29uZHMGBgYC
A/ABBAMEAQEFFlBtckRhdGVUaW1lBgcGAAEEAQIICGRhdGUICHRpbWUGBgYCA/I
BBAMEAQEFFFBtck1lc3NhZ2UGBwYAAQQBAggQc2VsZWN0b3IIEmFyZ3VtZW50cw
YGBgID9AEEAwQBAQUkUG1yRGlyZWN0ZWRNZXNzYWdlBgcGAAEEAQMAAQ8AAQ4IE
HJlY2VpdmVyBgYGAgP2AQQDBAEBBSpQbXJOYW1lc3BhY2VSZWZlcmVuY2UGBwYA
AQQBAQgSbmFtZXNwYWNlBgYGAgP4AQQDBAEBBSJQbXJDbGFzc1JlZmVyZW5jZQY
HBgABBAECAAEOCBZjbGFzc1N5bWJvbAYGBgID+gEEAwQBAQUqUG1yTWV0YWNsYX
NzUmVmZXJlbmNlBgcGAAEEAQIAARwAAQ4GBgYCA/wBBAMEAQEFJFBtckdsb2Jhb
FJlZmVyZW5jZQYHBgABBAECAAEpCAhuYW1lBgYGAgP+AQQDBAEBBSBQbXJNZXRo
b2RCaW5kaW5nBgcGAAEEAQMAATcIDG1ldGhvZAgGa2V5BgYG')! !

!SrpConfiguration classSide methodsFor: 'default configuration' stamp: ''!
registerDefaultMetastateTables
	"	 self registerDefaultMetastateTables. 	"

	| register |
	register := SrpConfiguration metastateTables.
	SrpPortal
		keyedCollection: self defaultMetastateTables 
		associationsDo: [:assoc | register add: assoc ].

! !

!SrpConfiguration classSide methodsFor: 'SrpPortableObjectRepresentations' stamp: ''!
registerPmrDefaults1
	"	 self registerPmrDefaults1 	"

	^self metastateTables add: (Association key: 'PmrDefaults1' value: self pmrDefaults1).! !

!SrpConfiguration classSide methodsFor: 'SrpPortableObjectRepresentations' stamp: ''!
registerPmrDefaults2
	"	 self registerPmrDefaults2 	"

	^self metastateTables add: (Association key: 'PmrDefaults2' value: self pmrDefaults2).! !

!SrpConfiguration classSide methodsFor: 'default configuration' stamp: ''!
resetDefaults
	"	self resetDefaults.	"

	DefaultInstances := nil.
	SrpConfiguration metastateTables: nil.
	SrpPortableBehavior portal: nil.
! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
addExchangeRule: aRule
	"Used to add to the collection of selected class-based exchange
	 rules that are enabled for this configuration. Instance-defined
	 rules (instances of SrpDynamicMappingRule) are normally
	 added here; however, class-defined rules (subclasses of
	 SrpSaveMappingRule or SrpLoadMappingRule) can also be
	 added."

	self exchangeRules add: aRule.
! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
addExchangeRules: rules
	"Used to add to the collection of selected class-based exchange
	 rules that are enabled for this configuration. Instance-defined
	 rules (instances of SrpDynamicMappingRule) are normally
	 added here; however, class-defined rules (subclasses of
	 SrpSaveMappingRule or SrpLoadMappingRule) can also be
	 added."

	self exchangeRules addAll: rules.
! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
addMetastates: aCollection

	metastates addAll: aCollection! !

!SrpConfiguration methodsFor: 'exchange rules-instance based' stamp: ''!
afterLoadingAnyNamed: collectionOfClassNames doWithContext: twoArgBlock
	"This creates a configurable exchange rule that will apply to all
	 metastates named in collectionOfClassNames. The twoArgBlock
	 will be evaluated with each instance of the types named in the
	 collectionOfClassNames. The second argument passed to the	
	 block is the context the instance appeared in. The twoArgBlock
	 should return the instance or another object that is to be used instead."

	| newRule |
	newRule := self configurableExchangeRuleClass newForLoading
		applicableClassNames: collectionOfClassNames;
		command: twoArgBlock.
	newRule applicationPriority: 1000.
	self exchangeRules add: newRule.
	^newRule
	! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
allLoadExchangeRules
	"Answer a collection of all the class-based exchange rules that are
	 to be in effect when loading using this configuration."

	^Set new
		addAll: self dialectLoadExchangeRules;
		addAll: self defaultLoadExchangeRules;
		addAll: self selectedLoadExchangeRules;
		yourself.! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
allSaveExchangeRules
	"Answer a collection of all the class-based exchange rules that are
	 to be in effect when saving using this configuration."

	^Set new
		addAll: self dialectSaveExchangeRules;
		addAll: self defaultSaveExchangeRules;
		addAll: self selectedSaveExchangeRules;
		yourself.! !

!SrpConfiguration methodsFor: 'exchange rules-instance based' stamp: ''!
beforeSavingAnyNamed: collectionOfClassNames doWithContext: twoArgBlock
	"This creates a configurable exchange rule that will apply to all
	 metastates named in collectionOfClassNames. The twoArgBlock
	 will be evaluated with each instance of the types named in the
	 collectionOfClassNames. The second argument passed to the	
	 block is the context the instance appeared in. The twoArgBlock
	 should return the instance or another object that is to be used instead."

	| newRule |
	newRule := self configurableExchangeRuleClass newForSaving
		applicableClassNames: collectionOfClassNames;
		command: twoArgBlock.
	newRule applicationPriority: 1000.
	self exchangeRules add: newRule.
	^newRule
	! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
beingUsedForLoadContext: aLoadContext
	"Notification that aLoadContext is using the
	 receiver for configuration settings. This is a great
	 chance to do any special processing that couldn't
	 be achived through other overrides. Do nothing
	 by default."! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
beingUsedForObjectLoader: anObjectLoader
	"Notification that anObjectLoader is using the
	 receiver for configuration settings. This is a great
	 chance to do any special processing that couldn't
	 be achived through other overrides. Do nothing
	 by default."! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
beingUsedForObjectSaver: anObjectSaver
	"Notification that anObjectSaver is using the
	 receiver for configuration settings. This is a great
	 chance to do any special processing that couldn't
	 be achived through other overrides. Do nothing
	 by default."! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
beingUsedForObjectStream: anObjectStream
	"Notification that anObjectStream is using the
	 receiver for configuration settings. This is a great
	 chance to do any special processing that couldn't
	 be achived through other overrides. Do nothing
	 by default."! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
beingUsedForSaveContext: aSaveContext
	"Notification that aSaveContext is using the
	 receiver for configuration settings. This is a great
	 chance to do any special processing that couldn't
	 be achived through other overrides. Do nothing
	 by default."! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
configurableExchangeRuleClass
	"Answer an instance-based exchange rule that allows load and/or
	 save rules to be defined without creating class definitions."

	^SrpConfigurableRule
! !

!SrpConfiguration methodsFor: 'context creation' stamp: ''!
contextUserData
	"The context object for each traversal can be the holder of
	 user data collected or referenced during a traversal. User
	 data is expected to behave as a dictionary. Nil can be
	 returned by this method to avoid dictionary creation for every
	 traversal unless one is necessary."
	^nil
! !

!SrpConfiguration methodsFor: 'metastates' stamp: ''!
createdMetastate: aMetastate modeledAfter: aClass
	"Notification that aClass was used to create a metastate."
! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
defaultLoadExchangeRules
	"Answer a collection of class-defined class-based load
	 exchange rules that are intended to be active by default."

	| answer |
	answer := Set new.
	SrpPortal behaviorWithAllSubclasses: SrpSubstituteObject do: [:aRule |
		((aRule useWhenLoading and: [aRule isDefaultRule])
			ifTrue:  [aRule applicableClassNames isEmpty not]
			ifFalse: [self shouldApplyRule: aRule])
				ifTrue: [answer add: aRule].
	].
	^answer! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
defaultSaveExchangeRules
	"Answer a collection of class-defined class-based save
	 exchange rules that are intended to be active by default."

	| answer |
	answer := Set new.
	SrpPortal behaviorWithAllSubclasses: SrpSubstituteObject do: [:aRule |
		((aRule useWhenSaving and: [aRule isDefaultRule])
			ifTrue:  [aRule applicableClassNames isEmpty not]
			ifFalse: [self shouldApplyRule: aRule])
				ifTrue: [answer add: aRule].
	].
	^answer! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
dialectLoadExchangeRules
	"Answer a collection of class-defined class-based load
	 exchange rules that are intended to be active by default."

	^SrpPortal exchangeRules select: [:aRule | aRule useWhenLoading ].
! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
dialectSaveExchangeRules
	"Answer a collection of class-defined class-based save
	 exchange rules that are intended to be active by default."

	^SrpPortal exchangeRules select: [:aRule | aRule useWhenSaving]
! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
duplicationTypes
	"Answer an array of class symbols whose identical
	 instances should be dumped more than once. This is mostly
	 a space saving issue for instances that are automatically
	 made canonic by the object space. This can only save space
	 when the save size is generally smaller than a backpointer."

	^#(Character True False SmallInteger UndefinedObject Boolean)
	"GS/S uses Boolean instead of True/False."! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
encodingVersion
	"Answer an unsigned integer that represents the default SRP
	 encoding standard that is in effect for the data. A context will
	 know the actual encoding version."

	^0
! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
exchangeRules 
	"Answer a collection of class-based exchange rules 
	 for this configuration. This collection can contain
	 both loading and saving rules that aren't usually
	 active by default."

	^exchangeRules! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
extendedAttributesFor: anObject
	"Answer an object that is to be saved along with the receiver.
	 The object will be sent as an argument to #srpExtendedAttributes:
	 when loading."

	^anObject srpExtendedAttributes
! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
extendedAttributesFor: anObject are: eaObject
	"eaObject was saved along with anObject as a holder of additional
	 information related to anObject. This method is sent when loading to
	 reassociate the two objects."

	^anObject srpExtendedAttributes: eaObject

! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
initialEncodings
	"Answer a collection of subclasses of SrpStreamEncoding
	 that are to be used in sequence to augment data written to a data
	 stream."

	^#()! !

!SrpConfiguration methodsFor: 'initialization' stamp: ''!
initialFinishingActionsBufferSize
	"Finishing actions are performed after a traversal has completed.
	 Each object that was a part of the traversal is given an opportunity
	 to register some finishing actions. The default is to use a small buffer
	 size because this may not be used frequently."
	^2! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
initialMetastateTableName
	"Used for saving only because loaded data always knows
	 the name of a table it was saved with. Answer a string or nil."
	^nil! !

!SrpConfiguration methodsFor: 'initialization' stamp: ''!
initialObjectDepthBufferSize
	"Two stack-like collections are created for each traversal to maintain
	 full knowledge of the context in which an object of the traversal is
	 used. One collection has the original referencing objects prior to any
	 mapping operations; the other collection has placeholders
	 that will contain the final result of any mapping operations. Use a
	 low size for simple objects and a large size for deeply nested objects."
	^30! !

!SrpConfiguration methodsFor: 'initialization' stamp: ''!
initialObjectsExpectedBufferSize
	"Each traversal will touch n number of objects. Those objects are kept in a
	 collection for the sole purpose of being referenced later by their index
	 position. This is important for an OODBMS server to tell a client--'by the way,
	 the object you gave me in slot 31 should now become a forwarder to oop xxx'."
	^256! !

!SrpConfiguration methodsFor: 'initialization' stamp: ''!
initialStreamBufferSize
	"Answer the size a read/write stream should attempt to be
	 preallocated to--if permitted with the dialect being used."
	^2048! !

!SrpConfiguration methodsFor: 'initialization' stamp: ''!
initialize

	exchangeRules := Set new.
	metastates := OrderedCollection new.
! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
loadMetastateTableNamed: metastateTableName
	"Metastate tables contain data descriptions used to read
	 state from SRP encoded data, they serve the same purpose
	 as an XML DTD file. This method loads an .mst file containing
	 a collection of metastates."
	
	"	self new loadMetastateTableNamed: 'RtcDefaults1'.	"

	| loaded filename |
	filename := self metastateTableDirectory, metastateTableName, '.mst'.
	loaded := SrpConfiguration default loadObjectFromBinaryFilename: filename.
	loaded isNil ifFalse: [self metastateTables at: metastateTableName put: loaded].
	^loaded
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
loadObjectFrom: aCharacterCollection
	"A convenience method to load one object from aCharacterCollection
	 using the receiver as the configuration. The collection is expected
	 to be of the same type (binary or text) as the value returned
	 by #useBinaryAccess."
	^(self onStream: (ReadStream on: aCharacterCollection)) next! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
loadObjectFromBinary: aByteArray
	"A convenience method to load one object from
	 aByteArray using the receiver as the configuration."
	^(self onBinaryStream: (ReadStream on: aByteArray)) next! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
loadObjectFromBinaryFilename: aFilename
	"A convenience method to load one object from a file
	 using the receiver as the configuration. The file is expected
	 to be contain binary data."

	| objectFileStream answer |
	(SrpPortal stringExistsAsFile: aFilename) ifFalse: [^nil].
	objectFileStream := self newStreamOpenBinaryFilename: aFilename.
	answer := objectFileStream next.
	objectFileStream close.
	^answer
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
loadObjectFromFilename: filename
	"A convenience method to load one object from a file
	 using the receiver as the configuration. The file is expected
	 to be of the same type (binary or text) as the value returned
	 by #useBinaryAccess."

	^self useBinaryAccess
		ifTrue: [self loadObjectFromBinaryFilename: filename]
		ifFalse: [self loadObjectFromTextFilename: filename]
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
loadObjectFromText: aString
	"A convenience method to load one object from
	 aString using the receiver as the configuration."
	^(self onTextStream: (ReadStream on: aString)) next! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
loadObjectFromTextFilename: aFilename
	"A convenience method to load one object from a file
	 using the receiver as the configuration. The file is expected
	 to be contain text data."

	| objectFileStream answer |
	(SrpPortal stringExistsAsFile: aFilename) ifFalse: [^nil].
	objectFileStream := self newStreamOpenTextFilename: aFilename.
	answer := objectFileStream next.
	objectFileStream close.
	^answer
! !

!SrpConfiguration methodsFor: 'exchange rules-inheritance' stamp: ''!
loadPostMapCommand
	"Used to declare the code that should be executed to
	 apply inheritance-based mapping rules for all objects
	 as they are traversed for a load just after class-based
	 mapping rules are applied."
	^[:object :context | object srpLoadPostMap: context ].! !

!SrpConfiguration methodsFor: 'exchange rules-inheritance' stamp: ''!
loadPreMapCommand
	"Used to declare the code that should be executed to
	 apply inheritance-based mapping rules for all objects
	 as they are traversed for a load just before class-based
	 mapping rules are applied."
	^[:object :context | object srpLoadPreMap: context ].! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
loaded: anObject using: anObjectLoader
	"Notification that anObjectLoader was used to save
	 anObject. Do nothing by default. Answer the object
	 that was loaded, or an object to take place of that object."

	^anObject! !

!SrpConfiguration methodsFor: 'metastates' stamp: ''!
loadedMetastate: aMetastate context: aContext
	"Notification that aMetastate was loaded."

! !

!SrpConfiguration methodsFor: 'exchange rules-instance based' stamp: ''!
loadedProxy: proxyObject context: aContext
	"A proxy object was loaded. This allows you to perform
	 any actions on the proxy or to answer another object to be
	 used in place of the proxy."

	^proxyObject
! !

!SrpConfiguration methodsFor: 'exchange rules-instance based' stamp: ''!
loadedProxy: proxyObject context: aContext forPosition: streamPosition
	"A proxy object was loaded. This allows you to perform
	 any actions on the proxy or to answer another object to be
	 used in place of the proxy."

	^proxyObject
! !

!SrpConfiguration methodsFor: 'metastates' stamp: ''!
metastateClass
	"Answer the class whose instances describe
	 instances of the receiver."

	^SrpMetastate! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
metastateTableDirectory
	"Answer the directory from which .mst (metastate table) files
	 will be loaded. The default is the current directory."

	^String new
! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
metastateTables

	^self class metastateTables! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
metastates

	"Lazy init is only here to help loading newer version of code."
	metastates isNil ifTrue: [metastates := OrderedCollection new].
	^metastates! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
metastatesUsedToSave: anObject
	"This is a convenience method that can be used when saving new
	 metastate tables. The method saves anObject and answers all
	 the metastates that were used to save that object."

	^self newStream writer
		nextPut: anObject;
		allMetastates! !

!SrpConfiguration methodsFor: 'exchange rules-post load' stamp: ''!
missingBehaviorOf: loadedObject loadedFrom: aMarshaler
	"This will be evaluated for each State that is loaded whose
	 metastate behavior could not be found. The return value is 
	 used in place of the loadedObject. Graph references are
	 maintained dispite this exchange."

	^loadedObject! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
newHitList
	"Answer kind of IdentityDictionary that will be used to
	 track the relative position of objects that are written to
	 the stream. Mapped objects are also added. Some form
	 of optimized high capacity weakly keyed identity dictionary
	 would be ideal here for persisting very large objects. This
	 method can be overridden in the saver class in case one
	 is available."

	^SrpPortal optionalWeakKeyIdentityDictionaryClass 
		new: self newHitListSize
! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
newHitListSize
	"Answer the initial size of an IdentityDictionary that will be used to
	 track the relative position of objects that are written to
	 the stream."

	^100
! !

!SrpConfiguration methodsFor: 'context creation' stamp: ''!
newLoadContext
	^ SrpLoadContext newConfiguration: self
! !

!SrpConfiguration methodsFor: 'metastates' stamp: ''!
newMetastateForClass: aClass

	^self metastateClass 
		newLikeClass: aClass 
		configuration: self
! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
newMetastateTable
	"A metastate table acts as an efficient way to access metastates
	 by both their ID and by their name. It is a collection of metastates
	 that is indexing for fast access."

	^SrpMetastateTable new
		addAllMetastates: self metastates;
		yourself
! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
newMetastateTableReferenceNamed: aString

	^SrpMetastateTableReference newNamed: aString configuration: self! !

!SrpConfiguration methodsFor: 'exchange rules-post load' stamp: ''!
newPlaceholderFor: anObject atLoaderIndex: index
	"Placeholders are used to maintain reference integrity when loading.
	 It allows for an object containing a reference to itself is exchanged
	 for another object. Without a placeholder, the reference will still point to the
	 old 'self' rather than the new 'self'. You would normally want to 
	 use load placeholders, but there are cases where it is safe to
	 not use them and doing without them would load faster."	

	^(SrpPlaceholder baseObject: anObject) srpLoaderIndex: index; yourself

! !

!SrpConfiguration methodsFor: 'context creation' stamp: ''!
newSaveContext
	^SrpSaveContext newConfiguration: self
! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
newStream
	"Answer a new object stream using the receiver as the configuration."
	^self objectStreamClass
		newConfiguration: self! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
newStreamOpenBinaryFilename: filename
	| aStream |
	aStream := SrpPortal stringAsBinaryFileStream: filename.
 	^self onBinaryStream: aStream! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
newStreamOpenFilename: filename

	^self useBinaryAccess
		ifTrue: [self newStreamOpenBinaryFilename: filename]
		ifFalse: [self newStreamOpenTextFilename: filename]
! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
newStreamOpenTextFilename: filename
	| aStream |
	aStream := SrpPortal stringAsTextFileStream: filename.
 	^self onTextStream: aStream! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
objectStream
	"Answer a the object stream for this configuration.
	 The default is to return a new object stream."

	^self newStream
! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
objectStreamClass
	"Allows for user-defined subclasses of SrpObjectStream to be used instead."
	^SrpObjectStream! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
onBinaryStream: aStream
	"Answer a new object stream in binary mode using the
	 receiver as the configuration for access to aStream."
	^self newStream binary on: aStream! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
onStream: aStream
	"Answer a new object stream using the receiver as the
	 configuration for access to aStream. The stream is
	 expected to be of the same type (binary or text) as
	 the value returned by #useBinaryAccess."
	^self newStream on: aStream! !

!SrpConfiguration methodsFor: 'object stream creation' stamp: ''!
onTextStream: aStream
	"Answer a new object stream in text mode using the
	 receiver as the configuration for access to aStream."
	^self newStream text on: aStream
! !

!SrpConfiguration methodsFor: 'exchange rules-post load' stamp: ''!
postLoadActionFor: loadedObject loader: aMarshaler
	"All loaded objects go through here. This is an opportunity for a
	 configuration to declare actions to be taken with objects after the
	 traversal has been completely loaded, but before load placeholders
	 are stripped. Answer nil if no action is necessary; otherwise answer
	 an object that responds to #value and answers an object to take the
	 place of the loadedObject."

	^loadedObject srpPostLoadActionForLoader: aMarshaler! !

!SrpConfiguration methodsFor: 'exchange rules-post load' stamp: ''!
processPostLoadActions: finishingActions for: aLoader
	"Finishing actions are processed after the root object is
	 completely loaded or saved. Anything that responds to
	 #value can be used as a finishing action."

	finishingActions do: [:ea | ea value ].
! !

!SrpConfiguration methodsFor: 'exchange rules-post load' stamp: ''!
processPostSaveActions: finishingActions for: aSaver
	"Finishing actions are processed after the root object is
	 completely loaded or saved. Anything that responds to
	 #value can be used as a finishing action."

	finishingActions do: [:ea | ea value ].
! !

!SrpConfiguration methodsFor: 'exchange rules-instance based' stamp: ''!
proxyFor: anObject context: aContext
	"This is used to save a proxy in place of an object when 
	 #srpDataTypeMap specifies the data type #Proxy.
	 Override this method to return the proxy you want to save."

	^anObject
! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
readerClass
	"Allows for user-defined subclasses of SrpObjectLoader to be used instead."
	^SrpObjectLoader! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
removeExchangeRules: aCollection
	"Used to remove from the collection of selected class-based exchange
	 rules that are enabled for this configuration."

	aCollection do: [:ea | self exchangeRules remove: ea ifAbsent: []].
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
replicate: anObject
	"A convenience method to replicate anObject using
	 the receiver for configuration settings."
		"	self new replicate: #(one two three).	"
	^self newStream 
		nextPut: anObject; 
		reset;
		next! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
reportError: errorString
	"Most errors raised by SRP are reported through this method."

	SrpPortal reportError: errorString! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
resolveMetastateTableNamed: aString
	"Answer a registered or local metastate table for the given name."
	
	^self metastateTables 
		at: aString 
		ifAbsent: [self loadMetastateTableNamed: aString]
! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
resolveMetastateTableReference: aReference
	"Do what can be done to find a metastate table for the given reference."
	
	| aCollection |
	aCollection := self resolveMetastateTableNamed: aReference name.
	aCollection isNil ifTrue: [aCollection := aReference fetchTable].
	^aCollection
		! !

!SrpConfiguration methodsFor: 'metastate table' stamp: ''!
saveMetastateTableNamed: metastateTableName containing: aCollectionOfClassesOrMetastates
	"Metastate tables contain data descriptions used to read
	 state from SRP encoded data, they serve the same purpose
	 as an XML DTD file. This method save a collection of metastates
	 to an .mst file."
	
	"	self new saveMetastateTableNamed: 'Example' containing: #(Array String).	"
	"	self new loadMetastateTableNamed: 'Example'.	"

	| someMetastates filename |
	filename := self metastateTableDirectory, metastateTableName, '.mst'.
	someMetastates := SrpMetastateTableReference codedMetastatesForAll: aCollectionOfClassesOrMetastates.
	SrpConfiguration default saveObject: someMetastates toBinaryFilename: filename.
	^someMetastates
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
saveObject: anObject
	"A convenience method to save anObject as a character collection
	 using the receiver as the configuration. The collection will be of
	 the type (binary or text) as the value returned by #useBinaryAccess."
	"	self new saveObject: #(one two three).	"
	^self newStream
		nextPut: anObject; 
		streamContents! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
saveObject: anObject toBinaryFilename: aFilename
	"A convenience method to load one object from a file
	 using the receiver as the configuration. The file is expected
	 to be contain binary data."

	| objectFileStream |
	objectFileStream := self newStreamOpenBinaryFilename: aFilename.
	objectFileStream nextPut: anObject.
	objectFileStream close.

! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
saveObject: anObject toFilename: filename
	"A convenience method to save one object from a file
	 using the receiver as the configuration. The file is expected
	 to be of the same type (binary or text) as the value returned
	 by #useBinaryAccess."

	^self useBinaryAccess
		ifTrue: [self saveObject: anObject toBinaryFilename: filename]
		ifFalse: [self saveObject: anObject toTextFilename: filename]
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
saveObject: anObject toTextFilename: aFilename
	"A convenience method to load one object from a file
	 using the receiver as the configuration. The file is expected
	 to be contain text data."

	| objectFileStream |
	objectFileStream := self newStreamOpenTextFilename: aFilename.
	objectFileStream nextPut: anObject.
	objectFileStream close.

! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
saveObjectAsBinary: anObject
	"A convenience method to save anObject as a
	 byte array using the receiver as the configuration."
	"	self new saveObjectAsBinary: #(one two three).	"
	^self newStream binary
		nextPut: anObject; 
		streamContents! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
saveObjectAsText: anObject
	"A convenience method to save anObject as a
	 string using the receiver as the configuration."
	"	self new saveObjectAsText: #(one two three).	"
	^self newStream text
		nextPut: anObject; 
		streamContents! !

!SrpConfiguration methodsFor: 'exchange rules-inheritance' stamp: ''!
savePostMapCommand
	"Used to declare the code that should be executed to
	 apply inheritance-based mapping rules for all objects
	 as they are traversed for a save just after class-based
	 mapping rules are applied."
	^[:object :context | object srpSavePostMap: context ].! !

!SrpConfiguration methodsFor: 'exchange rules-inheritance' stamp: ''!
savePreMapCommand
	"Used to declare the code that should be executed to
	 apply inheritance-based mapping rules for all objects
	 as they are traversed for a load just before class-based
	 mapping rules are applied."
	^[:object :context | object srpSavePreMap: context ].! !

!SrpConfiguration methodsFor: 'notification' stamp: ''!
saved: anObject using: anObjectSaver
	"Notification that anObjectSaver was used to save
	 anObject. Do nothing by default."

! !

!SrpConfiguration methodsFor: 'metastates' stamp: ''!
savingMetastate: aMetastate context: aContext
	"Notification that aMetastate is about to be saved. This is
	 a good place to hook if you want to associate other objects
	 (via extended attributes) with the metastate being saved."

! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
selectedLoadExchangeRules
	^self exchangeRules select: [:ea | ea useWhenLoading ]! !

!SrpConfiguration methodsFor: 'exchange rules-selected' stamp: ''!
selectedSaveExchangeRules
	^self exchangeRules select: [:ea | ea useWhenSaving ]! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
shouldApplyRule: aMappingRule
	"Subclasses of SrpMappingRule that respond false to #isDefaultRule
	 are given this opportunity to be considered part of a configuration.
	 This method can be overridden to selectively enable class-defined
	 class-based mapping rules that aren't enabled by default."

	^false
! !

!SrpConfiguration methodsFor: 'convenience methods' stamp: ''!
smalltalkSourceForObject: anObject
	"Answer a string with Smalltalk source code that can be
	 executed to recreate anObject."

	"	self new smalltalkSourceForObject: #(one two three).	"

	| string |
	string := self saveObjectAsText: anObject.
	^(WriteStream on: String new)
		nextPut: $(;
		nextPutAll: self class name;
		space; nextPutAll: #'new';
		space; nextPutAll: #'loadObjectFromText:';
		space; nextPut: $';
		cr; nextPutAll: (SrpPortal stringToSourceText: string);
		nextPutAll: ''')';
		contents! !

!SrpConfiguration methodsFor: 'exchange rules-class based' stamp: ''!
srpDataTypeMapForClass: aClass
	"You can customize this method if you want your
	 mapping rules to be configuration based rather than
	 just behavior based."

	^SrpPortal fixupLiteralArray: aClass srpDataTypeMap	! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
textAccessEncodings
	"Answer a collection of subclasses of SrpStreamEncoding that
	 are to ensure a data stream containing text can be read."

	^Array
		with: SrpNativeStreamWrapper
		with: SrpBase64
! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
useBinaryAccess
	"Answer boolean whether the data stream contains binary or text data.
	 This is really a default value because a stream can be treated on
	 demand as either binary or text--regardless of this setting."

	^true
! !

!SrpConfiguration methodsFor: 'exchange rules-post load' stamp: ''!
usePlaceholders
	"Answer boolean whether placeholders should be used to
	 maintain reference integrity of objects exchanged with others
	 while loading. You would normally want to do this, but there are
	 cases where it is safe to not use them and doing without them
	 would load faster."	

	^true

	"One scenario to watch out for is the case where an object
	 containing a reference to itself is exchanged for another object. 
	 If you don't use placeholders then the reference will still
	 point to the old 'self' rather than the new 'self'."

! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
versionNumberArrayFor: aBehavior
	"Answer nil or an array of unsigned integers that is to be saved
	 with the metastate information for aBehavior. The information
	 is user defined and can be used as an optional migration feature."

	^aBehavior srpVersionNumberArray.

! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
wrapAppendedText
	"Answer boolean whether text written to the end of a stream
	 should be wrapped."

	^true
! !

!SrpConfiguration methodsFor: 'data encoding' stamp: ''!
writerClass
	"Allows for user-defined subclasses of SrpObjectSaver to be used instead."
	^SrpObjectSaver! !

SrpConfiguration subclass: #SrpNonMappingConfiguration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpNonMappingConfiguration classSide methodsFor: 'documentation' stamp: ''!
documentation
"
	This configuration is used to save objects without using any of the
	default mapping rules that are visible to SrpConfiguration.
	This is used to save native (non-portable) representations of objects.

	A Dictionary is implemented differently by just about every
	Smalltalk dialect. If you were to use this configuration to save
	a Dictionary instance then that may only be loadable by the
	same dialect (and perhaps platform) that the instance was saved.
"! !

!SrpNonMappingConfiguration methodsFor: 'mapping rules-class based' stamp: ''!
allLoadExchangeRules
	"Answer a collection of all the class-based exchange rules that are
	 to be in effect when loading using this configuration."

	^Set new
		addAll: self dialectLoadExchangeRules;
		addAll: self selectedLoadExchangeRules;
		yourself.! !

!SrpNonMappingConfiguration methodsFor: 'mapping rules-class based' stamp: ''!
allSaveExchangeRules
	"Answer a collection of all the class-based exchange rules that are
	 to be in effect when saving using this configuration."

	^Set new
		addAll: self dialectSaveExchangeRules;
		addAll: self selectedSaveExchangeRules;
		yourself.! !

SrpConfiguration subclass: #SrtNboConfiguration
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtNboConfiguration methodsFor: 'Not categorized' stamp: ''!
initialEncodings
	"Answer an ordered collection of subclasses of SrpStreamEncoding
	 that are to be used in sequence to augment data written to a data
	 stream."

	^Array with: SrtNaturalBitOrder! !

SrpConfiguration subclass: #SrtProxyTestConfiguration
	instanceVariableNames: 'toProxyBlock fromProxyBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtProxyTestConfiguration methodsFor: 'mapping rules-instance based' stamp: ''!
fromProxyBlock: oneArgBlock

	fromProxyBlock := oneArgBlock! !

!SrtProxyTestConfiguration methodsFor: 'mapping rules-instance based' stamp: ''!
loadedProxy: proxyObject context: aContext
	"A proxy object was loaded. This allows you to perform
	 any actions on the proxy or to answer another object to be
	 used in place of the proxy."

	^fromProxyBlock value: proxyObject
! !

!SrtProxyTestConfiguration methodsFor: 'mapping rules-instance based' stamp: ''!
loadedProxy: proxyObject context: aContext forPosition: streamPosition
	"A proxy object was loaded. This allows you to perform
	 any actions on the proxy or to answer another object to be
	 used in place of the proxy."

	^fromProxyBlock value: proxyObject
! !

!SrtProxyTestConfiguration methodsFor: 'mapping rules-instance based' stamp: ''!
proxyFor: anObject context: aContext
	"This is used to save a proxy in place of an object when 
	 #srpDataTypeMap specifies the data type #Proxy.
	 Override this method to return the proxy you want to save."

	^toProxyBlock value: anObject
! !

!SrtProxyTestConfiguration methodsFor: 'mapping rules-instance based' stamp: ''!
toProxyBlock: oneArgBlock

	toProxyBlock := oneArgBlock! !

SrpConfiguration subclass: #SrtSequenceCheckConfiguration
	instanceVariableNames: 'userData'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
assert: aBoolean
	aBoolean ifFalse: [self error: 'Assertion failed']! !

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
contextUserData
	"The context object for each traversal can be the holder of
	 user data collected or referenced during a traversal. User
	 data is expected to behave as a dictionary. Nil can be
	 returned by this method to avoid dictionary creation for every
	 traversal unless one is necessary."

	userData isNil ifTrue: [userData := Dictionary new].
	^userData
! !

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
lastLoader

	^userData at: #lastLoader! !

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
lastSaver

	^userData at: #lastSaver! !

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
loaded: anObject using: anObjectLoader
	"Notification that anObjectLoader was used to save
	 anObject. Do nothing by default. Answer the object
	 that was loaded, or an object to take place of that object."

	userData at: #lastLoader put: anObjectLoader.
	^anObject
! !

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
replicate: obj

	| sSeq lSeq answer |
	answer := super replicate: obj.
	sSeq := self lastSaver hitSequence.
	lSeq := self lastLoader hitSequence.
	self assert: sSeq size = lSeq size.
	1 to: sSeq size do: [:i |
		self assert: (sSeq at: i) class = (lSeq at: i) class.
		"Couldn't send #size to VW LargePositiveIntegers"
		"self assert: (sSeq at: i) size = (lSeq at: i) size."
	].
	^answer! !

!SrtSequenceCheckConfiguration methodsFor: 'Not categorized' stamp: ''!
saved: anObject using: anObjectSaver
	"Notification that anObjectSaver was used to save
	 anObject. Do nothing by default."

	userData at: #lastSaver put: anObjectSaver.
! !

Object subclass: #SrpContext
	instanceVariableNames: 'originalCurrent mappedCurrent originalStack mappedStack exchangeRules configuration userData encodingVersion'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpContext classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 An instance of this class is created for each SRP traversal 
 (a save or load operation). Instances maintain a stack of
 objects that indicate all objects (from the root object) that
 were traversed to get to the current traversal object. An
 instance is passed for use in mapping operations.

 Instances know the full stack of both the original object
 (prior to mapping rules), as well as the stack of objects
 that resulted from mapping operations.  

 An instance knows all the mapping rules that are in effect
 for a traversal, and allows rules to be changed for objects
 yet to be processed. An instance can hold any user data 
 for the traversal. 
"! !

!SrpContext classSide methodsFor: 'instance creation' stamp: ''!
newConfiguration: aConfig
	^self new initializeConfiguration: aConfig! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
addExchangeRules: aCollection
	"Adds aCollection of exchange rules that will be used to
	 perform special operations on instances as they are
	 encountered."

	| rules |
	rules := self exchangeRules.
	aCollection do: [:each |
		each applicableClassNames do: [:native |
			(rules includesKey: native)
				ifTrue: [rules at: native put: (rules at: native), each]
				ifFalse: [rules at: native put: each].
		].
	].
! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
applicableExchangeRules
	"Answer a collection off all the exchange rules that apply
	 to the type of operations performed by the receiver."

	^self subclassResponsibility! !

!SrpContext methodsFor: 'accessors' stamp: ''!
basicUserData
	"Answer an object (usually a dictionary) that anyone can
	 use to hold data relevant for a traversal. This accessor
	 doesn't lazy-initialize."

	^userData
! !

!SrpContext methodsFor: 'accessors' stamp: ''!
configuration
	"Answer a kind of SrpConfiguration that knows settings
	 to be used for load and/or save traversals."

	^configuration! !

!SrpContext methodsFor: 'private' stamp: ''!
debugPrintOn: aStream
	"Private."

	self printOn: aStream.
	aStream cr; tab; nextPutAll: 'stack depth: '.
	self stackDepth printOn: aStream.
	aStream cr; tab; nextPutAll: 'exchange rules: '.
	self exchangeRules size printOn: aStream.
	aStream cr; tab; nextPutAll: 'was remapped: '.
	self wasCurrentRemapped printOn: aStream.
	aStream cr; tab; nextPutAll: 'any remapped: '.
	self wereAnyRemapped printOn: aStream.! !

!SrpContext methodsFor: 'accessors' stamp: ''!
encodingVersion
	"Answer an unsigned integer that represents the SRP encoding
	 standard that is in effect for the data."

	^encodingVersion isNil 
		ifTrue: [configuration encodingVersion] 
		ifFalse: [encodingVersion].! !

!SrpContext methodsFor: 'private' stamp: ''!
exchange
	"Private. Remap mappedCurrent using whatever mapping rules
	 are in effect."

	mappedCurrent := self preMapCommand value: mappedCurrent value: self.
	"mappedCurrent :=" self classBasedExchange.
	mappedCurrent := self postMapCommand value: mappedCurrent value: self.
	^mappedCurrent
! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
exchangeRules
	"Answer an IdentityDictionary containing of exchange rules
	 that will be used to perform special operations on instances
	 as they are encountered."

	^exchangeRules
! !

!SrpContext methodsFor: 'initialization' stamp: ''!
initializeConfiguration: aConfig
	"Private."
	configuration := aConfig.
	originalStack := OrderedCollection new: aConfig initialObjectDepthBufferSize.
	mappedStack := OrderedCollection new: aConfig initialObjectDepthBufferSize.
	userData := aConfig contextUserData.
	self initializeExchangeRules.! !

!SrpContext methodsFor: 'initialization' stamp: ''!
initializeExchangeRules
	"Private."
	| rules |
	rules := self applicableExchangeRules.
	exchangeRules := IdentityDictionary new: rules size.
	self addExchangeRules: rules.
! !

!SrpContext methodsFor: 'testing' stamp: ''!
isRoot
	"Answer boolean whether the object being mapped
	 is the root object."

	^originalStack isEmpty! !

!SrpContext methodsFor: 'accessors' stamp: ''!
mappedCurrent
	"Answer an object that may be a mapped version of the
	 originalCurrent object."
	
	^mappedCurrent! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
mappedCurrent: anObject
	"Use this method with care when loading. You don't want to
	 prematurely kill a placeholder--preventing a thorough
	 cleanup of other placeholders."

	mappedCurrent := anObject.! !

!SrpContext methodsFor: 'accessors' stamp: ''!
mappedReference
	"Answer the mapped object that referenced the
	 current mapped object being procesed."

	^mappedStack size > 0
		ifTrue: [ mappedStack last ]
		ifFalse: [ nil ].! !

!SrpContext methodsFor: 'accessors' stamp: ''!
mappedRoot
	"Answer the mapped version of the nativeRoot if
	 the root object went through any mapping."

	mappedStack isEmpty ifTrue: [^mappedCurrent].
	^mappedStack first! !

!SrpContext methodsFor: 'accessors' stamp: ''!
mappedStack
	"Answer an ordered collection of objects that were all traversed
	 down a network of objects to get to the current object. Mapped
	 objects are the result of any mappings."

	^mappedStack! !

!SrpContext methodsFor: 'accessors' stamp: ''!
originalCurrent
	"Answer the original object that existed prior to any
	 mapping."

	^originalCurrent! !

!SrpContext methodsFor: 'accessors' stamp: ''!
originalReference
	"Answer the original object that *may* have referenced the
	 current original object being procesed. The originalCurrent
	 could have multiple original references but only the first
	 encountered by the traversal will get this message."

	^originalStack size > 0
		ifTrue: [ originalStack last ]
		ifFalse: [ nil ].! !

!SrpContext methodsFor: 'accessors' stamp: ''!
originalRoot
	"Answer the original object that existed prior to
	 any mapping."

	originalStack isEmpty ifTrue: [^originalCurrent].
	^originalStack first! !

!SrpContext methodsFor: 'accessors' stamp: ''!
originalStack
	"Answer an ordered collection of objects that were all traversed
	 down a network of objects to get to the current object. Original
	 objects are those that existed before direct mapping was done
	 to them. Original objects could have been indirectly mapped as
	 another object (earlier in the originalStack) was mapped."

	^originalStack! !

!SrpContext methodsFor: 'private' stamp: ''!
popStack
	"Private. This should not be sent by anything other than
	 the dumper/loader."
	
	originalStack isEmpty ifFalse: [
		"[ObjectStudio] OrderedCollection>>removeLast doesn't return the item removed."
		originalCurrent := originalStack at: originalStack size.
		originalStack removeLast.
		mappedCurrent := mappedStack at: mappedStack size.
		mappedStack removeLast.
	].
! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
postMapCommand
	"The pre/post map commands are sent #value:value: for each object
	 that is traversed."

	^self subclassResponsibility! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
preMapCommand
	"The pre/post map commands are sent #value:value: for each object
	 that is traversed."

	^self subclassResponsibility! !

!SrpContext methodsFor: 'private' stamp: ''!
pushStack
	"Private. This should not be sent by anything other than
	 the dumper/loader."

	originalCurrent isNil ifFalse: [
		originalStack addLast: originalCurrent.
		mappedStack addLast: mappedCurrent.
	].
	originalCurrent := mappedCurrent := nil.
! !

!SrpContext methodsFor: 'testing' stamp: ''!
recursiveMappingFor: anOriginal
	"Answers nil or the mapped replacement for anOriginal if
	 anOriginal is in the context stack--if it is a recursive reference."

	| oati mati |
	1 to: originalStack size do: [:i |
		(anOriginal == (oati := originalStack at: i)
			and: [oati ~~ (mati := mappedStack at: i)]) ifTrue: [^mati].
	].
	^nil

! !

!SrpContext methodsFor: 'mapping rules' stamp: ''!
removeExchangeRules: aCollection
	"Removes aCollection of exchange rules that will be used to
	 perform special operations on instances as they are
	 encountered."

	| rules remaining |
	rules := self exchangeRules.
	aCollection do: [:rule |
		rule applicableClassNames do: [:native |
			(rules includesKey: native) ifTrue: [
				remaining := (rules at: native) without: rule.
				remaining isNil
					ifTrue:  [ rules removeKey: native ]
					ifFalse: [ rules at: native put: remaining ].
			].
		].
	].
! !

!SrpContext methodsFor: 'accessors' stamp: ''!
stackDepth
	"Answer the number of objects traversed to get to
	 the current object--not including the current object."

	^self originalStack size! !

!SrpContext methodsFor: 'accessors' stamp: ''!
userData
	"Answer a dictionary that anyone can use to hold data
	 relevant for a traversal. The receiver is available for
 	 all mapping operations."

	userData isNil ifTrue: [userData := Dictionary new: 4].
	^userData
! !

!SrpContext methodsFor: 'testing' stamp: ''!
wasCurrentRemapped

	^originalCurrent ~~ mappedCurrent! !

!SrpContext methodsFor: 'testing' stamp: ''!
wereAnyRemapped

	1 to: originalStack size do: [:i |
		(originalStack at: i) ~~ (mappedStack at: i) ifTrue: [^true].
	].
	^false

! !

SrpContext subclass: #SrpLoadContext
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpLoadContext methodsFor: 'accessors' stamp: ''!
applicableExchangeRules
	"Answer a collection off all the exchange rules that apply
	 to the type of operations performed by the receiver."

	^configuration allLoadExchangeRules! !

!SrpLoadContext methodsFor: 'private' stamp: ''!
classBasedExchange
	"Private. Remap the mappedCurrent using whatever class-based
	 exchange rules are in effect--and perform any mapping on the
	 results of a previous exchange. Answer the object that will take
	 the place of the nativeCurrent object."

	| attempts mappingRule original |
	attempts := 0.
	[ 
		original := mappedCurrent srpBaseObject.
		mappingRule := exchangeRules 
			at: (SrpPortal shortNameForBehavior: original srpClass)
			ifAbsent: [^mappedCurrent].
		mappedCurrent srpBaseObject: (
			mappingRule 
				applyTo: original 
				context: self
		).
		(mappedCurrent srpClass ~~ original srpClass)
	] whileTrue: [
		attempts := attempts + 1.
		attempts > 31 ifTrue: [self configuration reportError: 'Recursive mapping rules.'].
	].
	^mappedCurrent! !

!SrpLoadContext methodsFor: 'initialization' stamp: ''!
initializeConfiguration: aConfig

	super initializeConfiguration: aConfig.
	aConfig beingUsedForLoadContext: self.! !

!SrpLoadContext methodsFor: 'accessors' stamp: ''!
loaderIndex
	"Each unique object is numbered according to the
	 traversal sequence. Answer the sequence number for
	 the current object. The number could be used to refer
	 to objects in a saving (OODBMS client) object space for
	 the purpose of making them #become: proxies."

	^mappedCurrent srpLoaderIndex! !

!SrpLoadContext methodsFor: 'mapping rules' stamp: ''!
postMapCommand
	"The pre/post map commands are sent #value:value: for each object
	 that is traversed."

	^configuration loadPostMapCommand.
! !

!SrpLoadContext methodsFor: 'mapping rules' stamp: ''!
preMapCommand
	"The pre/post map commands are sent #value:value: for each object
	 that is traversed."

	^configuration loadPreMapCommand.
! !

!SrpLoadContext methodsFor: 'private' stamp: ''!
startedToLoad: aPlaceholder
	"Private. Remap anInstance using whatever mapping rules are in effect."

	originalCurrent := aPlaceholder srpBaseObject.
	mappedCurrent := aPlaceholder.
! !

!SrpLoadContext methodsFor: 'testing' stamp: ''!
wasCurrentRemapped
	"The originalCurrent will never have a load placeholder. The
	 mappedCurrent will almost always have a load placeholder--it
	 doesn't if a value hasn't yet been assigned."

	^originalCurrent ~~ mappedCurrent srpWithoutPlaceholder! !

!SrpLoadContext methodsFor: 'testing' stamp: ''!
wereAnyRemapped
	"The originals will never have a load placeholder. The
	 mapped will almost always have a placeholder."

	1 to: originalStack size do: [:i |
		(originalStack at: i) ~~ 
		(mappedStack at: i) srpWithoutPlaceholder 
			ifTrue: [^true].
	].
	^false

! !

SrpContext subclass: #SrpSaveContext
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpSaveContext methodsFor: 'accessors' stamp: ''!
applicableExchangeRules
	"Answer a collection off all the exchange rules that apply
	 to the type of operations performed by the receiver."

	^configuration allSaveExchangeRules! !

!SrpSaveContext methodsFor: 'mapping rules' stamp: 'mu 6/10/2011 14:09'!
classBasedExchange
	"Private. Remap the mappedCurrent using whatever class-based
	 exchange rules are in effect--and perform any exchange on the
	 results of a previous exchange. Answer the object that will take
	 the place of the nativeCurrent object."

	| attempts mappingRule original |
	attempts := 0.
	[ 
		original := mappedCurrent.
		mappingRule := exchangeRules 
			at: (SrpPortableBehavior portal shortNameForBehavior: original srpClass)
			ifAbsent: [^mappedCurrent].
		mappedCurrent := mappingRule 
			applyTo: original 
			context: self.
		(mappedCurrent srpClass ~~ original srpClass)
	] whileTrue: [
		attempts := attempts + 1.
		attempts > 31 ifTrue: [self configuration reportError: 'Recursive mapping rules.'].
	].
	^mappedCurrent! !

!SrpSaveContext methodsFor: 'mapping rules' stamp: ''!
exchangeOriginal: anInstance
	"Private. Remap anInstance using whatever mapping rules
	 are in effect. Answer anInstance or an object that will take
	 the place of anInstance."

	originalCurrent := mappedCurrent := anInstance.
	self exchange.
	^mappedCurrent
! !

!SrpSaveContext methodsFor: 'initialization' stamp: ''!
initializeConfiguration: aConfig
	"Private."
	super initializeConfiguration: aConfig.
	aConfig beingUsedForSaveContext: self.! !

!SrpSaveContext methodsFor: 'mapping rules' stamp: ''!
postMapCommand
	"The pre/post map commands are sent #value:value: for each object
	 that is traversed."

	^configuration savePostMapCommand.
! !

!SrpSaveContext methodsFor: 'mapping rules' stamp: ''!
preMapCommand
	"The pre/post map commands are sent #value:value: for each object
	 that is traversed."

	^configuration savePreMapCommand.
! !

Object subclass: #SrpMarshaler
	instanceVariableNames: 'stream classTable context hitSequence finishingActions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpMarshaler classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 An instance of this class is created each time there is a load or save of
 a root object to a data stream. It is the main holder of information for a
 traversal. It also specifies how to encode basic data types onto the
 data stream. It is responsible for applying mapping rules and knowing
 how to handle special directions that are intended only for the
 marshaler.
"! !

!SrpMarshaler classSide methodsFor: 'accessing' stamp: ''!
loaderClass
	"Answer the class that can be used for reloading
	 a saved object."

	^SrpObjectLoader! !

!SrpMarshaler classSide methodsFor: 'instance creation' stamp: ''!
new
	"This method probably won't be used, but could be."
	^self newConfiguration: SrpConfiguration default! !

!SrpMarshaler classSide methodsFor: 'instance creation' stamp: ''!
newConfiguration: aConfig

	^super new initializeConfiguration: aConfig! !

!SrpMarshaler methodsFor: 'Finishing Actions' stamp: ''!
addFinishingAction: somethingThatRespondsToValue
	"Private. Finishing actions are processed after the root
	 object is completely loaded or saved. Anything that responds to
	 #value can be used as a finishing action."

	finishingActions add: somethingThatRespondsToValue
! !

!SrpMarshaler methodsFor: 'Accessors' stamp: ''!
allMetastates
	"Answer a collection of all metastates currently known by the marshaler.
	 This is a convenience method used for saving metastate tables."

	^classTable allMetastates
! !

!SrpMarshaler methodsFor: 'private' stamp: ''!
classTable

	^classTable! !

!SrpMarshaler methodsFor: 'Accessors' stamp: ''!
configuration
	"Answer a kind of SrpConfiguration that knows settings
	 to be used for load and/or save traversals."

	^context configuration
! !

!SrpMarshaler methodsFor: 'Accessors' stamp: ''!
context
	"The context is passed as an argument to mapping rules
	 to provide the rules with useful information about the state of
	 the operation without giving direct access to a saver
	 or loader. The context also determines the mapping
	 rules that are in effect."

	^context
! !

!SrpMarshaler methodsFor: 'Accessors' stamp: ''!
hitSequence
	"Answer an OrderedCollection containing objects in the
	 sequence they were traversed. This can be useful for
	 object databases. For example, objects sent to a server
	 can be assigned an oop, and the index position can
	 be used to identify client objects that should become
	 a proxy to the new server object."

	^hitSequence! !

!SrpMarshaler methodsFor: 'private' stamp: ''!
initializeConfiguration: aConfig
	"Private."

	classTable := aConfig newMetastateTable.
	hitSequence := OrderedCollection new: aConfig initialObjectsExpectedBufferSize.
	finishingActions := OrderedCollection new: aConfig initialFinishingActionsBufferSize.! !

!SrpMarshaler methodsFor: 'Testing' stamp: ''!
isLoader
	"Answer boolean whether this instance
	 will be used to load objects."

	^false
! !

!SrpMarshaler methodsFor: 'Testing' stamp: ''!
isSaver
	"Answer boolean whether this instance
	 will be used to save objects."

	^false
! !

!SrpMarshaler methodsFor: 'Initialization' stamp: ''!
setEncodingType: encodingType

	stream := stream srpSetEncodingType: encodingType.
! !

!SrpMarshaler methodsFor: 'private' stamp: ''!
srpSavePreMap: aContext
	"Private. In case there is a reference back to the saver, we don't want
	 to save the instance of the tool we are using to save objects. If this
	 is sent, then you are probably saving way more than you intended."

	^nil
! !

!SrpMarshaler methodsFor: 'Accessors' stamp: ''!
stream
	"Answer the byte stream that contains data
	 converted to or from objects. The stream will
	 either be for read or write--but never both."

	^stream! !

!SrpMarshaler methodsFor: 'private' stamp: ''!
stream: aStream encodings: aCollection
	"By default, aStream is expected to accept binary data (on a
	 ByteArray or a file stream set to binary mode). The type of
	 stream that will work (text, binary, file, etc.) can be influenced
	 by any encodings that are used."

	stream := aCollection
		inject: (SrpNativeStreamWrapper on: aStream for: self)
		into: [:last :ea | ea on: last ].
! !

SrpMarshaler subclass: #SrpObjectLoader
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpObjectLoader methodsFor: 'private' stamp: ''!
attemptDiscardOfPlaceholder: mutatedObjectPlaceholder
	"Private. The placeholder can be removed at this stage under certain
	 circumstances. Answer mutatedObject without the load placeholder
	 if it is safe to do so."

	mutatedObjectPlaceholder srpIsPlaceholder ifFalse: [^mutatedObjectPlaceholder].
	mutatedObjectPlaceholder srpCanDiscard ifFalse: [^mutatedObjectPlaceholder].
	hitSequence
		at: mutatedObjectPlaceholder srpLoaderIndex
		put: mutatedObjectPlaceholder srpBaseObject.
	"ObjectStore #at:put: doesn't return the put object."
	^mutatedObjectPlaceholder srpBaseObject
! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
basicLoadMetastate
	"Metastates describe data so it can be loaded. It is
	 one of the few data structures directly written by SRP
	 instead of stored as an object that can be mapped."

	| classCodeOrZero classDef shapeCode namePath |
	(classCodeOrZero := self loadUnsigned) = 0
		ifTrue: [ ^self configuration reportError: 'Unexpected class code.' ].
	shapeCode := self loadUnsigned.
	namePath := self loadCharacterArrayArray.
	1 to: namePath size do: [:i | namePath at: i put: (namePath at: i) asSymbol ].
	classDef := self configuration metastateClass
		code: classCodeOrZero
		shapeCode: shapeCode
		behaviorName: namePath.
	classDef hasVersionNumberArray ifTrue: [
		classDef versionNumberArray: self loadUnsignedArray.
	].
	classDef hasNamedSlots ifTrue: [
		classDef allInstVarNames: (self loadCharacterArrayArray collect: [:ea | SrpPortal stringAsInstVarName: ea]).
		classDef hasTypedNamedSlots ifTrue: [
			classDef namedStoreTypeCodes: self loadUnsignedArray.
		].
	].
	classTable
		atCode: classDef code
		putMetastate: classDef.
	classDef hasExtendedAttributes ifTrue: [
		classDef extendedAttributes: self loadObject.
	].
	^classDef! !

!SrpObjectLoader methodsFor: 'Accessors' stamp: ''!
hitSequence
	"Answer an OrderedCollection containing objects in the
	 sequence they were traversed. This can be useful for
	 object databases. For example, objects sent to a server
	 can be assigned an oop, and the index position can
	 be used to identify client objects that should become
	 a proxy to the new server object."

	^hitSequence collect: [:ea | ea srpWithoutPlaceholder ]! !

!SrpObjectLoader methodsFor: 'Initialization' stamp: ''!
initializeConfiguration: aConfig
	"Private."

	super initializeConfiguration: aConfig.
	context := aConfig newLoadContext.
	aConfig beingUsedForObjectLoader: self.! !

!SrpObjectLoader methodsFor: 'Testing' stamp: ''!
isLoader
	"Answer boolean whether this instance
	 will be used to load objects."

	^true
! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
loadActionsFor: aPlaceholder
	"Private. Do anything that should be done to
	 an object just after it has been loaded, but
	 prior to the root object being completely loaded."

	| answer postLoadAction |
	context mappedCurrent == aPlaceholder
		ifFalse: [self configuration reportError: 'Assertion check failed when loading an object.'].
	answer := context exchange.
	postLoadAction := self configuration postLoadActionFor: answer loader: self.
	postLoadAction isNil 
		ifTrue: [ ^self attemptDiscardOfPlaceholder: answer ].
	finishingActions add: postLoadAction.
	^answer! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadAvoid
	"An object was mapped to nil by #srpDataTypeMap when saving. The
	 stream contains no state for the object, so we'll just answer nil."

	^nil! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadBoolean

	^self loadUnsigned ~= 0
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadByte

	^stream next! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadByteArray
	"Answer a ByteArray or String containing bytes (octets)."

	^stream next: self loadUnsigned.
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadCharacter
	"SRP supports character values of all sizes--yet wastes little
	 space. The font used to display a character determines if
	 non-English glyphs will show properly."

	^SrpPortal unsignedToCharacter: self loadUnsigned
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadCharacterArray
	"Load a collection of characters to be loaded in
	 the native single or double byte string class.
	 There is no encoding difference between a string
	 and an array of characters, the primary difference
	 is that the sign bit will be off for a character array."

	^self loadString
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadCharacterArrayArray
	"Answer an array of strings. Similar to #loadStringArray, but 
	 stored in an encoding that is more space efficient by at least
	 one byte when array typically contains fewer than 64 items."

	| namesCount stringSize answer string |
	stringSize := self loadSigned.
	stringSize < 0 
		ifTrue: [ namesCount := stringSize negated. stringSize := self loadUnsigned ]
		ifFalse: [ namesCount := 1 ].
	answer := Array new: namesCount.
	1 to: namesCount do: [:nameIndex |
		string := String new: stringSize.
		1 to: stringSize do: [:stringIndex | 
			string at: stringIndex put: (SrpPortal unsignedToCharacter: self loadUnsigned).
		].
		answer at: nameIndex put: string.
		nameIndex < namesCount ifTrue: [ stringSize := self loadUnsigned ].
	].
	^answer! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadDistinct
	"An object saved distinctly has a known length and has little
	 or no knowledge of state outside of it's data sequence. It can
	 be positionally loaded or skipped."

	| skipLength proxy startPosition bytes |
	(skipLength := self loadUnsigned) = 0
		ifTrue: [ ^self configuration reportError: 'Unexpected skip length.' ].
	startPosition := stream position.
	bytes := stream next: skipLength.
	proxy := self loadState.
	bytes yourself.
	^self configuration 
		loadedProxy: proxy 
		context: self context 
		forPosition: startPosition.! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadEncodingVersion
	"The data is declaring that it is encoding according to
	 a specific SRP encoding standard."

	context encodingVersion: self loadUnsigned.
	^self signalIgnoreLoad! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadMarshalDirective
	"A marshal directive is a special command intended
	 for the marshaler itself. It is indicated by a class code
	 of zero followed by the directive number."

	| directiveNumber |
	directiveNumber := self loadUnsigned.
	directiveNumber = 0 ifTrue: [ ^self specialOperationSignal ].
	^self perform: (self marshalDirectiveLoadArray at: directiveNumber).
! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadMetastate
	"Metastates describe data so it can be loaded. It is
	 one of the few data structures directly written by SRP
	 instead of stored as an object that can be mapped."

	| aMetastate |
	aMetastate := self basicLoadMetastate.
	context postMapCommand value: aMetastate value: context.
	^self signalIgnoreLoad! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadMetastateTableName
	"Private. This is a marshal directive to specify that metastates
	 be imported from some known or external source."

	| mstName metastates |
	mstName := self loadString.
	metastates := self configuration resolveMetastateTableNamed: mstName.
	metastates isNil ifTrue: [self configuration reportError: 'Data refers to a metastate table that could not be located (', mstName, ').'].
	classTable addAllMetastates: metastates.
	^self signalIgnoreLoad! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadMetastateTableReference
	"Private. This is a marshal directive to specify that metastates
	 be imported from some known or external source."

	| tableReference metastates |
	tableReference := self loadObject.
	metastates := tableReference allMetastates.
	classTable addAllMetastates: metastates.
	^self signalIgnoreLoad! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadObject
	"Load the next object on the stream and perform
	 any post load processing."

	| placeholder |
	placeholder := self loadState.
	self processFinishingActions.
	^placeholder srpIsPlaceholder 
		ifTrue:  [placeholder srpStripPlaceholders]
		ifFalse: [placeholder]
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadProxy
	"The next object in the stream is mapped to nil by #srpDataTypeMap.
	 Whatever is loaded will be skipped as if it doesn't exist in the stream.
	 A nil will be loaded and skipped if the nil mapping existed when saving."

	| proxy |
	proxy := self loadState.
	^self configuration loadedProxy: proxy context: self context.! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadRelativeOffset
	"Private. Load a reference to an object that already started
	 loading. The offset index position of the object is relative
	 to the index position of the object currently being loaded
	 rather than relative to the index position of the root object."

	| offsetOrZero |
	(offsetOrZero := self loadUnsigned) = 0 
		ifTrue: [ ^self configuration reportError: 'Unexpected relative offset load.' ].
	^hitSequence at: (hitSequence size - (offsetOrZero - 1))! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadSigned
	"Save an integer from zero to negative or positive infinity.
	 The least significant bit of the number is ON if the value
	 is a negative number."

	| unsigned |
	unsigned := self loadUnsigned.
	^unsigned \\ 2 = 0
		ifTrue: [ unsigned // 2 ]
		ifFalse: [ (unsigned // 2) negated ]

	"Another way to do the same thing:
		| unsigned |
		unsigned := self loadUnsigned.
		^(unsigned bitAnd: 1) = 0
			ifTrue: [ unsigned bitShift: -1 ]
			ifFalse: [ (unsigned bitShift: -1) negated ]
	"
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadSignedWidth
	"Load a signed integer that was saved with
	 a fixed width. We don't need to know the width
	 to read it."

	stream srpFinishByteGroupRead.
	^self loadSigned
! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadSkipBytes
	"Private. This is a marshal directive to skip a physical number
	 of bytes in the stream when loading."

	| skipLength |
	(skipLength := self loadUnsigned) = 0
		ifTrue: [ ^self configuration reportError: 'Unexpected skip length.' ].
	stream skip: skipLength.
	^self signalIgnoreLoad! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadSkipObject
	"Private. This is a marshal directive to skip the
	 next object in the stream when loading. The
	 object is loaded, remapped, and sent #srpLoaded
	 like a normal object. The object loaded is not
	 expected to be part of the answer though."

	self loadObject.
	^self signalIgnoreLoad
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadState
	"Private. Sent when loading objects referenced from a root
	 object. Leaves any required load placeholder in place for
	 post load processing."

	| classCodeOrZero aPlaceholder loadClassResponse |
	[(classCodeOrZero := self loadUnsigned) = 0] whileTrue: [
		(loadClassResponse := self loadMarshalDirective) == self signalIgnoreLoad 
			ifFalse: [^loadClassResponse].
	].
	aPlaceholder := (classTable metastateForCode: classCodeOrZero)
		loadInstanceUsing: self.
	^self loadActionsFor: aPlaceholder
! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
loadStateExtended
	"Private. A marshal directive that loads state and then an
	 object that is to be associated with the loaded state."

	^self configuration
		extendedAttributesFor: self loadState 
		are: self loadState.
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadString
	"Load a collection of characters to be loaded in
	 the native single or double byte string class."

	| stringSize answer isOneByte int chr |
	stringSize := self loadSigned.
	stringSize < 0 ifTrue: [^self loadStringNumLines: stringSize negated].
	answer := String new: stringSize.
	isOneByte := true.
	1 to: stringSize do: [:index | 
		chr := SrpPortal unsignedToCharacter: (int := self loadUnsigned). 
		(isOneByte and: [int > 16rFF]) ifTrue: [
			isOneByte := false. 
			answer := SrpPortal stringAsTwoByteString: answer.
		].
		answer srpIndexableValueAt: index put: chr.
	].
	^answer

! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadStringArray
	"Load an array of strings."

	| tableSize answer |
	tableSize := self loadUnsigned.
	answer := Array new: tableSize.
	1 to: tableSize do: [:index | answer at: index put: self loadString ].
	^answer
! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
loadStringNumLines: numLines
	"Load a collection of characters to be loaded in the native single
	 or double byte string class. This also uses the local end-of-line convention."

	| lineLengths stringSize stringPos answer isOneByte delimiter lineLength int chr |
	lineLengths := Array new: numLines.
	delimiter := SrpPortal platformLineDelimiter.
	stringSize := (numLines - 1) * delimiter size.
	1 to: numLines do: [:lineIndex |
		lineLength := self loadUnsigned.
		lineLengths at: lineIndex put: lineLength.
		stringSize := stringSize + lineLength.
	].
	answer := String new: stringSize.
	stringPos := 0.
	isOneByte := true.
	1 to: numLines do: [:lineIndex |
		(lineLengths at: lineIndex) timesRepeat: [
			chr := SrpPortal unsignedToCharacter: (int := self loadUnsigned). 
			(isOneByte and: [int > 16rFF]) ifTrue: [
				isOneByte := false. 
				answer := SrpPortal stringAsTwoByteString: answer.
			].
			answer srpIndexableValueAt: (stringPos := stringPos + 1) put: chr.
		].
		lineIndex = numLines ifFalse: [
			(self loadUnsigned = 13 and: [self loadUnsigned = 10])
				ifFalse: [ self configuration reportError: 'Bad data or encoding. Expected standard line separator in string.' ].
			delimiter do: [:delimChar | answer srpIndexableValueAt: (stringPos := stringPos + 1) put: delimChar ].
		].
	].
	^answer! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadUnsigned
	"Load an unsigned integer (a whole number) from
	 zero to positive infinity. Performance is greatly
	 affected by this method. This is a good candidate
	 for a primitive."

	| answer shifter atLast readValue |
	answer := 0.
	shifter := 1.
	[
		atLast := ((readValue := stream next) // 128) = 0.
		atLast ifFalse: [ readValue := readValue - 128 ].
		answer := answer + (readValue * shifter).
		shifter := shifter * 128.
		atLast 
	] whileFalse: [].
	^answer

	"Another way to do the same thing:
		| answer shifter readValue |
		answer := 0.
		shifter := 0.
		[ 
			readValue := stream next.
			answer := answer + ((readValue bitAnd: 127) bitShift: shifter).
			shifter := shifter + 7.
			readValue < 128
		] whileFalse: [].
		^answer
	"! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadUnsignedArray
	"Load an array of unsigned integers."

	| answerSize answer |
	answerSize := self loadUnsigned.
	answer := Array new: answerSize.
	1 to: answerSize do: [:index | answer at: index put: self loadUnsigned ].
	^answer
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
loadUnsignedWidth
	"Load an unsigned integer that was saved with
	 a fixed width. We don't need to know the width
	 to read it."

	stream srpFinishByteGroupRead.
	^self loadUnsigned
! !

!SrpObjectLoader methodsFor: 'Marshal Directives' stamp: ''!
marshalDirectiveLoadArray
	"Answer an array of selectors that is used to fetch the
	 selector to be performed for a marshal directive code."

	^#(
		loadRelativeOffset	
		loadMetastate
		loadSkipBytes
		loadSkipObject
		loadMetastateTableReference
		loadMetastateTableName
		loadEncodingVersion
		loadStateExtended
	  )
! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
missingBehaviorOf: loadedObject
	"This will be evaluated for each State that is loaded whose
	 metastate behavior could not be found. The return value is 
	 used in place of the loadedObject. Graph references are
	 maintained dispite this exchange."

	^self configuration missingBehaviorOf: loadedObject loadedFrom: self! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
next
	"Load the next object on the stream and perform
	 any post load processing."

	^self configuration 
		loaded: self loadObject
		using: self
! !

!SrpObjectLoader methodsFor: 'Loading' stamp: ''!
processFinishingActions
	"Private. Finishing actions are processed after the root
	 object is completely loaded or saved. Anything that responds to
	 #value can be used as a finishing action."

	self configuration
		processPostLoadActions: finishingActions
		for: self.
	stream srpFinishByteGroupRead.
! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
signalIgnoreLoad
	"Answer an object that can be given to
	 tell the loader to ignore the last object
	 that was loaded."

	^#SrpIgnoreLoadSignal
! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
specialOperationSignal
	"Private. This message is sent when a marshal directive code
	 of zero is loaded. No meaning is currently given for marshal
	 directive zero, but the zero code is reserved for future use."

	^self configuration reportError: 'Invalid data or unrecognized data sequence.'.! !

!SrpObjectLoader methodsFor: 'private' stamp: ''!
startedToLoad: anInstance 
	"Private. Answer anInstance with a placeholder around it. Relative
	 pointers as well as object slot pointers will point to the
	 placeholder instead of directly to anInstance. This allows
	 anInstance to be replaced without losing pointers that
	 were to anInstance."

	| aPlaceholder |
	aPlaceholder := self configuration usePlaceholders 
		ifTrue: [ self configuration newPlaceholderFor: anInstance atLoaderIndex: hitSequence size + 1 ]
		ifFalse: [ anInstance ].
	hitSequence addLast: aPlaceholder.
	context startedToLoad: aPlaceholder.
	^aPlaceholder! !

SrpMarshaler subclass: #SrpObjectSaver
	instanceVariableNames: 'hitList duplicationTypes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpObjectSaver methodsFor: 'Accessors' stamp: ''!
duplicationTypes
	"Answer a set containing class names whose duplicate
	 instances should be saved more than once."

	^duplicationTypes
! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
flush
	"Private. Flush any data buffers. This is done after a
	 root object has been completely written to the stream.
	 Some dialects only implement #flush on file streams."

	(stream respondsTo: #flush)
		ifTrue: [stream flush].! !

!SrpObjectSaver methodsFor: 'Initialization' stamp: ''!
initializeConfiguration: aConfig
	"Private."

	super initializeConfiguration: aConfig.
	hitList := aConfig newHitList.
	duplicationTypes := aConfig duplicationTypes asSet.
	context := aConfig newSaveContext.
	aConfig beingUsedForObjectSaver: self.! !

!SrpObjectSaver methodsFor: 'Testing' stamp: ''!
isSaver
	"Answer boolean whether this instance
	 will be used to save objects."

	^true
! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
lineLengthsForString: aString
	"Answer a collection of lengths for each line in aString. The length
	 of each line doesn't include the length of the delimiter."

	| lineLengths delimiter startPos indexOfDelimiter |
	lineLengths := OrderedCollection new.
	delimiter := SrpPortal platformLineDelimiter.
	startPos := 1.
	[(indexOfDelimiter := SrpPortal sequenceableCollection: aString indexOfSubCollection: delimiter startingAt: startPos) == 0]
	whileFalse: [
		lineLengths add: indexOfDelimiter - startPos.
		startPos := indexOfDelimiter + delimiter size.
	].
	lineLengths add: (aString size + 1) - startPos.
	^lineLengths
! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
maxSignedOfWidth: width
	"Answer the maximum signed integer value that can be stored
	 for the given number of save units (octets/bytes). This would be
	 useful for preparing to save unsigned integers into fixed-width spaces."

	^(self maxUnsignedOfWidth: width) // 2! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
maxSizeToEncodeBytesOfSize: numBytes
	"Answer the theoretical maximum number of physical bytes
	 that could be required to encode numBytes as an unsigned
	 integer stream."

	| unsignedIntegerSize |
	unsignedIntegerSize := numBytes * 8 + 6 // 7.
	^(self stream respondsTo: #srpMaxSizeToEncodeBytesOfSize:)
		ifTrue: [self stream srpMaxSizeToEncodeBytesOfSize: unsignedIntegerSize]
		ifFalse: [unsignedIntegerSize]
	! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
maxUnsignedOfWidth: width
	"Answer the maximum unsigned integer value that can be stored
	 for the given number of save units (octets/bytes). This would be
	 useful for preparing to save unsigned integers into fixed-width spaces."

	^(2 raisedTo: (width * 7)) - 1! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
minSignedOfWidth: width
	"Answer the minimum signed integer value that can be stored
	 for the given number of save units (octets/bytes). This would be
	 useful for preparing to save unsigned integers into fixed-width spaces."

	^((self maxUnsignedOfWidth: width) // 2) negated! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
nextPut: anObject

	self saveObject: anObject.
	self configuration 
		saved: anObject
		using: self.
	^anObject
! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
noteAndExchange: anObject
	"Private."

	| allowDuplicates relativeOffset saveObject |
	(allowDuplicates := self duplicationTypes includes: anObject class name) 
		ifFalse: [ relativeOffset := self relativeOffsetOfSavedObject: anObject ].
	(relativeOffset notNil and: [self shouldSaveRelativeOffsetTo: anObject])
		ifTrue: [ self saveRelativeOffset: relativeOffset. ^#'SRP_FLAG_SAVED_RELATIVE_OFFSET' ].
	hitSequence addLast: anObject.
	relativeOffset notNil 
	ifTrue: [ saveObject := anObject ]
	ifFalse: [
		anObject isNil | allowDuplicates ifFalse: [ hitList at: anObject put: hitSequence size ].
		saveObject := context exchangeOriginal: anObject.
		saveObject ~~ anObject ifTrue: [
			saveObject isNil | allowDuplicates ifFalse: [ hitList at: saveObject put: hitSequence size ].
		].
	].
	^saveObject
! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
processFinishingActions
	"Private. Finishing actions are processed after the root
	 object is completely loaded or saved. Anything that responds to
	 #value can be used as a finishing action."

	self configuration
		processPostSaveActions: finishingActions
		for: self.
	stream srpFinishByteGroupWrite.! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
relativeOffsetOfSavedObject: anObject
	"Private. Zero should never be used for a load offset
	 because zero is used to announce special commands."

	| idx |
	idx := hitList at: anObject ifAbsent: [].
	idx isNil ifTrue: [^nil].
	^hitSequence size - idx + 1.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveAvoid: anObject
	"This is for objects that are mapped to nil by #srpDataTypeMap.
	  No state will be saved for this object."

! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveBoolean: aBoolean

	self saveUnsigned: (aBoolean ifTrue: [1] ifFalse: [0])
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveByte: anInteger

	stream nextPut: anInteger.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveBytes: aByteCollection

	self saveUnsigned: aByteCollection size.
	stream nextPutAll: aByteCollection.
		
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveCharacter: aCharacter
	"SRP supports character values of all sizes--yet wastes little
	 space. The font used to display a character determines if
	 non-English glyphs will show properly."

	self saveUnsigned: (SrpPortal characterToUnsigned: aCharacter).

! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveCharacterArray: aString
	"Save a collection of characters to be loaded in the
	 native String class. This doesn't do any special portability
	 processing for end-of-line characters."

	| str |
	str := aString asString.
	self saveSigned: str size.
	1 to: str size do: [:i | self saveCharacter: (str at: i) ].
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveCharacterArrayArray: anArray
	"Similar to #saveStringArray, but storage encoding is more
	 space efficient by at least one byte when anArray typically
	 contains fewer than 64 items. End-of-line conversions are
	 not performed. Multi-byte characters are supported by
	 encoding, but not loader isn't required to expect them."

	| str |
	1 = anArray size
	ifTrue: [ self saveSigned: anArray last size ]
	ifFalse: [ 
		self saveSigned: anArray size negated.
		self saveUnsigned: anArray first size. 
	].
	1 to: anArray size do: [:nameIndex |
		str := anArray at: nameIndex.
		1 to: str size do: [:i | self saveCharacter: (str at: i) ].
		nameIndex < anArray size ifTrue: [
			self saveUnsigned: (anArray at: nameIndex + 1) size
		].
	].! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveDistinct: anObject

	| bytes proxy |
	self halt: 'Unfinished code. This would loop if the halt were removed.'.
	bytes := self configuration saveObjectAsBinary: anObject.
	proxy := self configuration proxyFor: anObject context: self context.
	self saveSkipBytes: bytes.
	self saveState: proxy.
	anObject srpSavedAs: proxy.
	! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveEncodingVersion: encodingVersion
	"The data is declaring that it is encoding according to
	 a specific SRP encoding standard."

	self saveUnsigned: 0.		"Marshal directive"
	self saveUnsigned: 8.		"Encoding version directive"
	self saveUnsigned: encodingVersion.
! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveMetastate: initialClassDef
	"Metastate describe data so it can be loaded. It is
	 one of the few data structures directly written by SRP
	 instead of stored as an object that can be mapped."

	| classDef |
	classDef := context preMapCommand value: initialClassDef value: context.
	self saveUnsigned: 0.		"Marshal directive"
	self saveUnsigned: 2. 	"loadMetastate"
	self saveUnsigned: classDef code.
	self saveUnsigned: classDef shapeCode.
	self saveCharacterArrayArray: (SrpPortal namespacePathForBehavior: classDef).
	classDef versionNumberArray notNil
		ifTrue: [self saveUnsignedArray: classDef versionNumberArray].
	classDef hasNamedSlots ifTrue: [
		self saveCharacterArrayArray: classDef allInstVarNames.
		classDef hasTypedNamedSlots ifTrue: [
			self saveUnsignedArray: classDef namedStoreTypeCodes.
		].
	].
	classDef hasExtendedAttributes 
		ifTrue: [self saveObject: classDef extendedAttributes].
	initialClassDef srpSavedAs: classDef.
	^classDef
! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveMetastateTableName: metastateTableName
	"Private. This is a marshal directive to specify
	 that metastates be imported from some
	 known or external source identified by the
	 tableReference."

	self saveUnsigned: 0. 	"Marshal directive"
	self saveUnsigned: 6. 	"Marshal directive code"
	self saveString: metastateTableName.


! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveMetastateTableReference: tableReference
	"Private. This is a marshal directive to specify
	 that metastates be imported from some
	 known or external source identified by the
	 tableReference."

	self saveUnsigned: 0. 	"Marshal directive"
	self saveUnsigned: 5. 	"Marshal directive code"
	self saveState: tableReference.


! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveObject: anObject

	| extendedAttributes |
	self context encodingVersion > 0
		ifTrue: [self saveEncodingVersion: self context encodingVersion].
	(extendedAttributes := self configuration extendedAttributesFor: anObject) isNil
		ifTrue:  [self saveState: anObject]
		ifFalse: [self saveState: anObject extended: extendedAttributes].
	self processFinishingActions.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveProxy: anObject
	"This is for objects that to be exchanged with a proxy as mapped
	 by #srpDataTypeMap."

	| proxy |
	proxy := self configuration proxyFor: anObject context: self context.
	self saveState: proxy.
! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveRelativeOffset: relativeOffset
	"Private. A marshal directive to refer to an
 	object that has already been saved."

	self saveUnsigned: 0. 	"Marshal directive"
	self saveUnsigned: 1. 	"Marshal directive code"
	self saveUnsigned: relativeOffset.


! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveRelativeOffsetOf: anObject
	"Private. A convenience method to save a marshal
	directive to refer to anObject that must have already
	been saved."

	| offset |
	offset := self relativeOffsetOfSavedObject: anObject.
	offset isNil ifFalse: [ ^self configuration reportError: 'Attempt to reference unsaved object.' ].
	self saveRelativeOffset: offset.! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSigned133Bit: anInteger
	"	(2 raisedTo: 133) - 1 = 10889035741470030830827987437816582766591	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveSigned: anInteger width: 19.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSigned21Bit: anInteger
	"	(2 raisedTo: 21) - 1 = 2097151	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveSigned: anInteger width: 3.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSigned35Bit: anInteger
	"	(2 raisedTo: 35) - 1 = 34359738367	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveSigned: anInteger width: 5.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSigned70Bit: anInteger
	"	(2 raisedTo: 70) - 1 = 1180591620717411303423	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveSigned: anInteger width: 10.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSigned: anInteger
	"Save an integer from zero to negative or positive infinity.
	 The least significant bit of the number is ON if the value
	 is a negative number."

	| int |
	int := anInteger asInteger.
	int < 0
		ifTrue:  [ self saveUnsigned: (int negated * 2) + 1 ]
		ifFalse: [ self saveUnsigned: (int * 2) ].

	" Another way to do the same thing:
		| int |
		int := anInteger asInteger.
		int < 0
			ifTrue:  [ self saveUnsigned: (int negated bitShift: 1) + 1 ]
			ifFalse: [ self saveUnsigned: (int bitShift: 1) ].
	"! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSigned: anInteger width: width
	"Save an integer from zero to negative or positive infinity.
	 The least significant bit of the number is ON if the value
	 is a negative number."

	| int |
	int := anInteger asInteger.
	int < 0
		ifTrue:  [ self saveUnsigned: (int negated * 2) + 1 width: width ]
		ifFalse: [ self saveUnsigned: (int * 2) width: width ].

	" Another way to do the same thing:
		| int |
		int := anInteger asInteger.
		int < 0
			ifTrue:  [ self saveUnsigned: (int negated bitShift: 1) + 1 width: width ]
			ifFalse: [ self saveUnsigned: (int bitShift: 1) width: width ].
	"! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveSignedArray: anArray width: width
	"Save an array of signed integers of a fixed width."

	self saveUnsigned: anArray size.
	anArray do: [:anInteger | self saveSigned: anInteger width: width ].
! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveSkipBytes: aByteArray
	"Private. A marshal directive to save a given
	 byte array to the stream expecting that the
	 loader will simply ignore those bytes."

	self saveUnsigned: 0. 	"Marshal directive"
	self saveUnsigned: 3. 	"Marshal directive code"
	(stream respondsTo: #srpFinishByteGroupWrite) ifTrue: [stream srpFinishByteGroupWrite].
	self saveBytes: aByteArray.
	(stream respondsTo: #srpFinishByteGroupWrite) ifTrue: [stream srpFinishByteGroupWrite].


! !

!SrpObjectSaver methodsFor: 'Marshal directives' stamp: ''!
saveSkipObject: anObject
	"Private. A marshal directive to save anObject
	 to the stream expecting that the loader will
	 simply ignore the object when loading."

	self saveUnsigned: 0. 	"Marshal directive"
	self saveUnsigned: 4. 	"Marshal directive code"
	self saveState: anObject.


! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveState: anObject
	"Private."

	| saveObject classDef |
	context pushStack.
	saveObject := self noteAndExchange: anObject.
	saveObject == #'SRP_FLAG_SAVED_RELATIVE_OFFSET' ifFalse: [
		classDef := self savedMetastateForClass: saveObject srpClass.
		classDef saveInstance: saveObject using: self.
		anObject srpSavedAs: saveObject.
	].
	context popStack.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveState: anObject extended: extendedAttributes
	"Private"

	self saveUnsigned: 0.		"Marshal directive"
	self saveUnsigned: 8. 	"loadStateExtended"
	self saveState: anObject.
	self saveState: extendedAttributes.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveString: aString
	"Save a collection of characters to be loaded in the
	 native String class. If the string contains any LF
	 characters then it will be saved in a portable
	 end-of-line form."

	| str |
	str := aString asString.
	(SrpPortal sequenceableCollection: str indexOfSubCollection: SrpPortal platformLineDelimiter startingAt: 1) > 0
		ifTrue: [ self saveStringAsLines: str ]
		ifFalse: [ self saveCharacterArray: str ].
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveStringArray: anArray
	"Save an array of strings. Multi-byte characters are
	 supported by encoding, and loader is required to
	 expect them. End-of-line conversions are performed."

	self saveUnsigned: anArray size.
	anArray do: [:string | self saveString: string ].! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
saveStringAsLines: aString
	"Save aString in a form that allows it to be quickly loaded using
	 whatever end-of-line convention applies to the loading platform.
	 Performance was more important than code clarity."

	| str lineLengths strIndex lfChar crChar delimiter lineLength char |
	delimiter := SrpPortal platformLineDelimiter.
	lfChar := SrpPortal unsignedToCharacter: 10.
	crChar := SrpPortal unsignedToCharacter: 13.
	str := aString asString.
	lineLengths := self lineLengthsForString: str.
	self saveSigned: lineLengths size negated.		"number of lines"
	lineLengths do: [:eachLength | self saveUnsigned: eachLength ].
	strIndex := 0.
	1 to: lineLengths size do: [:lineIndex |
		lineLength := lineLengths at: lineIndex.
		lineLength timesRepeat: [
			char := str at: (strIndex := strIndex + 1).
			self saveCharacter: char.
		].
		lineIndex = lineLengths size ifFalse: [
			strIndex := strIndex + delimiter size.
			self saveCharacter: crChar.
			self saveCharacter: lfChar.
		].
	].
	strIndex = str size
		ifFalse: [ self error: 'SRP internal error while saving a string' ].
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsigned133Bit: anInteger
	"	(2 raisedTo: 133) - 1 = 10889035741470030830827987437816582766591	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveUnsigned: anInteger width: 19.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsigned21Bit: anInteger
	"	(2 raisedTo: 21) - 1 = 2097151	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveUnsigned: anInteger width: 3.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsigned35Bit: anInteger
	"	(2 raisedTo: 35) - 1 = 34359738367	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveUnsigned: anInteger width: 5.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsigned70Bit: anInteger
	"	(2 raisedTo: 70) - 1 = 1180591620717411303423	"
	"Save an unsigned integer in an addressable position
	 with a fixed data width so the value can be updated."

	self saveUnsigned: anInteger width: 10.
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsigned: anInteger
	"Save an unsigned integer (a whole number) from
	 zero to positive infinity. Performance is 
	 affected by this method. This is a good candidate
	 for a primitive."

	| writeValue carry |
	carry := anInteger asInteger.
	[
		writeValue := carry \\ 128.
		(carry := carry // 128) = 0 
	] whileFalse: [ stream nextPut: writeValue + 128 ].
	stream nextPut: writeValue.

	"Another way to do the same thing:
		| writeValue carry |
		carry := anInteger asInteger.
		carry < 0 ifTrue: [self error].
		[
			writeValue := carry bitAnd: 127.
			(carry := carry bitShift: -7) = 0 
		] whileFalse: [ stream nextPut: (writeValue + 128) ].
		stream nextPut: writeValue.
	"
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsigned: anInteger width: width
	"Save an unsigned integer (a whole number) from
	 zero to ((2 raisedTo: (width * 7)) - 1). Performance is 
	 affected by this method. This is a good candidate
	 for a primitive."

	"	(0 to: 20) collect: [:width | (2 raisedTo: (width * 7)) - 1 ].	"

	| writeValue carry available |
	stream srpFinishByteGroupWrite.
	available := width.
	carry := anInteger asInteger.
	[
		writeValue := carry \\ 128.
		(carry := carry // 128) = 0 
	] whileFalse: [ 
		stream nextPut: writeValue + 128.
		(available := available - 1) = 0 ifTrue: [self error: 'Space assumption failure.'].
	].
	available - 1 timesRepeat: [
		stream nextPut: writeValue + 128.
		available := available - 1.
		writeValue := 0.
	].
	available = 1 ifTrue: [stream nextPut: writeValue].
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsignedArray: anArray
	"Save an array of strings."

	self saveUnsigned: anArray size.
	anArray do: [:anInteger | self saveUnsigned: anInteger ].
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
saveUnsignedArray: anArray width: width
	"Save an array of unsigned integers of a fixed width."

	self saveUnsigned: anArray size.
	anArray do: [:anInteger | self saveUnsigned: anInteger width: width ].
! !

!SrpObjectSaver methodsFor: 'Saving' stamp: ''!
savedMetastateForClass: aClass
	"Private."

	| classDef |
	classDef := classTable metastateForClass: aClass.
	classDef == nil ifTrue: [
		classDef := classTable addMetastate: (self configuration newMetastateForClass: aClass).
		self saveMetastate: classDef.
	].
	^classDef
! !

!SrpObjectSaver methodsFor: 'Initialization' stamp: ''!
setMetastateTable: tableReference
	"A schema table is a collection of SrpMetastate that is
	 referenced by the data instead of encoded in the data. Using
	 a metastate table improves performance and reduces data size. Use
	 also requires that the metastate table is defined or can be retrieved
	 by a loading object space. The argument tableReference can be
	 any object that can respond to #allMetastates when asked in
	 both saving and loading object spaces."

	| metastates |
	metastates := tableReference allMetastates.
	metastates isNil 
	ifTrue: ["This error is optional" self error: 'Unable to resolve a metastate table named ', tableReference name]
	ifFalse: [
		tableReference wouldNameSuffice
		ifTrue: [self saveMetastateTableName: tableReference name]
		ifFalse: [self saveMetastateTableReference: tableReference].
		classTable addAllMetastates: metastates.
	].
! !

!SrpObjectSaver methodsFor: 'private' stamp: ''!
shouldSaveRelativeOffsetTo: anObject
	"You would usually want to save a reference to a previously
	 saved object, but there are special circumstances in which you
	 wouldn't. If mapping rules exchange an object, and the new object
	 references the original (recursion), then you may want to now
	 save the original object rather than a reference to the new object."

	| mappedRecursive |
	(mappedRecursive := context recursiveMappingFor: anObject)
		isNil ifTrue: [^true].
	^mappedRecursive srpRecursiveReferencesShouldReferenceMapped
! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
widthOfSigned: anInteger
	"Answer the number of save units (octets/bytes) that would be
	 required to save anInteger. This would be useful for preparing
	 to save unsigned integers into fixed-width spaces."

	| int |
	int := anInteger asInteger.
	^int < 0
		ifTrue:  [ self widthOfUnsigned: (int negated * 2) + 1 ]
		ifFalse: [ self widthOfUnsigned: (int * 2) ].
! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
widthOfSignedToEncodeBytesOfSize: numBytes

	| unsigned |
	unsigned := self maxSizeToEncodeBytesOfSize: numBytes.
	^self widthOfSigned: unsigned
! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
widthOfUnsigned: anInteger
	"Answer the number of save units (octets/bytes) that would be
	 required to save anInteger. This would be useful for preparing
	 to save unsigned integers into fixed-width spaces."

	| answer |
	answer := 0.
	"	(0 to: 16) collect: [:width | (2 raisedTo: (width * 7)) - 1 ]	"
	#(0 127 16383 2097151 268435455 34359738367 4398046511103 562949953421311
	 72057594037927935 9223372036854775807 1180591620717411303423
	 151115727451828646838271 19342813113834066795298815
	 2475880078570760549798248447 316912650057057350374175801343
	 40564819207303340847894502572031 5192296858534827628530496329220095 )
	do: [:max | 
		max >= anInteger ifTrue: [^answer].
		answer := answer + 1.
	].
	"Wow, this is a big one!!"
	[ ((2 raisedTo: (answer * 7)) - 1) >= anInteger ]	 whileFalse: [answer := answer + 1].
	^answer


! !

!SrpObjectSaver methodsFor: 'Utility' stamp: ''!
widthOfUnsignedToEncodeBytesOfSize: numBytes

	| unsigned |
	unsigned := self maxSizeToEncodeBytesOfSize: numBytes.
	^self widthOfUnsigned: unsigned
! !

Object subclass: #SrpMetastate
	instanceVariableNames: 'code instanceShape behaviorName versionNumberArray behavior allInstVarNames namedStoreTypeCodes mapTable extendedAttributes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpMetastate classSide methodsFor: 'Instance creation' stamp: ''!
code: classCode shapeCode: shapeCode behaviorName: behaviorName

	^self new
		code: classCode;
		shapeCode: shapeCode;
		behaviorName: behaviorName.! !

!SrpMetastate classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 Instances of this class are created to describe the state of instances that
 are saved to a stream. They are usually saved to to the data stream along
 with instances to describe to a loader how instances should be loaded.
 They are also responsible for mapping instance variables.

 Direct data stream access is permitted, but is not recommended because
 it requires precise loading rules to be known and maintained in all loading
 object spaces. Likewise, data typing is permitted but is not recommended.
"! !

!SrpMetastate classSide methodsFor: 'Instance creation' stamp: ''!
new
	^super new initialize! !

!SrpMetastate classSide methodsFor: 'Instance creation' stamp: ''!
newLikeClass: aClass configuration: config

	(aClass isKindOf: self) ifTrue: [^aClass shallowCopy].
	^self new
		conformTo: aClass configuration: config;
		yourself! !

!SrpMetastate classSide methodsFor: 'Not categorized' stamp: ''!
storeTypeArrays
	"Only add to the bottom of this array to avoid breaking backward compatibility."

	^#(
		(State							saveState:							loadState)
		(Unsigned					saveUnsigned:					loadUnsigned)
		(Character					saveCharacter:					loadCharacter)
		(Signed						saveSigned:						loadSigned)
		(Byte							saveByte:							loadByte)
		(String							saveString:							loadString)
		(CharacterArray			saveCharacterArray:			loadCharacterArray)
		(CharacterArrayArray	saveCharacterArrayArray:	loadCharacterArrayArray)
		(ByteArray					saveByteArray:					loadByteArray)
		(Boolean						saveBoolean:						loadBoolean)
		(StringArray					saveStringArray:					loadStringArray)
		(UnsignedArray			saveUnsignedArray:			loadUnsignedArray)
		(Proxy							saveProxy:							loadProxy)
		(nil								saveAvoid:							loadAvoid)
		(Distinct						saveDistinct:						loadDistinct)
	  )
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
allInstVarNames
	"Answer an indexable collection containing all the instance variable
	 names specified by the receiver and its superclasses. The order of the
	 collection is the same as it is in an instance of the receiver."

	^allInstVarNames
! !

!SrpMetastate methodsFor: 'private' stamp: ''!
allInstVarNames: anArray
	"Private. Set an indexable collection containing all the instance variable
	 names specified by the receiver and its superclasses. The order of the
	 collection is the same as it is in an instance of the receiver."

	allInstVarNames := anArray.
! !

!SrpMetastate methodsFor: 'Object Loading' stamp: ''!
asBehavior
	"Answer the behavior (class) residing in the current object space
	 that is resolved using behaviorName. The behavior instance
	 variable is used as a cache to improve load performance."

	| resident |
	behavior isNil ifTrue: [
		resident := SrpPortal behaviorForNamespacePath: behaviorName.
		resident isNil
		ifFalse: [ self behavior: resident ]
		ifTrue: [ self ignoreResidentBehavior ].
	].
	^behavior! !

!SrpMetastate methodsFor: 'Object Loading' stamp: ''!
behavior
	"Answer the cached behavior that will be used for creating instances."

	^behavior! !

!SrpMetastate methodsFor: 'Object Loading' stamp: ''!
behavior: aBehavior
	"Answer the cached behavior that will be used for creating instances."

	behavior := aBehavior.
	mapTable := nil.! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
behaviorName
	"Answer the name of a behavior that will be used to
	 retrieve a behavior in a loading objectspace. The
	 behaviorName is an array of symbols with
	 namespace names (ex. #(Smalltalk Core String))."

	^behaviorName! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
behaviorName: anArray
	"Set the name of a behavior that will be used to
	 retrieve a behavior in a loading objectspace. The
	 behaviorName is an array of symbols with
	 namespace names (ex. #(Smalltalk Core String))."

	behaviorName := anArray! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
behaviorNameString
	"Answer the name of the behavior as a String
	 with namespaces separated by a period."

	| ans |
	self behaviorName size = 0 ifTrue: [String new].
	ans := self behaviorName last asString.
	self behaviorName size - 1 to: 1 by: -1 
		do: [:i | ans := (self behaviorName at: i), '.', ans ].
	^ans
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
behaviorShortName
	"Answer the name of a behavior that will be used to
	 retrieve a behavior in a loading objectspace. The
	 behaviorName is an array of symbols with
	 namespace names (ex. #(Smalltalk Core String))."

	^self behaviorName last! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
behaviorShortName: aSymbol
	"Set the name of a behavior that will be used to
	 retrieve a behavior in a loading objectspace. The
	 behaviorName is an array of symbols with
	 namespace names (ex. #(Smalltalk Core String))."

	self behaviorName: (Array with: aSymbol)! !

!SrpMetastate methodsFor: 'private' stamp: ''!
code

	^code
! !

!SrpMetastate methodsFor: 'private' stamp: ''!
code: anInteger

	code := anInteger
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
conformTo: aClass configuration: config
	"Make the receiver conform to the characteristics of aClass."

	| typeMap primaryStoreType selfTyped slotIndex |
	behaviorName := SrpPortal namespacePathForBehavior: aClass.
	instanceShape initialize.
	allInstVarNames := aClass allInstVarNames asArray.
	typeMap := config srpDataTypeMapForClass: aClass.
	primaryStoreType := self defaultPrimaryStoreTypeForClass: aClass.
	selfTyped := false.
	typeMap do: [:array | 
		(array at: 2) == nil 
			ifTrue: [allInstVarNames := allInstVarNames reject: [:stringOrSymbol | SrpPortal sequenceableCollection: (array at: 1) equals: stringOrSymbol]]
			ifFalse: [instanceShape isTyped: true].
		((array at: 1) == #self or: [(array at: 1) == #IndexableType]) ifTrue: [
			primaryStoreType := array at: 2.
			(array at: 1) == #self ifTrue: [selfTyped := true].
		].
	].
	self primaryStoreType: primaryStoreType.
	selfTyped ifTrue: [
		instanceShape isNamed:  false.
		instanceShape isIndexed: false.
	] ifFalse: [
		instanceShape isNamed:  allInstVarNames isEmpty not.
		instanceShape isIndexed: aClass srpHasIndexableValues.
		typeMap do: [:array |
			slotIndex := 0.
			allInstVarNames do: [:stringOrSymbol |
				slotIndex := slotIndex + 1.
				(SrpPortal sequenceableCollection: (array at: 1) equals: stringOrSymbol) ifTrue: [
					self namedStoreTypeCodes at: slotIndex put: (self storeTypeCodeForType: (array at: 2)).
				].
			].
		].
	].
	self behavior: aClass.
	self versionNumberArray: (config versionNumberArrayFor: aClass).
	self extendedAttributes: (config extendedAttributesFor: aClass).
	config createdMetastate: self modeledAfter: aClass.! !

!SrpMetastate methodsFor: 'Shape - Primary Store Type' stamp: ''!
defaultPrimaryStoreTypeForClass: aClass
	"The primary store type applies to either indexed slots
	 of an object or to the entire object. For most objects
	 the primary store type is #State. The default store
	 type is overridden by type declarations in 
	 #srpDataTypeMap for each class of object."

	(SrpPortal behaviorIsPointers: aClass) ifTrue: [^#State].
	(SrpPortal behaviorIsBytes: aClass) ifTrue: [^#Byte].
	^#Unsigned
! !

!SrpMetastate methodsFor: 'Extended attributes' stamp: ''!
extendedAttributes
	"Answer nil or an object that is saved and loaded as an object along with
	 a metastate. The for simplicity, object is expected to conform to
	 basic Smalltalk dictionary protocol. Attributes are associated to
	 avoid potential recursion problems while saving or loading metastate."

	^extendedAttributes! !

!SrpMetastate methodsFor: 'Extended attributes' stamp: ''!
extendedAttributes: aDictionary
	"Answer nil or an object that is saved and loaded as an object along with
	 a metastate. The for simplicity, object is expected to conform to
	 basic Smalltalk dictionary protocol. Attributes are associated to
	 avoid potential recursion problems while saving or loading metastate."

	extendedAttributes := aDictionary.
	instanceShape hasExtendedAttributes: extendedAttributes notNil.
! !

!SrpMetastate methodsFor: 'Extended attributes' stamp: ''!
hasExtendedAttributes

	^instanceShape hasExtendedAttributes 
! !

!SrpMetastate methodsFor: 'Shape - State Structure' stamp: ''!
hasNamedSlots
	"Answer boolean whether the stored definition includes
	 named slots."

	^instanceShape isNamed
! !

!SrpMetastate methodsFor: 'Shape - State Structure' stamp: ''!
hasTypedNamedSlots
	"Answer boolean whether the stored definition includes
	 typed named slots that need to saved and loaded with
	 the metastate."

	^instanceShape isTyped 
		"This second test is for old use of #isTyped and is otherwise unnecessary."
		and: [instanceShape isNamed]
! !

!SrpMetastate methodsFor: 'Behavior version number' stamp: ''!
hasVersionNumberArray
	"A version number array is application defined data stored with
	 a metastate. Answer boolean whether a version number array
	 has been set."

	^instanceShape hasVersionNumberArray
! !

!SrpMetastate methodsFor: 'Object Loading' stamp: ''!
ignoreResidentBehavior
	"Set the receiver to ignore any resident behavior of the same name
	 that might exist in the image. When set, instances of SrpState will
	 be created instead of instances of a resident behavior."
	
	behavior := self.
! !

!SrpMetastate methodsFor: 'private' stamp: ''!
initialize

	instanceShape := self newShapeDefinition.	"Default: #Empty"
	allInstVarNames := #().
	"namedStoreTypeCodes := nil.
	behavior := nil.
	mapTable := nil."! !

!SrpMetastate methodsFor: 'private' stamp: ''!
initializeCache
	"Initialize instance variables only used to
	 cache objects for better performance."

	mapTable := nil.! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
instSize
	"Answer the number of named instance variables specified by the receiver
	 and its superclasses."

	^allInstVarNames size! !

!SrpMetastate methodsFor: 'Mapping' stamp: ''!
instVarMappingFrom: oldInstVariables
	"Private - Return a mapping from the old instance string
	 to the current one.

	 The returned structure is an array. The size of the array
	 is the current size of the instances. Each slot contains
	 the offset of the instance variable under the old format.
	 The value is 0 if there is no such variable."

	| oldInstVarsCopy currentInstVariables mapping count index |
	oldInstVarsCopy := oldInstVariables copy.
	currentInstVariables := self allInstVarNames.
	mapping := Array new: currentInstVariables size.
	count := 1.
	currentInstVariables do: [:string |
		index := oldInstVarsCopy indexOf: string.
		mapping at: count put: index.
		index = 0 ifFalse: [
			oldInstVarsCopy at: index put: nil].
		count := count + 1].
	^mapping! !

!SrpMetastate methodsFor: 'Shape' stamp: ''!
instanceShape

	^instanceShape
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
instanceVariableString
	"Answer a String containing all the instance variable
	 names defined by the receiver.  The names are separated
	 with spaces."

	| stream |
	self allInstVarNames isEmpty ifTrue: [^''].
	stream := WriteStream on: (String new: 20).
	self allInstVarNames do: [:var |
		stream nextPutAll: var; space].
	^stream contents
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isBits
	"Added for polymorphism with Smalltalk behaviors.
	 Answer whether the indexable variable slots contain
	 objects considered bit-like objects by Smalltalk."

	instanceShape isSelfTyped ifTrue: [^false].
	^#(Unsigned Character Byte)
		includes: self primaryStoreType.

	"An integer,  character, or byte by itself isn't normally
	 considered a bit object by Smalltalk, but those objects
	 are considered bits when placed in indexable slots. Signed
	 integers don't qualify as bits because Smalltalk only allows
	 unsigned integers to be placed in indexable slots."

	! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isBytes
	"Added for polymorphism with Smalltalk behaviors."

	^self primaryStoreType == #Byte! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isFixed
	"Answer whether the receiver does not have a variable (indexable) part."

	^self isVariable not! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isLongs
	"Added for polymorphism with Smalltalk behaviors.
	 Longs in Smalltalk are unsigned values of a given size.
	 SRP can store any size values, so answer true even
	 though we don't know the original size of indexable data."

	self == self asBehavior
		ifFalse: [self asBehavior isLongs].
	^self primaryStoreType == #Unsigned! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isPointers
	"Answer whether the variables of the receiver's instances
	 point to other objects instead of bytes."

	^self isBits not! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isVariable
	"Answer whether the receiver has a variable (indexable) part."

	^instanceShape isIndexed! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
isWords
	"Added for polymorphism with Smalltalk behaviors.
	 Wongs in Smalltalk are unsigned values of a given size.
	 SRP can store any size values, so answer true even
	 though we don't know the original size of indexable data."

	self == self asBehavior
		ifFalse: [self asBehavior isWords].
	^self primaryStoreType == #Unsigned! !

!SrpMetastate methodsFor: 'Object Loading' stamp: ''!
loadInstanceUsing: aLoader
	"Create an instance for the receiver and then load the contents
	 of the anInstance from aLoader."

	| anInstance isVarLength aPlaceholder containsPlaceholders mappingTable inst loadSelector loadedObject mappedIndex indexableSize |
	instanceShape isStateless ifTrue: [
		inst := self asBehavior srpNew.
		inst := aLoader startedToLoad: inst.
		inst srpCanDiscard: true.
		^inst
	].
	instanceShape isSelfTyped ifTrue: [
		inst := aLoader perform: self primaryLoadingSelector.
		inst := aLoader startedToLoad: inst.
		inst srpCanDiscard: true.
		^inst
	].
	anInstance := (isVarLength := self isVariable)
		ifTrue:  [ self asBehavior srpNew: (indexableSize := aLoader loadUnsigned) ]
		ifFalse: [ self asBehavior srpNew ].
	aPlaceholder := aLoader startedToLoad: anInstance.
	aLoader context pushStack.
	containsPlaceholders := false.
	"Load indexed variables -- if there are any."
	isVarLength ifTrue: [
		loadSelector := self primaryLoadingSelector.
		1 to: indexableSize do: [:indexedVarIndex |
			loadedObject := aLoader perform: loadSelector.
			anInstance srpIndexableValueAt: indexedVarIndex put: loadedObject.
			containsPlaceholders := loadedObject srpIsPlaceholder or: [containsPlaceholders]
		].
	].
	"Load named variables -- if there are any."
	mappingTable := self mapTable.
	1  to: self instSize do: [:instVarIndex |
		loadSelector := (self storeTypeArrays
			at: (self namedStoreTypeCodes at: instVarIndex) + 1)
				at: 3.
		loadedObject := aLoader perform: loadSelector.
		(mappedIndex := mappingTable at: instVarIndex) isNil ifFalse: [
			anInstance srpNamedValueAt: mappedIndex put: loadedObject.
			containsPlaceholders := loadedObject srpIsPlaceholder or: [containsPlaceholders]
		].
	].
	aPlaceholder srpCanDiscard: containsPlaceholders not.
	aLoader context popStack.
	^aPlaceholder! !

!SrpMetastate methodsFor: 'Mapping' stamp: ''!
mapTable

	mapTable isNil
		ifTrue: [ mapTable := self newMapTable ].
	^mapTable

! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
name
	"Answer a string which if evaluated in the receiver's environment
	 returns the receiver."

	behaviorName isNil
		ifTrue: [^'an unnamed Class']
		ifFalse: [^behaviorName last]! !

!SrpMetastate methodsFor: 'Shape - State Structure' stamp: ''!
namedStoreTypeCodes
	"Answer an array of store type codes with one code
	 for each named instance variable. The default 
	 store type is #State."

	namedStoreTypeCodes isNil ifTrue: [
		namedStoreTypeCodes := allInstVarNames collect: [:ea | 0 ].
	].
	^namedStoreTypeCodes
! !

!SrpMetastate methodsFor: 'Shape - State Structure' stamp: ''!
namedStoreTypeCodes: anArray
	"Specify an array of store type codes with one code
	 for each named instance variable."

	namedStoreTypeCodes := anArray.
! !

!SrpMetastate methodsFor: 'Extended attributes' stamp: ''!
newExtendedAttributePool
	"Answer a new object that will be saved and later loaded along with
	 a metastate. The for simplicity, object is expected to conform to
	 basic Smalltalk dictionary protocol. Attributes are associated to
	 avoid potential recursion problems while saving or loading metastate."

	^Dictionary new! !

!SrpMetastate methodsFor: 'Mapping' stamp: ''!
newMapTable

	| destination |
	(destination := self asBehavior) == self
		ifTrue: [^1 to: self instSize].
	^self 
		newTableMapping: self allInstVarNames 
		to: self asBehavior allInstVarNames! !

!SrpMetastate methodsFor: 'Shape' stamp: ''!
newShapeDefinition

	^SrpMetastateShape new
! !

!SrpMetastate methodsFor: 'Mapping' stamp: ''!
newTableMapping: oldNames to: newNames

	| answer |
	answer := Array new: oldNames size.
	1 to: oldNames size do: [:i |
		answer at: i put: (
			newNames 
				indexOf: (oldNames at: i)
				ifAbsent: [ nil ]
		).
	].
	^answer
! !

!SrpMetastate methodsFor: 'Shape - Primary Store Type' stamp: ''!
primaryLoadingSelector

	^(self storeTypeArrays
		at: instanceShape primaryStoreTypeCode + 1)
			at: 3
! !

!SrpMetastate methodsFor: 'Shape - Primary Store Type' stamp: ''!
primarySavingSelector

	^(self storeTypeArrays
		at: instanceShape primaryStoreTypeCode + 1)
			at: 2
! !

!SrpMetastate methodsFor: 'Shape - Primary Store Type' stamp: ''!
primaryStoreType
	"The primary store type applies to either indexable slots
	 of an object or to the entire object. For most objects
	 the default store type is #State. The primary store
	 type is overridden by type declarations in 
	 #srpDataTypeMap for each class of object."

	^(self storeTypeArrays
			at: instanceShape primaryStoreTypeCode + 1)
				at: 1! !

!SrpMetastate methodsFor: 'Shape - Primary Store Type' stamp: ''!
primaryStoreType: aSymbol
	"The primary store type applies to either indexable slots
	 of an object or to the entire object. For most objects
	 the default store type is #State. The primary store
	 type is overridden by type declarations in 
	 #srpDataTypeMap for each class of object."

	instanceShape primaryStoreTypeCode: 
		(self storeTypeCodeForType: aSymbol)
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
printOn: aStream

	| storeType |
	code isNil ifFalse: [
		code printOn: aStream.
		aStream nextPutAll: ' '.
	].
	aStream nextPutAll: self behaviorNameString.
	instanceShape isSelfTyped ifTrue: [
		aStream nextPutAll: '-'.
		aStream nextPutAll: instanceShape primaryStoreType.
	] ifFalse: [
		instanceShape isIndexed ifTrue: [
			aStream nextPutAll: ' Indexed'.
			storeType := instanceShape primaryStoreType.
			storeType == #State ifFalse: [
				aStream nextPut: $-; nextPutAll: storeType.
			].
		].
	].
	allInstVarNames size > 0 ifTrue: [
		aStream nextPutAll: ' ('.
		1 to: allInstVarNames size do: [:i |
			aStream nextPutAll: (allInstVarNames at: i).
			namedStoreTypeCodes isNil ifFalse: [
				storeType := self storeTypeForCode: (namedStoreTypeCodes at: i).
				storeType == #State ifFalse: [
					aStream nextPut: $-; nextPutAll: storeType.
				].
			].
			aStream space.
		].		
		aStream nextPut: $).
	].
! !

!SrpMetastate methodsFor: 'Object Saving' stamp: ''!
saveInstance: anInstance using: aSaver
	"Write the classCode for anInstance, then write the contents
	 of anInstance to aSaver. Indexed slots are written before named
	 only because indexed can have a known size and that may
	 useful for some other purpose."

	| savingSelector |
	aSaver saveUnsigned: self code.
	instanceShape isStateless ifTrue: [^self].
	(instanceShape isSelfTyped and: [aSaver context isRoot not or: [self primaryStoreType ~~ #Distinct]]) ifTrue: [
		aSaver
			perform: self primarySavingSelector
			with: anInstance.
		^self
	].
	"Save indexed variables -- if there are any."
	self isVariable ifTrue: [
		aSaver saveUnsigned: anInstance srpIndexableSize.
		savingSelector := self primarySavingSelector.
		1 to: anInstance srpIndexableSize do: [:indexedVarIndex |
			aSaver perform: savingSelector with: (anInstance srpIndexableValueAt: indexedVarIndex).
		].
	].
	"Save named variables -- if there are any."
	1  to: self instSize do: [:instVarIndex |
		savingSelector := (self storeTypeArrays
			at: (self namedStoreTypeCodes at: instVarIndex) + 1)
				at: 2.
		aSaver
			perform: savingSelector
			with: (anInstance srpNamedValueAt: instVarIndex).
	].! !

!SrpMetastate methodsFor: 'private' stamp: ''!
shallowCopy
	"A metastate is copied before it is resaved. The copy needs to be fixed
	 up so that it doesn't refer to an old code and so that the copy acts as it's
	 own behavior when a real Behavior is absent."

	| myCopy |
	myCopy := super shallowCopy initializeCache.
	myCopy code: nil.
	behavior == self ifTrue: [myCopy behavior: myCopy].
	^myCopy! !

!SrpMetastate methodsFor: 'Shape' stamp: ''!
shapeCode

	^instanceShape integer
! !

!SrpMetastate methodsFor: 'Shape' stamp: ''!
shapeCode: anInteger

	instanceShape integer: anInteger
! !

!SrpMetastate methodsFor: 'Object Saving' stamp: ''!
srpDataTypeMap

	^self class srpDataTypeMap! !

!SrpMetastate methodsFor: 'Object Loading' stamp: ''!
srpLoadPostMap: aContext
	"Notification that the receiver was loaded."

	self initializeCache.
	(self extendedAttributes notNil and: [self asBehavior notNil]) ifTrue: [
		aContext configuration
			extendedAttributesFor: self asBehavior
			are: self extendedAttributes
	].
	aContext configuration
		loadedMetastate: self 
		context: aContext.

! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
srpNew
	"Answer a fixed size instance of the receiver. Do not
	 override this method."

	^SrpState newMetastate: self! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
srpNew: anInteger 
	"Answer an instance of the receiver with anInteger number
	 of indexable slots (either variables or bytes depending on
	 the receiver's type). Error if the receiver does not allow
	 variable length instances. Do not override this method."

	^SrpState new: anInteger metastate: self! !

!SrpMetastate methodsFor: 'Object Saving' stamp: ''!
srpSavePostMap: aContext
	"Answer an object that is to represent the receiver when
	 saved. This is sent after class-based save mapping rules
	 are applied."

	self initializeCache.
! !

!SrpMetastate methodsFor: 'Object Saving' stamp: ''!
srpSavePreMap: aContext
	"Before saving a metastate we want to ensure that the
	 class version name is updated to reflect whatever the class
	 says it should be."

	behavior isNil ifFalse: [
		self versionNumberArray: (aContext configuration versionNumberArrayFor: behavior).
	].
	aContext configuration
		savingMetastate: self 
		context: aContext.
! !

!SrpMetastate methodsFor: 'Object Saving' stamp: ''!
srpSavedAs: savedObject
	"Notification that the receiver was saved."

	self initializeCache.
! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
srpSlotNamed: stringOrSymbol
	"Private. Answer an integer value. Answer zero if not found."

	| myVarNames |
	myVarNames := self allInstVarNames.
	1 to: myVarNames size do: [:i |
		(SrpPortal sequenceableCollection: (myVarNames at: i) equals: stringOrSymbol)
			ifTrue: [^i].
	].
	^0! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
srpVersionNumberArray
	"Answer nil or an array of unsigned integers that represent the class
	 version number of the receiver. This method is sent when a metastate
	 is resaved."

	^self versionNumberArray! !

!SrpMetastate methodsFor: 'Shape' stamp: ''!
storeTypeArrays

	^self class storeTypeArrays! !

!SrpMetastate methodsFor: 'Shape - State Structure' stamp: ''!
storeTypeCodeForType: typeName

	| arrays |
	arrays := self storeTypeArrays.
	1 to: arrays size do: [:arrayIndex |
		typeName == ((arrays at: arrayIndex) at: 1) ifTrue: [^arrayIndex - 1].
	].
	^0
! !

!SrpMetastate methodsFor: 'Shape - Primary Store Type' stamp: ''!
storeTypeForCode: index
	"The default store type applies to either slots of
	 an object or to the entire object. For most objects
	 the default store type is #State. The default store
	 type is overridden by type declarations in 
	 #srpDataTypeMap for each class of object."

	^(self storeTypeArrays
			at: index + 1)
				at: 1! !

!SrpMetastate methodsFor: 'Behavior behaviors' stamp: ''!
symbol
	"Added for portability with ENVY images. SRP doesn't send this."

	^self behaviorShortName asSymbol! !

!SrpMetastate methodsFor: 'Behavior version number' stamp: ''!
versionNumberArray
	"The version number array is an array containing unsigned integers
	 indicating behavior version information. The contents of the array
	 is defined by the application and can be used to aid class migrations."
	
	^versionNumberArray
! !

!SrpMetastate methodsFor: 'Behavior version number' stamp: ''!
versionNumberArray: anArray
	"The version number array is an array containing unsigned integers
	 indicating behavior version information. The contents of the array
	 is defined by the application and can be used to aid class migrations."
	
	versionNumberArray := anArray.
	instanceShape hasVersionNumberArray: versionNumberArray notNil.
! !

Object subclass: #SrpMetastateShape
	instanceVariableNames: 'integer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpMetastateShape classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 An instance of this class is created for each instance of a SrpMetastate.
 Instances are used to encode/decode simple structural metastate data
 as an integer for space efficiency.
"! !

!SrpMetastateShape classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize
! !

!SrpMetastateShape classSide methodsFor: 'instance creation' stamp: ''!
newLikeClass: aClass configuration: config

	^self new conformTo: aClass configuration: config
! !

!SrpMetastateShape classSide methodsFor: 'Not categorized' stamp: ''!
test
	"	self test	"

	| inst |
	inst := self new.
	inst hasExtendedAttributes: true.
	inst integer = 2r1000000000000 ifFalse: [self error].
	inst primaryStoreTypeCode: 2r1111.
	inst integer = 2r1000011110000 ifFalse: [self error].
	inst primaryStoreTypeCode: 0.
	inst integer = 2r1000000000000 ifFalse: [self error].
	inst isTyped: true.
	inst integer = 2r1000000000001 ifFalse: [self error].
	inst isNamed: true.
	inst integer = 2r1000000000011 ifFalse: [self error].
	inst isIndexed: true.
	inst integer = 2r1000000000111 ifFalse: [self error].
	inst isTyped: false.
	inst integer = 2r1000000000110 ifFalse: [self error].
	inst isNamed: false.
	inst integer = 2r1000000000100 ifFalse: [self error].
	inst isIndexed: false.
	inst integer = 2r1000000000000 ifFalse: [self error].
! !

!SrpMetastateShape methodsFor: 'private' stamp: ''!
clearBit: index
	"Private convenience method. VW didn't implement #clearBit:"

	^integer := integer bitAnd: (1 bitShift: index - 1) bitInvert! !

!SrpMetastateShape methodsFor: 'private' stamp: ''!
debugPrintOn: aStream

	self printOn: aStream.
	aStream cr; tab; nextPutAll: (integer printStringRadix: 2 padTo: 10) reverse.
! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
hasExtendedAttributes
	"Bit 13"

	^self integerAllMask: 2r1000000000000! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
hasExtendedAttributes: aBool
	"Bit 13"

	aBool
		ifTrue: [self setBit: 13]
		ifFalse: [self clearBit: 13].! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
hasVersionNumberArray
	"A version number array is application defined data stored with
	 a metastate. Answer boolean whether a version number array
	 has been set. Bit 14"

	^self integerAllMask: 2r10000000000000! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
hasVersionNumberArray: aBool
	"Bit 14"

	aBool
		ifTrue: [self setBit: 14]
		ifFalse: [self clearBit: 14].
! !

!SrpMetastateShape methodsFor: 'private' stamp: ''!
initialize

	integer := 0.
! !

!SrpMetastateShape methodsFor: 'private' stamp: ''!
integer

	^integer
! !

!SrpMetastateShape methodsFor: 'private' stamp: ''!
integer: anInteger

	integer := anInteger
! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
integerAllMask: anInteger
	"VisualSmalltalk seemed to be the only dialect that didn't support
	 #allMask:. But all are likely to support #bitAnd:."

	^anInteger = (integer bitAnd: anInteger)
! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isDirect
	"Direct encoding was removed as of build 2.2 012.
	 This bit is now available for other uses."

	^self integerAllMask: 2r1000! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isDirect: aBool
	"Direct encoding was removed as of build 2.2 012.
	 This bit is now available for other uses."

	aBool
		ifTrue: [self setBit: 4]
		ifFalse: [self clearBit: 4].! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isIndexed

	^self integerAllMask: 2r100! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isIndexed: aBool

	aBool
		ifTrue: [self setBit: 3]
		ifFalse: [self clearBit: 3].! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isNamed

	^self integerAllMask: 2r10! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isNamed: aBool

	aBool
		ifTrue: [self setBit: 2]
		ifFalse: [self clearBit: 2].! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isSelfTyped
	"Answer boolean whether instances are represented
	 by one of the SRP basic data types."

	^(integer bitAnd: 2r111) = 1
! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isStateless
	"In other words, there is no contents to save or load.
	 The class code itself is all that is required to marshal
	 instances."

	^(integer bitAnd: 2r1111) = 0
! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isTyped

	^self integerAllMask: 2r1! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
isTyped: aBool

	aBool
		ifTrue: [self setBit: 1]
		ifFalse: [self clearBit: 1].! !

!SrpMetastateShape methodsFor: 'Primary Store Type' stamp: ''!
primaryStoreType

	^(SrpMetastate storeTypeArrays
			at: self primaryStoreTypeCode + 1)
				at: 1
! !

!SrpMetastateShape methodsFor: 'Primary Store Type' stamp: ''!
primaryStoreTypeCode
	"Bits 5 to 8. Note that bits 9 to 12 are currently unused."

	^(integer bitShift: -4) bitAnd: 2r1111.
! !

!SrpMetastateShape methodsFor: 'Primary Store Type' stamp: ''!
primaryStoreTypeCode: anInteger
	"Bits 5 to 8. Note that bits 9 to 12 are currently unused."

	anInteger > 2r1111 ifTrue: [self error].
	self integer: (
		(anInteger bitShift: 4) + (integer bitAnd: 2r11110000 bitInvert)
	)



		
	! !

!SrpMetastateShape methodsFor: 'State Structure' stamp: ''!
printOn: aStream

	self isSelfTyped ifTrue: [
		aStream nextPutAll: '['.
		aStream nextPutAll: self primaryStoreType.
		aStream nextPutAll: '] '.
	] ifFalse: [
		self isIndexed ifTrue: [
			aStream nextPutAll: 'Indexed-'.
			aStream nextPutAll: self primaryStoreType.
		].
		self isNamed ifTrue: [aStream nextPutAll: 'Named '].
	].
	self hasExtendedAttributes
		ifTrue: [aStream nextPutAll: ' hasExtendedAttributes'].
! !

!SrpMetastateShape methodsFor: 'private' stamp: ''!
setBit: index
	"Private convenience method. VW didn't implement #setBit:"

	^integer := integer bitOr: (1 bitShift: index - 1)! !

Object subclass: #SrpMetastateTable
	instanceVariableNames: 'dictionary index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpMetastateTable classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 Instances are effectively a collection of metastates that are maintained
 in a way that provides for quick access. Instances can also be saved
 and referenced by data to reduce the cost of storing and loading
 metastates with each traversal (much like an XML DTD file). While
 this can be more efficient, it means the data can't be loaded unless
 the metastate table is available for loading.
"! !

!SrpMetastateTable classSide methodsFor: 'instance creation' stamp: ''!
new

	^self basicNew initialize

! !

!SrpMetastateTable methodsFor: 'accessors' stamp: ''!
addAllMetastates: aCollection

	aCollection do: [:ea | self atCode: ea code putMetastate: ea copy ].
! !

!SrpMetastateTable methodsFor: 'accessors' stamp: ''!
addMetastate: aMetastate

	| existing |
	aMetastate code isNil 
		ifTrue: [aMetastate code: self nextAvailableCode]
		ifFalse: [
			existing := self metastateForCode: aMetastate code.
			existing isNil 
				ifFalse: [ dictionary removeKey: (SrpPortal shortNameForBehavior: existing) ].
		].
	existing := dictionary 
		at: (SrpPortal shortNameForBehavior: aMetastate)
		ifAbsent: [].
	existing isNil ifFalse: [ index at: existing code put: nil ].
	dictionary 
		at: (SrpPortal shortNameForBehavior: aMetastate)
		put: aMetastate.
	(aMetastate code - index size) timesRepeat: [ index add: nil ].
	index at: aMetastate code put: aMetastate.
	^aMetastate! !

!SrpMetastateTable methodsFor: 'accessors' stamp: ''!
allMetastates

	^index reject: [:ea | ea isNil ].
! !

!SrpMetastateTable methodsFor: 'private' stamp: ''!
atCode: integer putMetastate: aMetastate

	(integer - index size) timesRepeat: [ index add: nil ].
	aMetastate isNil 
		ifTrue: [index at: integer put: nil]
		ifFalse: [self addMetastate: aMetastate].
! !

!SrpMetastateTable methodsFor: 'private' stamp: ''!
initialize
	"Preallocate some slots for performance."
	dictionary := Dictionary new: 63.
	index := OrderedCollection new: 127.! !

!SrpMetastateTable methodsFor: 'accessors' stamp: ''!
metastateForClass: aClass

	^dictionary 
		at: (SrpPortal shortNameForBehavior: aClass)
		ifAbsent: []
! !

!SrpMetastateTable methodsFor: 'accessors' stamp: ''!
metastateForCode: code
	
	index size < code ifTrue: [^nil].
	^index at: code! !

!SrpMetastateTable methodsFor: 'private' stamp: ''!
nextAvailableCode
	"Codes 1-31 are reserved codes for marshal directive codes
	 for typed instance variables."

	32 to: index size do: [:i |
		(index at: i) == nil ifTrue: [ ^i ].
	].
	index addLast: nil. 
	^index size.


! !

!SrpMetastateTable methodsFor: 'private' stamp: ''!
printOn: aStream

	| record |
	super printOn: aStream.
	aStream nextPutAll: ' ('.
	1 to: index size do: [:indexPos |
		(record := index at: indexPos) isNil ifFalse: [
			aStream cr; tab.
			record code printOn: aStream.
			aStream nextPut: $:.
			record name printOn: aStream. 
			aStream space.
		].
	].
	aStream cr; nextPut: $).! !

!SrpMetastateTable methodsFor: 'accessors' stamp: ''!
removeMetastate: aMetastate

	index size < aMetastate code ifTrue: [^nil].
	index at: aMetastate code put: nil.
	^aMetastate initialize
! !

Object subclass: #SrpMetastateTableReference
	instanceVariableNames: 'name location accessCode configuration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpMetastateTableReference classSide methodsFor: 'utility' stamp: ''!
codedMetastatesForAll: aCollectionOfClassesClassNamesOrMetastates
	"This prepares a collection of classes, class names, or metastates
	 for storage in a metastate table file (.mst). Codes are assigned in
	 the most efficient manner possible."

	| someMetastates |
	someMetastates := aCollectionOfClassesClassNamesOrMetastates collect: [:classNameClassOrMetastate | 
		(classNameClassOrMetastate isKindOf: SrpConfiguration default metastateClass) 
		ifTrue: [classNameClassOrMetastate] 
		ifFalse: [
			SrpNonMappingConfiguration new newMetastateForClass: (
				(SrpPortal objectIsString: classNameClassOrMetastate)
					ifTrue: [SrpPortal behaviorForName: classNameClassOrMetastate]
					ifFalse: [classNameClassOrMetastate]
			)
		].
	].
	someMetastates := self recodeMetastates: someMetastates.
	^someMetastates
! !

!SrpMetastateTableReference classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 Instances provide basic information used to retrieve a metastate table,
 and are saved along with data.
"! !

!SrpMetastateTableReference classSide methodsFor: 'instance creation' stamp: ''!
newNamed: metastateTableName

	^self new name: metastateTableName copy
! !

!SrpMetastateTableReference classSide methodsFor: 'instance creation' stamp: ''!
newNamed: metastateTableName configuration: aConfiguration

	^self new 
		configuration: aConfiguration; 
		name: metastateTableName copy;
		yourself
! !

!SrpMetastateTableReference classSide methodsFor: 'utility' stamp: ''!
recodeMetastates: metastates
	"This assigns codes in the most space efficient manner possible."

	| table answer |
	table := SrpMetastateTable new.
	metastates do: [:metastate | table addMetastate: metastate ].
	answer := table allMetastates asArray.
	answer size < 128 ifTrue: [
		1 to: answer size do: [:i |
			(answer at: i) code: 127 - answer size + i.
		].
	].
	^answer! !

!SrpMetastateTableReference classSide methodsFor: 'utility' stamp: ''!
smalltalkSourceForClasses: classNames

	"	self smalltalkSourceForClasses: self pmrDefaultClassNames.	"
	"	self smalltalkSourceForClasses: self rtcDefaultClassNames. 	"

	| metastates |
	metastates := self codedMetastatesForAll: classNames.
	^SrpObjectStream smalltalkSourceForObject: metastates.! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
accessCode

	^accessCode
! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
accessCode: aString

	accessCode := aString.
! !

!SrpMetastateTableReference methodsFor: 'Resolution' stamp: ''!
allMetastates

	| resolvedMetastates |
	resolvedMetastates := self configuration resolveMetastateTableReference: self.
	resolvedMetastates isNil ifTrue: [self configuration reportError: 'Data refers to a metastate table that could not be located (', self name, ').'].
	^resolvedMetastates
! !

!SrpMetastateTableReference methodsFor: 'private' stamp: ''!
basicFetchTable
	"This is unfinished code. When finished, the receiver will
	 be used to fetch data through the internet that will be loaded
	 as a metastate table."

	^nil! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
configuration

	^configuration
! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
configuration: aConfiguration

	configuration := aConfiguration.
! !

!SrpMetastateTableReference methodsFor: 'Resolution' stamp: ''!
fetchTable

	| newTable |
	newTable := self basicFetchTable.
	newTable isNil ifFalse:[
		self configuration metastateTables 
			at: self name put: newTable.
	].
	^newTable.
! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
location

	^location
! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
location: aString

	location := aString.
! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
name

	^name
! !

!SrpMetastateTableReference methodsFor: 'Accessors' stamp: ''!
name: aString

	name := aString
! !

!SrpMetastateTableReference methodsFor: 'Resolution' stamp: ''!
wouldNameSuffice
	"Answer boolean whether simply saving and loading
	 the name of the receiver would be sufficient for
	 identifying the correct schema table."

	^self accessCode isNil
		and: [self location isNil]
! !

Object subclass: #SrpObjectStream
	instanceVariableNames: 'stream initialPosition previousPosition metastateTableReference binary config'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpObjectStream classSide methodsFor: 'documentation' stamp: ''!
copyright

	^'Copyright (C) 2000-2004, Paul L. Baumann.  All Rights Reserved.'! !

!SrpObjectStream classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 This is the main class that is used to save and load object to/from an
 SRP encoded data stream. Instances are meant to be treated in a
 manner similar to a standard stream. 
"! !

!SrpObjectStream classSide methodsFor: 'documentation' stamp: ''!
license

^'The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is State Replication Protocol (SRP).

The Initial Developer of the Original Code is Paul Baumann.  
Portions created by Paul Baumann are
Copyright (C) 2000-2004, Paul L. Baumann.

Recent email addresses for SRP and Paul Baumann are:
	srp@avervista.com
	plbaumann@avervista.com
	plbaumann@yahoo.com

The official product website is: 
	srp.avervista.com

Contributor(s):
	Claus Gittinger <cg@exept.de>
	Chris Grindstaff <chrisg@appliedReasoning.com>
	Carsten Haerle
	David Mitchell <dmitchell@appliedreasoning.com>
	Les Tyrrell
	Robert Withers <withers@vnet.net>
	Jon Anderson  <jonkanderson@yahoo.com>
'.! !

!SrpObjectStream classSide methodsFor: 'convenience methods' stamp: ''!
loadObjectFromBinary: aByteArray

	^SrpConfiguration default 
		loadObjectFromBinary: aByteArray! !

!SrpObjectStream classSide methodsFor: 'convenience methods' stamp: ''!
loadObjectFromText: aString

	^SrpConfiguration default 
		loadObjectFromText: (SrpPortal sourceTextFromString: aString)! !

!SrpObjectStream classSide methodsFor: 'instance creation' stamp: ''!
new
	"Answer an instance of the receiver for reading
	 and/or writing object structures on a new stream
	 in binary."

	^self newConfiguration: SrpConfiguration default.! !

!SrpObjectStream classSide methodsFor: 'instance creation' stamp: ''!
newConfiguration: aConfig
	"Answer an instance of the receiver for reading
	 and/or writing object structures on a new stream
	 in binary."

	^super new initializeConfiguration: aConfig! !

!SrpObjectStream classSide methodsFor: 'instance creation' stamp: ''!
onBinaryStream: aStream
	"Answer an instance of the receiver for reading
	 and/or writing object structures on aStream in binary."

	^self new on: aStream! !

!SrpObjectStream classSide methodsFor: 'instance creation' stamp: ''!
onTextStream: aStream
	"Answer an instance of the receiver for reading
	 and/or writing object structures on aStream in text
	 (binary data converted to Base64)."

	^self new text on: aStream! !

!SrpObjectStream classSide methodsFor: 'convenience methods' stamp: ''!
replicate: anObject
	"	self replicate: #(one two three).	"
	^SrpConfiguration default 
		replicate: anObject! !

!SrpObjectStream classSide methodsFor: 'convenience methods' stamp: ''!
saveObjectAsBinary: anObject
	"	self saveObjectAsBinary: #(one two three).	"
	^SrpConfiguration default 
		saveObjectAsBinary: anObject! !

!SrpObjectStream classSide methodsFor: 'convenience methods' stamp: ''!
saveObjectAsText: anObject
	"	self saveObjectAsText: #(one two three).	"
	^SrpConfiguration default 
		saveObjectAsText: anObject.
! !

!SrpObjectStream classSide methodsFor: 'convenience methods' stamp: ''!
smalltalkSourceForObject: anObject
	"Answer a string with Smalltalk source code that can be
	 executed to recreate anObject."

	| string |
	string := SrpBase64Configuration new saveObjectAsText: anObject.
	^(WriteStream on: String new)
		nextPut: $(;
		nextPutAll: #SrpObjectStream;
		space; nextPutAll: #'loadObjectFromText:';
		space; nextPut: $';
		cr; nextPutAll: (SrpPortal stringToSourceText: string);
		nextPutAll: ''')';
		contents! !

!SrpObjectStream methodsFor: 'reading' stamp: ''!
atEnd
	"Answer whether the receiver's input stream reached its end."

	^stream atEnd! !

!SrpObjectStream methodsFor: 'encoding' stamp: ''!
binary
	"A declaration that the stream contains bytes.
	 This is the default setting."

	self setBinary: true.! !

!SrpObjectStream methodsFor: 'initialize-release' stamp: ''!
close
	"Close the stream that the receiver is accessing."

	stream close! !

!SrpObjectStream methodsFor: 'accessing' stamp: ''!
configuration
	"Answer a kind of SrpConfiguration that knows settings
	 to be used for load and/or save traversals."

	^config
! !

!SrpObjectStream methodsFor: 'reading' stamp: ''!
contents
	"Answer all of the objects in the receiver's input stream."

	| cont startPos |
	startPos := stream position.
	stream reset.
	cont := WriteStream on: (Array new: 20).
	[self atEnd] whileFalse: [cont nextPut: self next].
	stream position: startPos.
	^cont contents! !

!SrpObjectStream methodsFor: 'private' stamp: ''!
defaultStream
	"Answer a read/write stream that is preallocated to a certain
	 size (if possible) using the correct type of character collection."

	^SrpPortal 
		newReadWriteStreamSize: config initialStreamBufferSize 
		binary: binary.! !

!SrpObjectStream methodsFor: 'initialize-release' stamp: ''!
flush
	"Write any unwritten information."

	stream flush! !

!SrpObjectStream methodsFor: 'private' stamp: ''!
initializeConfiguration: aConfig
	config := aConfig.
	initialPosition := 0.		"Set zero regardless of dialect root stream position."
	binary := config useBinaryAccess.
	self metastateTableName: config initialMetastateTableName.
	config beingUsedForObjectStream: self.! !

!SrpObjectStream methodsFor: 'encoding' stamp: ''!
isBinary
	"Answer whether the receiver's data is byte or character."

	^binary! !

!SrpObjectStream methodsFor: 'metastate table' stamp: ''!
metastateTableName

	metastateTableReference isNil ifTrue: [^nil].
	^metastateTableReference name! !

!SrpObjectStream methodsFor: 'metastate table' stamp: ''!
metastateTableName: metastateTableName

	| newRef |
	newRef := metastateTableName isNil
		ifTrue: [ nil ]
		ifFalse: [ self configuration newMetastateTableReferenceNamed: metastateTableName ].
	self metastateTableReference: newRef.! !

!SrpObjectStream methodsFor: 'metastate table' stamp: ''!
metastateTableReference

	^metastateTableReference! !

!SrpObjectStream methodsFor: 'metastate table' stamp: ''!
metastateTableReference: aRef

	metastateTableReference := aRef
! !

!SrpObjectStream methodsFor: 'reading' stamp: ''!
next
	^self reader next! !

!SrpObjectStream methodsFor: 'writing' stamp: ''!
nextPut: anObject

	self writer nextPut: anObject.
	^anObject! !

!SrpObjectStream methodsFor: 'writing' stamp: ''!
nextPutAll: aCollection 
	"Put each of the elements of aCollection on
	 the stream. It is far better to simply #nextPut:
	 the entire collection, but this might be useful
	 for something. Answer aCollection."

	aCollection do: [:v | self nextPut: v].
	^aCollection! !

!SrpObjectStream methodsFor: 'initialize-release' stamp: ''!
on: aStream 

	stream := aStream.
	initialPosition := previousPosition := stream position.! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
position
	" Answer the zero-based position of the underlying stream. "

	^stream position - SrpPortableBehavior portal streamBasePosition! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
position: newPos
	" Set the zero based position of the underlying stream. "

	stream position: newPos + SrpPortableBehavior portal streamBasePosition! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
previousByteSize
	" Answer the number of bytes traversed for
	 the last read/write operation. "

	^self position - self previousPosition! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
previousPosition
	" Answer the stream position before the last read/write operation. "

	^previousPosition! !

!SrpObjectStream methodsFor: 'private' stamp: ''!
reader
	" Answer a reader that will read
	one object structure from the receiver's stream. "

	| reader |
	previousPosition := stream position.
	reader := config readerClass newConfiguration: config.
	reader stream: stream encodings: config initialEncodings.
	binary ifFalse: [reader setEncodingType: #Text].
	^reader! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
reset
	self position: initialPosition! !

!SrpObjectStream methodsFor: 'encoding' stamp: ''!
setBinary: bool
	"Private. Declare whether the stream should be
	 considered to contain bytes (true) or characters
	 (false)."

	binary := bool.! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
setToEnd
	" Set the underlying stream to its end. "

	stream setToEnd! !

!SrpObjectStream methodsFor: 'positioning' stamp: ''!
stream
	^stream
! !

!SrpObjectStream methodsFor: 'reading' stamp: ''!
streamContents
	"Answer the bytes that represent objects in the stream."

	^stream contents
! !

!SrpObjectStream methodsFor: 'encoding' stamp: ''!
text
	"A declaration that the stream contains characters
	 instead of bytes. Binary data will be converted 
	 to/from Base64 encoded characters."

	self setBinary: false.! !

!SrpObjectStream methodsFor: 'private' stamp: ''!
writer
	" Answer a writer that will write one object structure
	 on the receiver's stream. "
	
	| writer |
	stream isNil ifTrue: [ stream := self defaultStream ].
	previousPosition := stream position.
	writer := config writerClass newConfiguration: config.
	writer stream: stream encodings: config initialEncodings.
	binary ifFalse: [
		(config wrapAppendedText and: [stream atEnd])
			ifTrue: [writer setEncodingType: #WrappedText]
			ifFalse: [writer setEncodingType: #Text].
	].
	self metastateTableReference isNil 
		ifFalse: [ writer setMetastateTable: self metastateTableReference ].
	^writer! !

Object subclass: #SrpPlaceholder
	instanceVariableNames: 'baseObject memento loaderIndex canDiscard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpPlaceholder classSide methodsFor: 'instance creation' stamp: ''!
baseObject: anObject

	^self new srpBaseObject: anObject! !

!SrpPlaceholder classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 An instance is created for each object that is loaded. This behaves
 as a value holder. Other objects refer to the holder instead of directly
 to a loaded object. This allows the loaded object to be remapped
 while still preserving references made by other loaded objects.

 Placeholders are stripped a little at a time from objects as soon
 as it is known that they are no longer necessary. After all load
 mapping operations have completed (even post-load operations)
 then any remaining placeholders are dereferenced.
"
! !

!SrpPlaceholder methodsFor: 'testing' stamp: ''!
= anObject
	"Answer a Boolean which is true when the receiver and
	 the argument anObject are equivalent, and false
	 otherwise. Subclasses may redefine = to alter the
	 meaning of equivalence. Equivalence implies that two
	 objects are similar in form and content. Objects that
	 are equivalent must answer the same value for the
	 message #hash."

	(self == anObject) 			ifTrue: [^true].
	(baseObject = anObject) 	ifTrue: [^true].
	(memento == anObject) 	ifTrue: [^true].
	^false
! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
doesNotUnderstand: aMessage
	"Forward the message to the baseObject. If the
	 answer is not the base object then answer the
	 answer; else answer yourself."

	| answer |
	answer := baseObject 
		perform: aMessage selector 
		withArguments: aMessage arguments.
	(answer == baseObject) ifFalse: [^answer].! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
hash
	"The hash value of the baseObject must be answered for proper
	 indexing in collections. Remember the placeholder itself is short lived."

	^baseObject hash! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
printOn: aStream
	"Private implementation. Public protocol."

	aStream nextPut: $`.
	baseObject printOn: aStream.
! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpAttribute: varNameString
	"This can be sent by mapping rules. It is a way to access named
	 variable slots regardless of whether accessor methods are defined."

	^baseObject srpAttribute: varNameString! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpAttribute: varNameString put: anObject
	"This can be sent by mapping rules. It is a way to access named
	 variable slots regardless of whether accessor methods are defined."

	^baseObject srpAttribute: varNameString put: anObject! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpBaseObject
	"Private protocol."

	^baseObject! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpBaseObject: anObject
	"Private protocol."

	memento := baseObject := anObject
! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpCanDiscard
	"Private. canDiscard can be either nil, true, or false."

	^canDiscard == true! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpCanDiscard: bool
	"Private protocol."

	canDiscard := bool! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpClass
	"Answer the class of the receiver. This can be sent by
	 mapping rules. It is preferable to #class because it
	 works with special loader constructs."

	^baseObject srpClass

! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpIndexableSize
	"Answer the basic size of the receiver. This can be sent by
	 mapping rules. It is preferable to #basicSize because it
	 works with special loader constructs."

	^baseObject srpIndexableSize! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpIndexableValueAt: index
	"Answer the object at the variable index position of the receiver.
	 This can be sent by mapping rules. It is preferable to #basicAt: because
	 it works with special loader constructs."

	^baseObject srpIndexableValueAt: index

! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpIndexableValueAt: index put: value
	"Store the Object argument value at the variable index position 
	 in the receiver. This can be sent by mapping rules. It is preferable to
	 #basicAt:put: because it works with special loader constructs."

	^baseObject srpIndexableValueAt: index put: value

! !

!SrpPlaceholder methodsFor: 'testing' stamp: ''!
srpIsPlaceholder
	"Objects are temporarily wrapped with a placeholder while
	 loading. This can be sent in load mapping rules or #srpLoaded
	 to test if the receiver is a placeholder."

	^true
! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpLoadPostMap: aContext
	"Answer an object that is to represent the receiver when
	 loaded. Keep in mind that slots of the receiver could 
	 still contain placeholders around an object that may
	 yet be mapped. Also be aware that when an object
	 references the receiver, and the receiver also references
	 that object, that the slots for that object may not be fully
	 loaded yet."

	"Because this is a placeholder, the baseObject can be changed
	 locally, and all references to the placeholder will not be affected."

	baseObject := baseObject srpLoadPostMap: aContext.

	"Answer the receiver because you don't want to lose the
	 placeholder behavior."
! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpLoadPreMap: aContext
	"Answer an object that is to represent the receiver when
	 loaded. Keep in mind that slots of the receiver could 
	 still contain placeholders around an object that may
	 yet be mapped. Also be aware that when an object
	 references the receiver, and the receiver also references
	 that object, that the slots for that object may not be fully
	 loaded yet. If unsure of loader placeholders then declare an
	 #srpPostLoadAction instead."

	baseObject := baseObject srpLoadPreMap: aContext.

	"Answer the receiver because you don't want to lose the
	 placeholder behavior."! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpLoaderIndex
	"Private protocol."

	^loaderIndex! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpLoaderIndex: loaderBackpointerIndexPos
	"Private protocol."

	loaderIndex := loaderBackpointerIndexPos! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpNamedValueAt: index
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt: because it
	 works with special loader constructs."

	^baseObject srpNamedValueAt: index

! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpNamedValueAt: index put: anObject
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt:put: because
	 it works with special loader constructs."

	^baseObject srpNamedValueAt: index put: anObject

! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpPostLoadActionForLoader: aMarshaler
	"Answer a command (directed message or block) that
	 will answer an object to take the place of the
	 receiver when sent #value. The command will be
	 sent #value after the root object has been completely
	 loaded. Answer nil if no further processing is required."

	| actionOrNil |
	actionOrNil := baseObject srpPostLoadActionForLoader: aMarshaler.
	actionOrNil isNil ifTrue: [^nil].
	^SrpActionItem
		forPlaceholder: self 
		action: actionOrNil
! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpStripPlaceholders
	"Private protocol. This is a recursive method that is used to
	 cleanup all the placeholders used for correctly resolving changing
	 object pointers when loading objects. This should only be
	 done after an entire persistent object has been loaded. To
	 do otherwise may leave some unmutated objects."

	^self srpStripPlaceholders: SrpPortal identitySetClass new
! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpStripPlaceholders: recursionSet
	"Private protocol. This is a recursive method that is used to cleanup
	 all the placeholders used for correctly resolving changing object pointers
	 when loading objects."

	| item newItem |
	(recursionSet includes: self) ifTrue: [^baseObject].
	recursionSet add: self.
	"Strip from any fixed variables -- if there are any."
	1  to: baseObject srpClass instSize do: [:instVarIndex |
		item := baseObject srpNamedValueAt: instVarIndex.
		(item srpIsPlaceholder or: [self srpWasMutated]) ifTrue: [
			newItem := item srpStripPlaceholders: recursionSet.
			item == newItem		"Avoid store in read-only"
				ifFalse: [baseObject srpNamedValueAt: instVarIndex put: newItem].
		].
	].
	"Strip from any indexed (pointer only) variables -- if there are any."
	(SrpPortal behaviorIsPointers: baseObject srpClass) ifTrue: [
		1 to: baseObject srpIndexableSize do: [:indexedVarIndex |
			item := baseObject srpIndexableValueAt: indexedVarIndex.
			(item srpIsPlaceholder or: [self srpWasMutated]) ifTrue: [
				newItem := item srpStripPlaceholders: recursionSet.
				item == newItem		"Avoid store in read-only"
					ifFalse: [baseObject srpIndexableValueAt: indexedVarIndex put: newItem].
			].
		].
	].
	^baseObject! !

!SrpPlaceholder methodsFor: 'private' stamp: ''!
srpWasMutated
	"Private protocol."

	^(baseObject == memento) not
! !

!SrpPlaceholder methodsFor: 'accessors' stamp: ''!
srpWithoutPlaceholder
	"This can be sent by mapping rules and in #srpLoaded 
	 when it is necessary to ensure an object is not wrapped
	 by a placeholder when loading. If you ever end up with
	 a placeholder after loading and mapping is completely
	 finished then this wasn't sent when it should have been. "

	^baseObject
! !

Object subclass: #SrpPortableBehavior
	instanceVariableNames: 'lineDelimiter exchangeRules'
	classVariableNames: 'SrpPortal'
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehavior classSide methodsFor: 'Behavior selection' stamp: ''!
initialize
	"	self initialize	"

	SrpPortal := nil.
	self portal.! !

!SrpPortableBehavior classSide methodsFor: 'Behavior selection' stamp: ''!
mightApplyForImage

	^false
! !

!SrpPortableBehavior classSide methodsFor: 'Behavior selection' stamp: ''!
newCorrectBehavior

	^self selectCorrectBehavior new
	
! !

!SrpPortableBehavior classSide methodsFor: 'Behavior selection' stamp: ''!
portal
	"Sent during initialization and as an accessor when loading
	 code that may not have been initialized in the proper sequence."

	"	self portal.	"

	| was |
	(was := SrpPortal) isNil 
		ifTrue: [SrpPortal := self newCorrectBehavior].
	^was! !

!SrpPortableBehavior classSide methodsFor: 'Behavior selection' stamp: ''!
portal: obj

	| was |
	was := SrpPortal.
	SrpPortal := obj.
	^was
! !

!SrpPortableBehavior classSide methodsFor: 'Behavior selection' stamp: 'sd 7/19/2006 16:54'!
selectCorrectBehavior
	"Discover the dialect being used. This code must be written so as to execute on all dialects."
	"	self selectCorrectBehavior.	"

	| candidates |
	candidates := (self respondsTo: #withAllSubclasses)
	ifTrue: [
		"[ObjectStudio] #withAllSubclasses returns names of subclasses."
		"[Squeak] #withAllSubclasses returns a Set."
		self withAllSubclasses asArray collect: [:ea | 
			(ea respondsTo: #asClass) 
				ifTrue: [ea asClass] 
				ifFalse: [ea]
		]
	] ifFalse: [ 
		"[GemStone/S] doesn't understand #withAllSubclasses."
		(Object perform: #'_resolveLiteralVar:' with: #ClassOrganizer) 
			value new allSubclassesOf: self.
	].
	candidates := candidates select: [:ea | ea mightApplyForImage ].
	candidates isEmpty ifTrue: [ ^self ].
	candidates size = 1 ifTrue: [ ^candidates first ].
	"Answer the last instead of the first so that we give preference to a subclass."
	^candidates last.! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
base256ToBase64Table
	"This is used for converting to Base64."

	" *The Array below has been populated with:
	SrpBase64 validCharacters asArray collect: [:each | each asInteger ]"

	^#(
		65 66 67 68 69 70 71 72 73 74 75 76 
		77 78 79 80 81 82 83 84 85 86 87 88 
		89 90 97 98 99 100 101 102 103 104 
		105 106 107 108 109 110 111 112 113 114 
		115 116 117 118 119 120 121 122 48 49 
		50 51 52 53 54 55 56 57 43 47
	)
! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
base64ToBase256Table
	"This is for converting from Base64."

	" *Use to populate the ByteArray below*
		| bytes answer |
		bytes := SrpBase64 validCharacters asByteArray.
		answer := (16r00 to: 16rFF) collect: [:byte | (bytes indexOf: byte) - 1 ].
		answer at: $= value + 1 put: 0.
		^answer
	"

	^#(
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 62 -1 -1 -1 63 52 53 54 55 
		56 57 58 59 60 61 -1 -1 -1 0 -1 -1 -1 0 1 2 3 4 
		5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 
		22 23 24 25 -1 -1 -1 -1 -1 -1 26 27 28 29 30 
		31 32 33 34 35 36 37 38 39 40 41 42 43 44 
		45 46 47 48 49 50 51 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 
		-1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1
	)
! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior basicNew: newSize
	"ObjectStudio doesn't implement #basicNew:."

	^aBehavior basicNew: newSize
! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior bindingFor: key

	^aBehavior bindingFor: key
! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorForName: className
	"Answer a behavior resident to the object space named className."

	"	self behaviorForName: #String.	"

	^self rootNamespace
		at: className asSymbol
		ifAbsent: [].! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorForNamespacePath: anArray
	"	(SrpPortal behaviorForNamespacePath: #(Smalltalk Core String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(Core String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(Absent String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(Array String)) == String.	"

	| namespace binding |
	namespace := self rootNamespace.
	1 to: anArray size - 1 do: [:i |
		binding := self namespace: namespace bindingFor: (anArray at: i).
		binding notNil 
			ifTrue: [namespace := binding value].
	].
	binding := self namespace: namespace bindingFor: anArray last.
	^binding isNil 
		ifTrue: [self rootNamespace at: anArray last ifAbsent: []] 
		ifFalse: [binding value].! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorHasIndexableValues: aBehavior
	"Behaviors in ObjectStudio don't know if they have variable length instances."

	^aBehavior isVariable! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorIsBytes: aBehavior

	^aBehavior isBytes! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior isMetaclass! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorIsPointers: aBehavior
	"Behaviors in ObjectStudio don't know if they have pointers."

	^aBehavior isPointers! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
behaviorWithAllSubclasses: aBehavior do: oneArgBlock
	"Answer a collection containing aBehavior and all subclasses of aBehavior."

	"	self portal behaviorWithAllSubclasses: Collection do: [:ea | ].	"

	oneArgBlock value: aBehavior.
	aBehavior allSubclassesDo: oneArgBlock.! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
byteArrayAsString: aByteArray

	^aByteArray asString! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  'ByteArray withAll: (65 to: 100)' works for VW but not VA."

	^(WriteStream on: ByteArray new)
		nextPutAll: receiver;
		contents.
! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
characterCollection: characterCollection at: index putCharacter: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW, Squeak, and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer instead of character, but that would be less
	 portable with other Smalltalk dialects."

	^characterCollection basicAt: index put: aCharacter
! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
characterToUnsigned: aCharacter

	^aCharacter asInteger! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
dateClassNewDay: day monthIndex: month year: year
	"OK for VA, Dol."

	^Date
		newDay: day 
		monthIndex: month 
		year: year! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
dateDayOfMonth: aDate 
	^aDate dayOfMonth! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
dateMonth: aDate 
	^aDate monthIndex! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
dateTimeClass

	^self reportError: 'Unknown or unsupported behavior'.! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
dateTimeClassDate: aDate time: aTime

	^self dateTimeClass
		date: aDate
		time: aTime! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
decimalClassFromString: aString

	^aString asDecimal
! !

!SrpPortableBehavior methodsFor: 'Dialect specific hooks' stamp: ''!
defaultExchangeRules
	"Answer a collection of mapping rule that apply to this Smalltalk dialect."

	^Array new! !

!SrpPortableBehavior methodsFor: 'Message related' stamp: ''!
directedMessageClass
	"Not all Smalltalk dialects have an DirectedMessage class."

	^self behaviorForName: #DirectedMessage.
! !

!SrpPortableBehavior methodsFor: 'Message related' stamp: ''!
directedMessageClassSelector: aSymbol arguments: anArray receiver: anObject
	"Not all dialects have a DirectedMessage class or use the same
	 class creation methods."

	^self directedMessageClass new
		selector: aSymbol;
		arguments: anArray;
		receiver: anObject;
		yourself
! !

!SrpPortableBehavior methodsFor: 'Dialect specific hooks' stamp: ''!
exchangeRules
	"Answer a collection of exchange rules that apply to this Smalltalk dialect."
	exchangeRules isNil 
		ifTrue: [exchangeRules := self defaultExchangeRules].
	^exchangeRules
! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
fixupLiteralArray: anArray
	"ObjectStudio doesn't parse literal arrays according to
	 normal Smalltalk convention. This is used to compensate."

	^anArray! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
floatAsNumeratorAndDenominator: aFloat
	"(VA, GS/S, ObjectStudio) use #asFraction, (VW, VSE) use #asRational"

	| temp |
	temp := aFloat asFraction.
	^Array with: temp numerator with: temp denominator! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
floatScale: aFloat
	"Answer an integer indicating the scale for a float.
	 For example 0.000065 scale has a scale of -5 which
	 means that the float can be expressed as 6.5e-5."

	| scale x |
	aFloat = 0.0 ifTrue: [^0].
	(x := aFloat abs) < 1 ifTrue: [
		scale := 1.
		[x < 10.0] whileTrue: [
			x := x * 10.
			scale := scale - 1.
		].
	] ifFalse: [
		scale := 0.
		[x < 10.0] whileFalse: [
			x := x / 10.
			scale := scale + 1.
		].
	]. 
	^scale
! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
identitySetClass
	"Not all Smalltalk dialects have an IdentitySet class.
	 Answer a normal Set class if identity isn't possible."

	|class|
	class := self behaviorForName: #IdentitySet.
	class == nil ifTrue: [^Set].
	^class.
! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
isNamespaceImage

	^false! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
keyedCollection: keyedCollection associationsDo: oneArgBlock
	"ObjectStudio #associationsDo: expects two arguments."
	"VisualAge doesn't implement #associations."
	"VisualWorks doesn't allow #addAll: for populating."

	keyedCollection associationsDo: oneArgBlock! !

!SrpPortableBehavior methodsFor: 'Message related' stamp: ''!
messageClass
	"Not all Smalltalk dialects have an Message class."

	^self behaviorForName: #Message.
! !

!SrpPortableBehavior methodsFor: 'Message related' stamp: ''!
messageClassSelector: aSymbol arguments: anArray
	"Not all dialects have a Message class or use the same
	 class creation methods."

	^self messageClass new
		selector: aSymbol;
		arguments: anArray;
		yourself
! !

!SrpPortableBehavior methodsFor: 'Environmental awareness' stamp: ''!
millisecondsToRun: aBlock
	"GemStone doesn't implement Time>>millisecondsToRun:."

	^Time millisecondsToRun: aBlock! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
namespace: ns bindingFor: key
	"A namespace should be considered anything with visibility
	 to variable bindings. It could be the Smalltalk dictionary, a
	 class, a pool dictionary, or a real dialect namespace. Answer
	 nil if ns isn't considered a namespace, or if the key wasn't
	 found in the namespace."

	^ns bindingFor: key! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
namespace: ns classAt: classSymbol

	| aClass |
	aClass := self namespace: ns resolve: classSymbol.
	(aClass isNil or: [aClass isBehavior not])
		ifTrue: [^nil].
	^aClass! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
namespace: ns ensureBindingFor: key

	| binding |
	binding := self namespace: ns bindingFor: key.
	binding isNil ifFalse: [^binding].
	^self namespace: ns newBindingFor: key.! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
namespace: ns newBindingFor: key

	^ns declareVariable: key.
! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
namespace: ns resolve: key
	"There are some reservations about the design here."

	| anAssociation |
	anAssociation := self namespace: ns bindingFor: key.
	^anAssociation isNil ifFalse: [anAssociation value]! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
namespacePathForBehavior: aBehavior
	"The code that goes here depends on the language and
	 dialect being used. Answer an array of symbols that another
	 objectspace can use to retrieve behavior later."

	aBehavior class == SrpMetastate ifTrue: [^aBehavior behaviorName].
	^Array with: (self shortNameForBehavior: aBehavior)! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
newBase256ToBase64Table
	"This is used for converting to Base64."

	^SrpBase64 validCharacters asArray collect: [:each | self characterToUnsigned: each ].
! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
newBase64ToBase256Table
	"This is for converting from Base64."

	| bytes answer |
	bytes := SrpBase64 validCharacters asByteArray.
	answer := (16r00 to: 16rFF) collect: [:byte | (bytes indexOf: byte) - 1 ].
	answer at: (self characterToUnsigned: $=) + 1 put: 0.
	^answer
! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
newReadOverwriteStreamWith: aCollection
	"We want a stream that will start at the beginning
	 and that will leave data alone that it does not write."

	^(ReadWriteStream with: aCollection) reset; yourself! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
newReadWriteStreamSize: initialSize binary: binary
	"Creates a stream preallocated with the appropriate
	 type of character collection. This was necessary
	 because some dialects need special treatment to
	 prevent preallocated contents from being considered
	 part of the written data."

	| aCollection |
	 aCollection := binary 
		ifTrue: [ ByteArray new: initialSize ] 
		ifFalse: [ String new: initialSize ].
	^(ReadWriteStream on: aCollection) reset; yourself! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
objectIsString: anObject
	"Dolphin and Squeak don't support #isString."
	^anObject isString! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
optionalWeakKeyIdentityDictionaryClass

	^IdentityDictionary! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
platformLineDelimiter
	"Writing #cr to a stream is a simple portable way to
	 get the local line delimiter convention. Cache the
	 result for performance. The cache must be flushed
	 if the same image is used on another platform."

	lineDelimiter isNil 
		ifTrue: [ lineDelimiter := (WriteStream on: String new) cr; contents ].
	^lineDelimiter
! !

!SrpPortableBehavior methodsFor: 'Dialect specific hooks' stamp: ''!
reportError: errorString
	"Most errors raised by SRP are reported through this method."

	self error: errorString! !

!SrpPortableBehavior methodsFor: 'Namespace related' stamp: ''!
rootNamespace
	"ObjectStudio and GemStone don't define Smalltalk.
	 ObjectStudio will just declare this as a new Global.
	 GemStone will show a compile error that doesn't cause a problem."

	^Smalltalk
! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
secondsFromTime: aTime
	"Answer the number of seconds since midnight for aTime.
	 It is OK to answer a non-integer value for fractional seconds.
	 This generic behavior should be overridden for optimization
	 for a particular dialect."

	^(aTime respondsTo: #asMilliseconds)
		ifTrue: [aTime asMilliseconds / 1000]
		ifFalse: [aTime asSeconds]

	"Note that VA 6.0 adds support for #asMilliseconds so this default
	 behavior works for all VA versions."
! !

!SrpPortableBehavior methodsFor: 'Environmental awareness' stamp: ''!
secondsFromUtc

	^3600 * self timeZonesFromGMT! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
sequenceableCollection: receiver equals: aCollection
	"Answer true if the receiver and aCollection contain the same elements
	 in the same order."

	receiver == aCollection ifTrue: [^true].
	receiver size = aCollection size ifFalse: [^false].
	1 to: receiver size do: [ :index |
		(receiver at: index) = (aCollection at: index)
			ifFalse: [^false ]].
	^true

	"VW doesn't support #equals:. VA does."! !

!SrpPortableBehavior methodsFor: 'Collection related' stamp: ''!
sequenceableCollection: sequenceableCollection indexOfSubCollection: aSubcollection startingAt: anIndex
	"VisualSmalltalk doesn't implement #indexOfSubCollection:startingAt:, but most others do."

	^sequenceableCollection indexOfSubCollection: aSubcollection startingAt: anIndex
! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
shortNameForBehavior: aClass
	"This should answer #Metaclass instead of the
	 class symbol that Metaclass is an instance for."

	aClass class == SrpMetastate ifTrue: [^aClass behaviorShortName].
	^(self behaviorIsMetaclass: aClass)
		ifTrue: [ aClass class name ]
		ifFalse: [ aClass name ]
! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^(self behaviorIsMetaclass: aClass)
		ifTrue: [ aClass primaryInstance name ]
		ifFalse: [ aClass name ]! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
sourceTextFromString: aString

	^aString! !

!SrpPortableBehavior methodsFor: 'Dialect specific hooks' stamp: ''!
srpLoadPreMap: aContext
	"It is strange to save an instance of a portable behavior,
	 but it could happen for SRP tests. This maps to an
	 instance for the loading environment."

	^self class portal! !

!SrpPortableBehavior methodsFor: 'Dialect specific hooks' stamp: ''!
srpSavePreMap: aContext
	"It is strange to save an instance of a portable behavior,
	 but it could happen for SRP tests. This maps to an
	 instance of the abstract superclass when saving."

	self class name == #SrpPortableBehavior
		ifTrue: [^self].
	^SrpPortableBehavior new! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
streamBasePosition
	"ST/X and GS/S use 1 based streams instead of 0 based
	 like all other Smalltalk dialects. Source comments in
	 ST/X indicated that this may change in a future release."
	^0! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
stringAsBinaryFileStream: filename
	^filename asFilename readWriteStream binary! !

!SrpPortableBehavior methodsFor: 'Behavior related' stamp: ''!
stringAsInstVarName: aString
	"Convert aString so that it can be found in #allInstVarNames.
	 GS/S expects symbols, other dialects use strings."

	^aString! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
stringAsTextFileStream: filename
	^filename asFilename readWriteStream! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
stringAsTwoByteString: aString
	"Convert aString so that it can contain
	 characters with values up to 16rFFFF (65535)."

	^aString asTwoByteString! !

!SrpPortableBehavior methodsFor: 'Stream related' stamp: ''!
stringExistsAsFile: filename
	^filename asFilename exists! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
stringToSourceText: aString

	^aString! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
symbolClassIntern: aString

	^aString asSymbol
! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
timeClassFromSeconds: seconds
	"Seconds could very well be a non-integer value."
	"OK for VA, VW, Dol. Not OK for GemStone."

	^Time fromSeconds: seconds! !

!SrpPortableBehavior methodsFor: 'String related' stamp: ''!
twoByteStringClass
	"Not all Smalltalk dialects have support for double byte strings.
	 Answer nil if support isn't available."

	^nil! !

!SrpPortableBehavior methodsFor: 'Magnitude related' stamp: ''!
unsignedToCharacter: anUnsignedInteger

	^anUnsignedInteger asCharacter! !

!SrpPortableBehavior methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#Unknown! !

SrpPortableBehavior subclass: #SrpPortableBehaviorDol
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorDol classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation		"

	The finishing touches to the first Dolphin port were completed
	at Camp Smalltalk 1. Dolphin has proven to be a very easy
	Smalltalk dialect to port to and it provides a very appealing
	development environment.

	SRP 3.0 was ported to Dolphin 5.1.

"! !

!SrpPortableBehaviorDol classSide methodsFor: 'As yet unclassified' stamp: ''!
mightApplyForImage
	"Should find better test than this."

	Smalltalk isNil ifTrue: [^false].
	Smalltalk at: #AttributeDescriptor ifAbsent: [^false].
	^true! !

!SrpPortableBehaviorDol methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior isMeta! !

!SrpPortableBehaviorDol methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  'ByteArray withAll: (65 to: 100)' works for VW but not VA."

	^receiver asByteArray! !

!SrpPortableBehaviorDol methodsFor: 'String related' stamp: ''!
characterCollection: characterCollection at: index putCharacter: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW, Squeak, and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer instead of character, but that would be less
	 portable with other Smalltalk dialects."

	^characterCollection basicAt: index put: aCharacter asInteger
! !

!SrpPortableBehaviorDol methodsFor: 'Magnitude related' stamp: ''!
dateTimeClass
	"Should use something faster than this later."

	^self behaviorForName: #TimeStamp! !

!SrpPortableBehaviorDol methodsFor: 'Environmental awareness' stamp: ''!
floatAsNumeratorAndDenominator: aFloat
	"Answer a Fraction representing aFloat."
	"	self new floatAsNumeratorAndDenominator: 1.2.	"

	| factor x y fract |
	aFloat = 0.0 ifTrue: [^Array with: 0 with: 1].
	x := y := aFloat truncated.
	x == 0 ifTrue: [y := 1].
	aFloat = x ifTrue: [^Array with: x with: 1].
	factor := 10 raisedToInteger: (((self floatScale: aFloat) abs + 14) min: 307).
	fract := x + ((((aFloat rem: y) * factor) asInteger) / factor).
	^Array with: fract numerator with: fract denominator! !

!SrpPortableBehaviorDol methodsFor: 'Namespace related' stamp: ''!
namespace: ns classAt: classSymbol

	| aClass |
	aClass := self namespace: ns resolve: classSymbol.
	(aClass isNil or: [(aClass isKindOf: Behavior) not])
		ifTrue: [^nil].
	^aClass! !

!SrpPortableBehaviorDol methodsFor: 'String related' stamp: ''!
objectIsString: anObject
	"Dolphin, GS/S, and Squeak don't support #isString."
	^anObject isKindOf: String! !

!SrpPortableBehaviorDol methodsFor: 'Magnitude related' stamp: ''!
secondsFromTime: aTime
	"Answer the number of seconds since midnight for aTime.
	 It is OK to answer a non-integer value for fractional seconds."

	^aTime asMilliseconds / 1000! !

!SrpPortableBehaviorDol methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^(self behaviorIsMetaclass: aClass)
		ifTrue: [ aClass instanceClass name ]
		ifFalse: [ aClass name ]! !

!SrpPortableBehaviorDol methodsFor: 'Stream related' stamp: ''!
stringAsBinaryFileStream: filename

	^(self behaviorForName: #FileStream)
		 open: filename mode: #open check: false text: false! !

!SrpPortableBehaviorDol methodsFor: 'Stream related' stamp: ''!
stringAsTextFileStream: filename

	^(self behaviorForName: #FileStream)
		 open: filename mode: #open check: false text: true! !

!SrpPortableBehaviorDol methodsFor: 'Stream related' stamp: ''!
stringExistsAsFile: filename

	^(self behaviorForName: #File) exists: filename ! !

!SrpPortableBehaviorDol methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#Dolphin! !

SrpPortableBehavior subclass: #SrpPortableBehaviorObjectStudio
	instanceVariableNames: 'base256ToBase64Table base64ToBase256Table'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorObjectStudio classSide methodsFor: 'documentation' stamp: ''!
documentation "

	ObjectStudio 6.9.0

	This dialect doesn't:
		...have a Transcript global--but one is declared by SUnit.
		...allow temporary variables to be declared anywhere except the method header.
		...support class instance variables.
		...have a Smalltalk global.
		...support #variableSubclass:* class definitions.

	ObjectStudio is by far the most difficult Smalltalk dialect to port to so far.

	This port is only 90% complete as of SRP release 3.0.

"! !

!SrpPortableBehaviorObjectStudio classSide methodsFor: 'Not categorized' stamp: ''!
mightApplyForImage

	| sys |
	(Object respondsTo: #environment) ifFalse: [^false].
	(Object environment respondsTo: #at:ifAbsent:) ifFalse: [^false].
	sys := Object environment at: #System ifAbsent: [^false].
	(sys name isKindOf: String) ifFalse: [^false].
	^sys name = 'ObjectStudio'! !

!SrpPortableBehaviorObjectStudio methodsFor: 'String related' stamp: ''!
base256ToBase64Table
	"Overridden because ObjectStudio doesn't properly parse
	 negative value literals like #(-1)."

	base256ToBase64Table isNil 
		ifTrue: [base256ToBase64Table := self newBase256ToBase64Table].
	^base256ToBase64Table

! !

!SrpPortableBehaviorObjectStudio methodsFor: 'String related' stamp: ''!
base64ToBase256Table
	"Overridden because ObjectStudio doesn't properly parse
	 negative value literals like #(-1)."

	base64ToBase256Table isNil 
		ifTrue: [base64ToBase256Table := self newBase64ToBase256Table].
	^base64ToBase256Table

! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior basicNew: newSize
	"ObjectStudio doesn't implement #basicNew:."

	^aBehavior new: newSize
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior bindingFor: key

	^aBehavior classPool
		associationAt: key asSymbol
		ifAbsent: [].! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behaviorHasIndexableValues: aBehavior
	"Behaviors in ObjectStudio don't know if they have variable length instances."

	"ObjectStudio support passing in either a symbol or a class.
	 It would be far more efficient to pass in a class, but using a symbol
	 allows for the code to be loaded in dialects that don't implement the
	 class. You'd probably want to optimize this for production use." 

	^aBehavior inheritsFrom: #IndexedCollection! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behaviorIsBytes: aBehavior
	"There didn't seem to be a meta way of doing this so this
	 relies on hard-coded class names."

	^(self behaviorIsPointers: aBehavior) not! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior type == #class! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behaviorIsPointers: aBehavior
	"There didn't seem to be a meta way of doing this so this
	 hard-codes specific class names."

	(aBehavior isKindOf: SrpMetastate) ifTrue: [^aBehavior isPointers].
	^(#(String Symbol) includes: aBehavior name) not! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
behaviorWithAllSubclasses: aBehavior do: oneArgBlock
	"Answer a collection containing aBehavior and all subclasses of aBehavior."

	"	self portal behaviorWithAllSubclasses: Collection do: [:ea | ].	"

	oneArgBlock value: aBehavior.
	aBehavior allSubclasses do: [:aSymbol | oneArgBlock value: aSymbol asClass ].! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  'ByteArray withAll: (65 to: 100)' works for VW but not VA."

	^receiver asByteArray
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'String related' stamp: ''!
characterCollection: characterCollection at: index putCharacter: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW, Squeak, and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer instead of character, but that would be less
	 portable with other Smalltalk dialects."

	characterCollection at: index put: aCharacter.
	"ObjectStore at:put: doesn't return the put object."
	^aCharacter
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Magnitude related' stamp: ''!
dateClassNewDay: day monthIndex: month year: year
	"OK for VA, Dol."

	^Date
		newDay: day 
		imonth: month 
		year: year! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Magnitude related' stamp: ''!
dateTimeClass
	"Should use something faster than this later."

	^self behaviorForName: #Timestamp! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Magnitude related' stamp: ''!
dateTimeClassDate: aDate time: aTime

	^self dateTimeClass
		newDate: aDate
		time: aTime! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Dialect specific hooks' stamp: ''!
defaultExchangeRules
	"ObjectStudio requires PmrSymbol as an absolute minimum
	 because of how Strings and Symbols are implemented in that
	 dialect."

	^Array with: (self behaviorForName: #PmrSymbol)! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Message related' stamp: ''!
directedMessageClass
	"Not all Smalltalk dialects have an DirectedMessage class."

	^self behaviorForName: #Action.
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Collection related' stamp: ''!
fixupLiteralArray: anArray
	"ObjectStudio doesn't parse literal arrays according to
	 normal Smalltalk convention. This is used to compensate."

	"Note: There are more issues than just the ones compensated
		for here. For example, #(-1) gets parsed as #- and 1."

	| new |
	^anArray collect: [:ea |
		new := ea.
		ea == #nil ifTrue: [new := nil].
		ea == #true ifTrue: [new := true].
		ea == #false ifTrue: [new := false].
		ea class == anArray class ifTrue: [ new := self fixupLiteralArray: ea ].
		new
	]
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Magnitude related' stamp: ''!
floatAsNumeratorAndDenominator: aFloat
	"	self new floatAsNumeratorAndDenominator: -1.2.	"
	"	self new floatAsNumeratorAndDenominator: (1/3) asFloat.	"
	"Note: I'm sure someone can come up with a more efficient way to do this."

	| multiple num den gcd |
	aFloat = 0.0 ifTrue: [^Array with: 0 with: 1].
	den := 1.
	[
		num := (multiple := aFloat * den) truncated.
		multiple = num
	] whileFalse: [den := den * 10].
	gcd := num gcd: den.
	^Array with: (num / gcd) asInteger with: (den / gcd) asInteger.
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Collection related' stamp: ''!
keyedCollection: keyedCollection associationsDo: oneArgBlock
	"ObjectStudio #associationsDo: expects two arguments."
	"VisualAge doesn't implement #associations."
	"VisualWorks doesn't allow #addAll: for populating."

	keyedCollection associationsDo: [:key :value | oneArgBlock value: (Association key: key value: value)]! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Message related' stamp: ''!
messageClass
	"Not all Smalltalk dialects have an Message class."

	^self behaviorForName: #SmalltalkMessage.
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Namespace related' stamp: ''!
namespace: ns bindingFor: aKey
	"A namespace should be considered anything with visibility
	 to variable bindings. It could be the Smalltalk dictionary, a
	 class, a pool dictionary, or a real dialect namespace. Answer
	 nil if ns isn't considered a namespace, or if the key wasn't
	 found in the namespace."

	ns isClass
		ifTrue: [ ^self behavior: ns bindingFor: aKey ].
	(ns respondsTo: #associationAt:ifAbsent:) ifTrue: [
		^ns
			associationAt: aKey
			ifAbsent: [].
	].
	^nil! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Namespace related' stamp: ''!
namespace: ns newBindingFor: key

	^ns 
		associationAt: key asSymbol
		ifAbsent: [
			self at: key put: nil.
			self associationAt: key asSymbol.
		]
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Namespace related' stamp: ''!
rootNamespace
	"System is the root namespace for ObjectStudio."

	^Object environment
! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Magnitude related' stamp: ''!
secondsFromTime: aTime
	"ObjectStudio stores time in 1/100 of a second."

	^aTime asNumber * 10

! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Collection related' stamp: ''!
sequenceableCollection: sequenceableCollection indexOfSubCollection: aSubcollection startingAt: anIndex
	"VisualSmalltalk and ObjectStudio don't implement #indexOfSubCollection:startingAt:, as most others do."

	| subcollectionSize first subIndex lastIndex 	element |
	(subcollectionSize := aSubcollection size) == 0 ifTrue: [^0].
	first := aSubcollection at: 1.
	anIndex to: sequenceableCollection size - subcollectionSize + 1 do: [:index |
		element := sequenceableCollection at: index.
		element = first ifTrue: [
			subIndex := index + 1.
			lastIndex := index + subcollectionSize.
			[
				subIndex >= lastIndex ifTrue: [^index].
				(sequenceableCollection at: subIndex) = (aSubcollection at: subIndex - index + 1)]
			whileTrue: [subIndex := subIndex + 1].
		].
	].
	^0! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^(self behaviorIsMetaclass: aClass)
		ifTrue: [ aClass instanceClass symbol ]
		ifFalse: [ aClass symbol ]! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Behavior related' stamp: ''!
stringAsInstVarName: aString
	"Convert aString so that it can be found in #allInstVarNames.
	 GS/S and ObjectStudio expect symbols, other dialects use strings."

	^aString asSymbol! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Stream related' stamp: ''!
stringExistsAsFile: filename
	^(filename asFilename readStreamOnError: []) notNil! !

!SrpPortableBehaviorObjectStudio methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#ObjectStudio! !

SrpPortableBehavior subclass: #SrpPortableBehaviorSqueak
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorSqueak classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation 	 "

	The port to Squeak was done at Camp Smalltalk 1
	and cleaned up for release shortly after.

	Some members of the Squeak community appear
	to use SRP 2.2 quite a bit.

	SRP 3.0 was tested with Squeak 3.7beta 5878.

"
! !

!SrpPortableBehaviorSqueak classSide methodsFor: 'As yet unclassified' stamp: 'sd 7/19/2006 16:50'!
mightApplyForImage
	"This test works for Squeak"

	^(Smalltalk respondsTo: #version) 
	and: ['Squeak*' match: Smalltalk version]! !

!SrpPortableBehaviorSqueak methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior isMeta! !

!SrpPortableBehaviorSqueak methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  'ByteArray withAll: (65 to: 100)' works for VW but not VA."

	^receiver asByteArray! !

!SrpPortableBehaviorSqueak methodsFor: 'String related' stamp: ''!
characterCollection: characterCollection at: index putCharacter: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW, Squeak, and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer instead of character, but that would be less
	 portable with other Smalltalk dialects."

	^characterCollection basicAt: index put: aCharacter asInteger
! !

!SrpPortableBehaviorSqueak methodsFor: 'Magnitude related' stamp: ''!
dateClassNewDay: day monthIndex: month year: year
	"OK for VA, Dol."

	^Date
		newDay: day 
		month: month 
		year: year! !

!SrpPortableBehaviorSqueak methodsFor: 'Magnitude related' stamp: ''!
dateTimeClass

	^self behaviorForName: #DateAndTime! !

!SrpPortableBehaviorSqueak methodsFor: 'Magnitude related' stamp: ''!
decimalClassFromString: aString

	^aString asFixedPoint
! !

!SrpPortableBehaviorSqueak methodsFor: 'Namespace related' stamp: ''!
namespace: ns bindingFor: key
	"A namespace should be considered anything with visibility
	 to variable bindings. It could be the Smalltalk dictionary, a
	 class, a pool dictionary, or a real dialect namespace. Answer
	 nil if ns isn't considered a namespace, or if the key wasn't
	 found in the namespace."

	ns == Smalltalk ifTrue: [
		^Smalltalk
			associationAt: key
			ifAbsent: []
	].
	^Smalltalk
		associationAt: (ns name, '::', key) asSymbol
		ifAbsent: [].! !

!SrpPortableBehaviorSqueak methodsFor: 'String related' stamp: ''!
objectIsString: anObject
	"Dolphin, GS/S, and Squeak don't support #isString."
	^anObject isKindOf: String! !

!SrpPortableBehaviorSqueak methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^aClass isMeta
		ifTrue: [ aClass soleInstance name ]
		ifFalse: [ aClass name ]! !

!SrpPortableBehaviorSqueak methodsFor: 'Stream related' stamp: ''!
stringAsBinaryFileStream: filename
	^(self stringAsTextFileStream: filename) binary! !

!SrpPortableBehaviorSqueak methodsFor: 'Stream related' stamp: ''!
stringAsTextFileStream: filename
	| fileStreamClass fileStream |
	fileStreamClass := (self behaviorForName: #FileStream) concreteStream.
	fileStream := fileStreamClass fileNamed: filename.
	fileStream resetToStart.
	^fileStream! !

!SrpPortableBehaviorSqueak methodsFor: 'String related' stamp: ''!
stringAsTwoByteString: aString

	^aString! !

!SrpPortableBehaviorSqueak methodsFor: 'Stream related' stamp: ''!
stringExistsAsFile: filename
	^(Smalltalk at: #FileDirectory) default fileExists: filename! !

!SrpPortableBehaviorSqueak methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#Squeak! !

SrpPortableBehavior subclass: #SrpPortableBehaviorStx
	instanceVariableNames: 'streamBasePosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorStx classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation		"

	Claus Gittinger worked on the first Smalltalk/X port while at
	Camp Smalltalk 1. He got the code working but he decided to 
	change ST/X to be more standard than to change this code. 
	He said he would provide a ST/X patch that would, among 
	other things, correct the problems with streams being 1 based 
	instead of the standard 0 based. I'm not sure if anyone ever
	used SRP 2.2 on ST/X.

	SRP 3.0 was ported to Smalltalk/X 5.1.4.

"
! !

!SrpPortableBehaviorStx classSide methodsFor: 'As yet unclassified' stamp: ''!
mightApplyForImage

	^(Smalltalk respondsTo: #isSmalltalkX)
		and: [Smalltalk isSmalltalkX].
! !

!SrpPortableBehaviorStx methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior basicNew: newSize
	"Floats return nil if you use #srpNew: to create a sized instance.
	 Numbers should know their own sizes. This rule might be applicable
	 to other dialects, but I'm reluctant to make that change so close to the
	 3.0 release. This is a slower implementation than extending base code,
	 but I wanted to keep the change from other untested dialects."

	(aBehavior inheritsFrom: Number) ifTrue: [^aBehavior srpNew].
	^aBehavior basicNew: newSize
! !

!SrpPortableBehaviorStx methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior bindingFor: key
	"ST/X doesn't have classPools so it stores class variables
	 in the Smalltalk dictionary with a funny name."

	^self rootNamespace
		associationAt: (aBehavior name, ':', key) asSymbol
		ifAbsent: [].! !

!SrpPortableBehaviorStx methodsFor: 'Behavior related' stamp: ''!
behaviorForNamespacePath: anArray
	"	(SrpPortal behaviorForNamespacePath: #(Smalltalk Core String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(Core String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(Absent String)) == String.	"
	"	(SrpPortal behaviorForNamespacePath: #(Array String)) == String.	"

	| stream |
	stream := WriteStream on: String new.
	1 to: anArray size - 1 do: [:i |
		stream 
			nextPutAll: (anArray at: i); 
			nextPutAll: '::'.
	].
	stream nextPutAll: anArray last.
	self rootNamespace
		at: stream contents asSymbol
		ifPresent: [:found | ^found ].
	^self rootNamespace 
		at: anArray last 
		ifAbsent: [].
! !

!SrpPortableBehaviorStx methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior isMeta! !

!SrpPortableBehaviorStx methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA and ST/X but not VW.
	  'ByteArray withAll: (65 to: 100)' works for VW but not VA."

	^receiver asByteArray! !

!SrpPortableBehaviorStx methodsFor: 'Magnitude related' stamp: ''!
dateClassNewDay: day monthIndex: month year: year

	^Date
		newDay: day 
		month: month 
		year: year! !

!SrpPortableBehaviorStx methodsFor: 'Magnitude related' stamp: ''!
decimalClassFromString: aString

	^aString asFixedPoint! !

!SrpPortableBehaviorStx methodsFor: 'Dialect specific hooks' stamp: ''!
defaultExchangeRules
	"Answer a collection of mapping rule that apply to this Smalltalk dialect."

	^Array with: (self behaviorForName: #PmrSymbol)

	"Using PmrSymbol here is a bit of hack used to get ST/X support
	 ready in time for the SRP 3.0 release. 'Symbol basicNew: 5'
	 returns a String instead of a Symbol in ST/X. Rules based on
	 class name could not be applied. Portable mapping rules
	 don't normally need to be loaded for basic SRP function, but
	 they do now--at least for ST/X."! !

!SrpPortableBehaviorStx methodsFor: 'Namespace related' stamp: ''!
namespace: ns bindingFor: key
	"A namespace should be considered anything with visibility
	 to variable bindings. It could be the Smalltalk dictionary, a
	 class, a pool dictionary, or a real dialect namespace. Answer
	 nil if ns isn't considered a namespace, or if the key wasn't
	 found in the namespace."

	ns == Smalltalk ifTrue: [
		^Smalltalk
			associationAt: key
			ifAbsent: []
	].
	^Smalltalk
		associationAt: (ns name, '::', key) asSymbol
		ifAbsent: [].! !

!SrpPortableBehaviorStx methodsFor: 'Namespace related' stamp: ''!
namespace: ns newBindingFor: key

	ns at: key put: nil.
	^self namespace: ns bindingFor: key.
! !

!SrpPortableBehaviorStx methodsFor: 'Namespace related' stamp: ''!
namespace: ns resolve: key

	^ns at: key ifAbsent: []! !

!SrpPortableBehaviorStx methodsFor: 'Namespace related' stamp: ''!
namespacePathForBehavior: aBehavior

	| oc subject |
	aBehavior class == SrpMetastate ifTrue: [^aBehavior behaviorName].
	oc := OrderedCollection new.
	subject := aBehavior.
	oc addLast: subject nameWithoutPrefix.
	[ subject nameSpace = Smalltalk ] whileFalse: [ 
		oc addFirst: (subject := subject nameSpace) nameWithoutPrefix.
	].
	^oc asArray! !

!SrpPortableBehaviorStx methodsFor: 'Collection related' stamp: ''!
optionalWeakKeyIdentityDictionaryClass

	^Smalltalk at: #WeakIdentityDictionary! !

!SrpPortableBehaviorStx methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^aClass isMeta
		ifTrue: [ aClass soleInstance nameWithoutPrefix ]
		ifFalse: [ aClass nameWithoutPrefix ]! !

!SrpPortableBehaviorStx methodsFor: 'Stream related' stamp: ''!
streamBasePosition
	"ST/X used to use 1 based streams instead of 0 based like
	 most other Smalltalk dialects. Version 5.1.4 is known to
	 use 0 based streams, but I don't know when that change
	 was first made."

	streamBasePosition isNil 
		ifTrue: [streamBasePosition := (ReadStream on: 'Test') reset; position].
	^streamBasePosition! !

!SrpPortableBehaviorStx methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#'Smalltalk/X'! !

SrpPortableBehavior subclass: #SrpPortableBehaviorVa
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorVa classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation  	"

	VisualAge was the first Smalltalk dialect that SRP worked with.
	SRP has been used on all VisualAge releases since 3.0.

	All SRP code is maintained through a VisualAge image to
	an ENVY/Developer code repository. The code is then
	exported to various forms for use by the other Smalltalk
	dialects.

	SRP 3.0 was developed with VisualAge 6.0.2.

"


! !

!SrpPortableBehaviorVa classSide methodsFor: 'As yet unclassified' stamp: ''!
mightApplyForImage

	| sys |
	Smalltalk isNil ifTrue: [^false].
	sys := Smalltalk at: #System ifAbsent: [^false].
	^sys vmType = 'ES'

! !

!SrpPortableBehaviorVa methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior bindingFor: key

	^aBehavior
		variableAssociationAt: key
		ifAbsent: [].! !

!SrpPortableBehaviorVa methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  'ByteArray withAll: (65 to: 100)' works for VW but not VA."

	^receiver asByteArray
! !

!SrpPortableBehaviorVa methodsFor: 'Magnitude related' stamp: ''!
dateTimeClass

	^self behaviorForName: #AbtTimestamp
! !

!SrpPortableBehaviorVa methodsFor: 'Collection related' stamp: ''!
identitySetClass
	"VA implements an EsIdentitySet. This method uses
	 an indirect reference to that class so that code
	 can be loaded into other Smalltalk dialects."

	^self behaviorForName: #EsIdentitySet
! !

!SrpPortableBehaviorVa methodsFor: 'Namespace related' stamp: ''!
namespace: ns bindingFor: aKey
	"A namespace should be considered anything with visibility
	 to variable bindings. It could be the Smalltalk dictionary, a
	 class, a pool dictionary, or a real dialect namespace. Answer
	 nil if ns isn't considered a namespace, or if the key wasn't
	 found in the namespace."

	ns isClass 
		ifTrue: [ ^ns classVariableAssociationFor: aKey ].
	(ns respondsTo: #associationAt:ifAbsent:) ifTrue: [
		^ns
			associationAt: aKey
			ifAbsent: [].
	].
	^nil
! !

!SrpPortableBehaviorVa methodsFor: 'Namespace related' stamp: ''!
namespace: ns newBindingFor: key

	^ns environment 
		declareVariable: key.
! !

!SrpPortableBehaviorVa methodsFor: 'Stream related' stamp: ''!
newReadWriteStreamSize: initialSize binary: binary
	"Creates a stream preallocated with the appropriate
	 type of character collection. This was necessary
	 because some dialects need special treatment to
	 prevent preallocated contents from being considered
	 part of the written data."

	^(super newReadWriteStreamSize: initialSize binary: binary)
		truncate
! !

!SrpPortableBehaviorVa methodsFor: 'Collection related' stamp: ''!
optionalWeakKeyIdentityDictionaryClass
	"Private. VisualAge provides a weakly keyed identity
	 dictionary that is faster than simply using an
	 IdentityDictionary. By being weakly keyed, this will
	 minimize memory consumption if saving extremely
	 large and complex objects."

	^Smalltalk 
		at: #AbtWeakKeyIdentityDictionary 
		ifAbsent: [IdentityDictionary]
! !

!SrpPortableBehaviorVa methodsFor: 'Collection related' stamp: ''!
sequenceableCollection: receiver equals: aCollection
	"Answer true if the receiver and aCollection contain the same elements
	 in the same order."

	^receiver equals: aCollection

	"VW doesn't support #equals:. VA does."! !

!SrpPortableBehaviorVa methodsFor: 'Stream related' stamp: ''!
stringAsBinaryFileStream: filename

	| cfsClass |
	cfsClass := self behaviorForName: #CfsReadWriteFileStream.
	^(cfsClass open: filename)
		isBytes: true;
		yourself
	! !

!SrpPortableBehaviorVa methodsFor: 'Stream related' stamp: ''!
stringAsTextFileStream: filename

	| cfsClass |
	cfsClass := self behaviorForName: #CfsReadWriteFileStream.
	^cfsClass open: filename
	! !

!SrpPortableBehaviorVa methodsFor: 'String related' stamp: ''!
stringAsTwoByteString: aString
	"Convert aString so that it can contain
	 characters with values up to 16rFFFF (65535)."

	^aString asDBString! !

!SrpPortableBehaviorVa methodsFor: 'Stream related' stamp: ''!
stringExistsAsFile: filename

	| statClass |
	statClass := self behaviorForName: #CfsStat.
	^(statClass stat: filename) isCfsError not.
	! !

!SrpPortableBehaviorVa methodsFor: 'Environmental awareness' stamp: ''!
timeZonesFromGMT
	"Answer the number of time zones from GMT that the user
	 is geographically located."

	^(self behaviorForName: #EmTimeStamp) timeZonesFromGMT! !

!SrpPortableBehaviorVa methodsFor: 'String related' stamp: ''!
twoByteStringClass
	"Not all Smalltalk dialects have support for double byte strings.
	 Answer nil if support isn't available."

	^Smalltalk at: #DBString ifAbsent: []! !

!SrpPortableBehaviorVa methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#VisualAge! !

SrpPortableBehaviorVa subclass: #SrpPortableBehaviorVaMainframe
	instanceVariableNames: 'base256ToBase64Table base64ToBase256Table'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorVaMainframe classSide methodsFor: 'Not categorized' stamp: ''!
documentation
	"This portal provides support for the EBCDIC character set found on mainframes.
	 The modification is courtesy of Jon Anderson [jonkanderson@yahoo.com]."
! !

!SrpPortableBehaviorVaMainframe classSide methodsFor: 'Not categorized' stamp: ''!
mightApplyForImage

	| sys |
	Smalltalk isNil ifTrue: [^false].
	sys := Smalltalk at: #System ifAbsent: [^false].
	^(sys image subsystemType: 'OS') = 'MVS'


! !

!SrpPortableBehaviorVaMainframe methodsFor: 'String related' stamp: ''!
base256ToBase64Table
	"This is used for converting to Base64."

	base256ToBase64Table isNil 
		ifTrue: [base256ToBase64Table := self newBase256ToBase64Table].
	^base256ToBase64Table

! !

!SrpPortableBehaviorVaMainframe methodsFor: 'String related' stamp: ''!
base64ToBase256Table
	"This is for converting from Base64."

	base64ToBase256Table isNil 
		ifTrue: [base64ToBase256Table := self newBase64ToBase256Table].
	^base64ToBase256Table

! !

!SrpPortableBehaviorVaMainframe methodsFor: 'String related' stamp: ''!
characterToUnsigned: aCharacter

	^aCharacter ebcdicToAscii asInteger! !

!SrpPortableBehaviorVaMainframe methodsFor: 'String related' stamp: ''!
sourceTextFromString: aString

	^aString ebcdicToAscii! !

!SrpPortableBehaviorVaMainframe methodsFor: 'String related' stamp: ''!
stringToSourceText: aString

	^aString asciiToEbcdic! !

!SrpPortableBehaviorVaMainframe methodsFor: 'Magnitude related' stamp: ''!
unsignedToCharacter: anUnsignedInteger

	^anUnsignedInteger asCharacter asciiToEbcdic! !

SrpPortableBehavior subclass: #SrpPortableBehaviorVs
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorVs classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation 	"

	SRP 3.0 was the first release to support VisualSmalltalk.
	VisualSmalltalk 3.1.0 was used for the port.

"
! !

!SrpPortableBehaviorVs classSide methodsFor: 'As yet unclassified' stamp: ''!
mightApplyForImage

	^(Smalltalk respondsTo: #product)
		and: [Smalltalk product = 'Visual Smalltalk']! !

!SrpPortableBehaviorVs methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior bindingFor: key

	^aBehavior classPool
		associationAt: key asString
		ifAbsent: [].! !

!SrpPortableBehaviorVs methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior isMetaClass! !

!SrpPortableBehaviorVs methodsFor: 'Behavior related' stamp: ''!
behaviorWithAllSubclasses: aBehavior do: oneArgBlock
	"Answer a collection containing aBehavior and all subclasses of aBehavior."

	"	self portal behaviorWithAllSubclasses: Collection do: [:ea | ].	"

	aBehavior withAllSubclassesDo: oneArgBlock.! !

!SrpPortableBehaviorVs methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  (ByteArray withAll: (65 to: 100)) works for VW but not VA.
	  (ByteArray withAll: 'ABC') doesn't work for VW."

	^(self objectIsString: receiver)
		ifTrue: [receiver asByteArray]
		ifFalse: [ByteArray withAll: receiver]! !

!SrpPortableBehaviorVs methodsFor: 'String related' stamp: ''!
characterCollection: characterCollection at: index putCharacter: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW, Squeak, and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer instead of character, but that would be less
	 portable with other Smalltalk dialects."

	characterCollection isDoubleByteString
		ifTrue: [characterCollection at: index put: aCharacter]
		ifFalse: [characterCollection basicAt: index put: aCharacter].
	^aCharacter! !

!SrpPortableBehaviorVs methodsFor: 'Magnitude related' stamp: ''!
dateClassNewDay: day monthIndex: month year: year

	^Date
		newDay: day 
		month: (Date nameOfMonth: month)
		year: year! !

!SrpPortableBehaviorVs methodsFor: 'Magnitude related' stamp: ''!
dateTimeClass

	^self behaviorForName: #TimeStamp
! !

!SrpPortableBehaviorVs methodsFor: 'Magnitude related' stamp: ''!
decimalClassFromString: aString

	| scale |
	scale := aString size - (aString indexOf: $. ifAbsent: [aString size]).
	^aString asNumber asFixedPoint: scale
! !

!SrpPortableBehaviorVs methodsFor: 'Message related' stamp: ''!
directedMessageClass
	"Not all Smalltalk dialects have an DirectedMessage class."

	^self behaviorForName: #Message.
! !

!SrpPortableBehaviorVs methodsFor: 'Magnitude related' stamp: ''!
floatAsNumeratorAndDenominator: aFloat
	"(VA, GS/S, ObjectStudio) use #asFraction, (VW, VSE) use #asRational"

	| temp |
	temp := aFloat asRational.
	^Array with: temp numerator with: temp denominator
! !

!SrpPortableBehaviorVs methodsFor: 'Namespace related' stamp: ''!
namespace: ns bindingFor: aKey
	"A namespace should be considered anything with visibility
	 to variable bindings. It could be the Smalltalk dictionary, a
	 class, a pool dictionary, or a real dialect namespace. Answer
	 nil if ns isn't considered a namespace, or if the key wasn't
	 found in the namespace."

	ns isClass
		ifTrue: [ ^self behavior: ns bindingFor: aKey ].
	(ns respondsTo: #associationAt:ifAbsent:) ifTrue: [
		^ns
			associationAt: aKey
			ifAbsent: [].
	].
	^nil! !

!SrpPortableBehaviorVs methodsFor: 'Namespace related' stamp: ''!
namespace: ns newBindingFor: key

	^ns 
		associationAt: key asSymbol
		ifAbsent: [
			self at: key put: nil.
			self associationAt: key asSymbol.
		]
! !

!SrpPortableBehaviorVs methodsFor: 'Stream related' stamp: ''!
newReadOverwriteStreamWith: aCollection
	"We want a stream that will start at the beginning
	 and that will leave data alone that it does not write."

	^(ReadWriteStream on: aCollection) reset; yourself! !

!SrpPortableBehaviorVs methodsFor: 'Stream related' stamp: ''!
newReadWriteStreamSize: initialSize binary: binary
	"Creates a stream preallocated with the appropriate
	 type of character collection. This was necessary
	 because some dialects need special treatment to
	 prevent preallocated contents from being considered
	 part of the written data."

	^(super newReadWriteStreamSize: initialSize binary: binary)
		truncate
! !

!SrpPortableBehaviorVs methodsFor: 'Collection related' stamp: ''!
optionalWeakKeyIdentityDictionaryClass

	^Smalltalk
		at: #EphemeronDictionary
		ifAbsent: [IdentityDictionary]
! !

!SrpPortableBehaviorVs methodsFor: 'Collection related' stamp: ''!
sequenceableCollection: receiver equals: aCollection
	"Answer true if the receiver and aCollection contain the same elements
	 in the same order."

	^receiver size = aCollection size
		and: [(receiver indexOfCollection: aCollection) = 1]

	"VW doesn't support #equals:. VA does. VS #equals: is class specific."! !

!SrpPortableBehaviorVs methodsFor: 'Collection related' stamp: ''!
sequenceableCollection: sequenceableCollection indexOfSubCollection: aSubcollection startingAt: anIndex
	"VisualSmalltalk and ObjectStudio don't implement #indexOfSubCollection:startingAt:, as most others do."

	| subcollectionSize first subIndex lastIndex 	element |
	(subcollectionSize := aSubcollection size) == 0 ifTrue: [^0].
	first := aSubcollection at: 1.
	anIndex to: sequenceableCollection size - subcollectionSize + 1 do: [:index |
		element := sequenceableCollection at: index.
		element = first ifTrue: [
			subIndex := index + 1.
			lastIndex := index + subcollectionSize.
			[
				subIndex >= lastIndex ifTrue: [^index].
				(sequenceableCollection at: subIndex) = (aSubcollection at: subIndex - index + 1)]
			whileTrue: [subIndex := subIndex + 1].
		].
	].
	^0! !

!SrpPortableBehaviorVs methodsFor: 'Behavior related' stamp: ''!
shortNameForBehavior: aClass
	"This should answer #Metaclass instead of the
	 class symbol that Metaclass is an instance for."

	aClass class == SrpMetastate ifTrue: [^aClass behaviorShortName].
	^(self behaviorIsMetaclass: aClass)
		ifTrue: [ aClass class symbol ]
		ifFalse: [ aClass symbol ]
! !

!SrpPortableBehaviorVs methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^(self behaviorIsMetaclass: aClass)
		ifTrue: [ aClass instanceClass symbol ]
		ifFalse: [ aClass symbol ]! !

!SrpPortableBehaviorVs methodsFor: 'Stream related' stamp: ''!
stringAsBinaryFileStream: filename

	| cfsClass |
	cfsClass := self behaviorForName: #CfsReadWriteFileStream.
	^(cfsClass open: filename)
		isBytes: true;
		yourself
	! !

!SrpPortableBehaviorVs methodsFor: 'Stream related' stamp: ''!
stringAsTextFileStream: filename

	| cfsClass |
	cfsClass := self behaviorForName: #CfsReadWriteFileStream.
	^cfsClass open: filename
	! !

!SrpPortableBehaviorVs methodsFor: 'String related' stamp: ''!
stringAsTwoByteString: aString
	"Convert aString so that it can contain
	 characters with values up to 16rFFFF (65535)."

	"This behavior lookup could be a performance problem where
	 double byte strings are used frequently."
	^(self behaviorForName: #DoubleByteString) fromString: aString! !

!SrpPortableBehaviorVs methodsFor: 'Stream related' stamp: ''!
stringExistsAsFile: filename
	^(Smalltalk at: #File) exists: filename ! !

!SrpPortableBehaviorVs methodsFor: 'Environmental awareness' stamp: ''!
timeZonesFromGMT

	| sysTimeClass kl localTime systemTime |
	sysTimeClass := self behaviorForName: #WinSystemTime.
	kl := self behaviorForName: #KernelLibrary.
	systemTime := sysTimeClass new.
	localTime := sysTimeClass new.
	kl getLocalTime: localTime asParameter.
	kl getSystemTime: systemTime asParameter.
	^(systemTime asDate asSeconds + systemTime asTime asSeconds) -
	(localTime asDate asSeconds + localTime asTime asSeconds) // -3600

! !

!SrpPortableBehaviorVs methodsFor: 'String related' stamp: ''!
twoByteStringClass
	"Not all Smalltalk dialects have support for double byte strings.
	 Answer nil if support isn't available."

	^self behaviorForName: #DoubleByteString! !

!SrpPortableBehaviorVs methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#VisualSmalltalk! !

SrpPortableBehavior subclass: #SrpPortableBehaviorVw
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorVw classSide methodsFor: 'As yet unclassified' stamp: ''!
comments

	"VW doesn't support (0 to: 255) asByteArray as VA would. Might suggest Cincom add it."
	"VW doesn't support SequenceableCollection>>doWithIndex:."

! !

!SrpPortableBehaviorVw classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation 		"

	VisualWorks (with ENVY) was the first port from the original 
	VisualAge-based code. The code was ported to VW 3.0
	(without ENVY) at Camp Smalltalk 1.

	SRP 2.2 was ported to VW 5i in both ENVY/Developer and 
	parcel form, but it didn't get much distribution.

	SRP 3.0 was ported to VisualWorks 7.3 beta. SRP 3.0
	parcels should still work with VW 5i. The ENVY DAT
	file for SRP 3.0 probably works for early versions
	of VisualWorks that have ENVY/Developer support.

"! !

!SrpPortableBehaviorVw classSide methodsFor: 'As yet unclassified' stamp: ''!
mightApplyForImage

	(Smalltalk respondsTo: #versionName) ifFalse: [^false].
	((Smalltalk versionName isKindOf: String)
		and: ['VisualWorks*' match: Smalltalk versionName]) ifFalse: [^false].
	^(Class respondsTo: #fullyQualifiedReference) not! !

!SrpPortableBehaviorVw methodsFor: 'Behavior related' stamp: ''!
behavior: aBehavior bindingFor: key
	"Default #bindingFor: only looks in the class itself."

	^aBehavior fullBindingFor: key
! !

!SrpPortableBehaviorVw methodsFor: 'Behavior related' stamp: ''!
behaviorIsBytes: aBehavior

	^aBehavior isBits! !

!SrpPortableBehaviorVw methodsFor: 'Behavior related' stamp: ''!
behaviorIsMetaclass: aBehavior

	^aBehavior isMeta! !

!SrpPortableBehaviorVw methodsFor: 'Collection related' stamp: ''!
byteArrayClassWithAll: receiver
	" '(65 to: 100) asByteArray' works for VA but not VW.
	  (ByteArray withAll: (65 to: 100)) works for VW but not VA.
	  (ByteArray withAll: 'ABC') doesn't work for VW."

	^(self objectIsString: receiver)
		ifTrue: [receiver asByteArray]
		ifFalse: [ByteArray withAll: receiver]! !

!SrpPortableBehaviorVw methodsFor: 'String related' stamp: ''!
characterCollection: characterCollection at: index putCharacter: aCharacter
	"Store aCharacter at the variable index position in the receiver.
	 #basicAt:put: in VW, Squeak, and Dolphin expect an integer. VA and ST/X
	 expect characters. Didn't use #at:put: (expects character) in VW
	 because symbols are read-only. Could have set the store type to
	 unsigned integer instead of character, but that would be less
	 portable with other Smalltalk dialects."

	^characterCollection basicAt: index put: aCharacter asInteger
! !

!SrpPortableBehaviorVw methodsFor: 'Magnitude related' stamp: ''!
dateClassNewDay: day monthIndex: month year: year
	"VW."

	^Date
		newDay: day 
		monthNumber: month 
		year: year! !

!SrpPortableBehaviorVw methodsFor: 'Magnitude related' stamp: ''!
floatAsNumeratorAndDenominator: aFloat
	"(VA, GS/S, ObjectStudio) use #asFraction, (VW, VSE) use #asRational"

	| temp |
	temp := aFloat asRational.
	^Array with: temp numerator with: temp denominator
! !

!SrpPortableBehaviorVw methodsFor: 'Collection related' stamp: ''!
identitySetClass
	"VW implements an IdentitySet. This method uses
	 an indirect reference to that class so that code
	 can be loaded into other Smalltalk dialects."

	^self behaviorForName: #IdentitySet
! !

!SrpPortableBehaviorVw methodsFor: 'Magnitude related' stamp: ''!
secondsFromTime: aTime
	"Answer the number of seconds since midnight for aTime.
	 It is OK to answer a non-integer value for fractional seconds."

	^aTime asSeconds
! !

!SrpPortableBehaviorVw methodsFor: 'Behavior related' stamp: ''!
shortNameForClass: aClass
	"This would return #String for both String and String 
	 class (not #Metaclass)."
	^aClass isMeta
		ifTrue: [ aClass soleInstance name ]
		ifFalse: [ aClass name ]! !

!SrpPortableBehaviorVw methodsFor: 'String related' stamp: ''!
twoByteStringClass
	"Not all Smalltalk dialects have support for double byte strings.
	 Answer nil if support isn't available."

	^Smalltalk at: #TwoByteString ifAbsent: []! !

!SrpPortableBehaviorVw methodsFor: 'Environmental awareness' stamp: ''!
vmTypeName

	^#VisualWorks! !

SrpPortableBehaviorVw subclass: #SrpPortableBehaviorVwNs
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol-SrpPortableBehaviors'!

!SrpPortableBehaviorVwNs classSide methodsFor: 'As yet unclassified' stamp: ''!
documentation "

	SrpPortableBehaviorVwNs is intended for VisualWorks 
	 releases with namespace support (releases since VW 5i).

	It has been tested on VW 5i.1 (envy), 7.0, and 7.3.

	Code is maintained in a VW 5i.1 envy image from which
	parcel files are generated. Parcels are generated from 
	with a modified version of the parcel generation framework
	provided by OTI/IBM in VW 5i.1.

"
! !

!SrpPortableBehaviorVwNs classSide methodsFor: 'As yet unclassified' stamp: ''!
mightApplyForImage

	(Smalltalk respondsTo: #versionName) ifFalse: [^false].
	(Class respondsTo: #fullyQualifiedReference) ifFalse: [^false].
	((Smalltalk versionName isKindOf: String) 
		and: ['VisualWorks*' match: Smalltalk versionName]) ifTrue: [^true].
	"Noticed in VW 7.3"
	^(Smalltalk versionName respondsTo: #defaultString)
		and: ['VisualWorks*' match: Smalltalk versionName defaultString]
! !

!SrpPortableBehaviorVwNs methodsFor: 'Namespace related' stamp: ''!
isNamespaceImage

	^true! !

!SrpPortableBehaviorVwNs methodsFor: 'Namespace related' stamp: ''!
namespacePathForBehavior: aBehavior
	"The code that goes here depends on the language and
	 dialect being used. Answer an array of symbols that another
	 objectspace can use to retrieve behavior later."

	"Note that this only works for VW since 5i. Superclass behavior
	 would work for images prior to 5i."
	aBehavior class == SrpMetastate ifTrue: [^aBehavior behaviorName].
 	^aBehavior fullyQualifiedReference path! !

Object subclass: #SrpState
	instanceVariableNames: 'metastate namedValues indexableValues extendedAttributes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpState classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 Instances are only created when loading state into an object space that
 doesn't have a corresponding behavior class as specified by a
 metastate. This is an alternative to failing a load that allows state
 to be loaded and then resaved even though behavior wasn't defined.
"! !

!SrpState classSide methodsFor: 'instance creation' stamp: ''!
new: intSize metastate: aMetastate

	^self new 
		srpMetastate: aMetastate;
		srpNamedValues: (Array new: aMetastate instSize);
		srpIndexableValues: (Array new: intSize);
		yourself
! !

!SrpState classSide methodsFor: 'instance creation' stamp: ''!
newLike: anInstance
	"Answer a new instance of the receiver that
	 resembles anInstance. This is a convenience
	 method not used by SRP. Use sparingly because
	 it creates a metastate each time."
	"	self newLike: Dictionary new "

	| mstate |
	mstate := SrpConfiguration default newMetastateForClass: anInstance srpClass.
	^self newLike: anInstance metastate: mstate.! !

!SrpState classSide methodsFor: 'instance creation' stamp: ''!
newLike: anInstance metastate: mstate
	"Answer a new instance of the receiver that
	 resembles anInstance. This is a convenience
	 method not used by SRP."

	| newInst |
	newInst := mstate isVariable 
		ifTrue: [ self new: anInstance size metastate: mstate ]
		ifFalse: [ self newMetastate: mstate ].
	newInst conformTo: anInstance.
	^newInst.! !

!SrpState classSide methodsFor: 'instance creation' stamp: ''!
newMetastate: aMetastate
	"Answer a new instance of the receiver associated
	 with aMetastate."

	^self new: 0 metastate: aMetastate
! !

!SrpState methodsFor: 'converting' stamp: ''!
asAttributeDictionary
	"Answer a dictionary with the attributes and indexable
	 contents of the receiver. This is a convenience method
	 not used by SRP."

	| answer |
	answer := Dictionary new.
	"Add indexed variables."
	1 to: self size do: [:indexedVarIndex |
		answer 
			at: indexedVarIndex 
			put: (self at: indexedVarIndex).
	].
	"Add named variables."
	metastate allInstVarNames do: [:varName |
		answer
			at: varName
			put: (self srpAttribute: varName).
	].
	^answer! !

!SrpState methodsFor: 'converting' stamp: ''!
asBehavior: aBehavior
	"Answer an instance of aBehavior with the state
	 of the receiver. This is a convenience method not
	 used by SRP."

	| inst |
	inst := (metastate isVariable and: [ SrpPortal behaviorHasIndexableValues: aBehavior ])
		ifTrue: [ aBehavior srpNew: self size ]
		ifFalse: [ aBehavior srpNew ].
	self embellish: inst.
	^inst! !

!SrpState methodsFor: 'customary behavior' stamp: ''!
at: index
	"Answer the object at the position of the Integer index variable in the receiver.

	 Fail if index is not an kind of Integer Fail if index is < 1 Fail if the receiver does not
	 have indexed variables Fail if index is greater than the number of indexed variables in the
	 receiver."

	^self srpIndexableValueAt: index
! !

!SrpState methodsFor: 'customary behavior' stamp: ''!
at: index put: value
	"Store the Object argument value in the variable
	 position of the receiver represented by the Integer
	 argument index. Answer the Object argument value that
	 was stored.

	 Fail if index is not a kind of Integer Fail if index
	 is < 1 Fail if the receiver does not have indexed
	 variables Fail if index is greater than the number of
	 indexed variables in the receiver Fail if the class
	 of value is not appropriate for the index variable
	 type of the receiver."

	^self srpIndexableValueAt: index put: value
! !

!SrpState methodsFor: 'converting' stamp: ''!
conformTo: anInstance
	"Borrow similarly named attributes and indexable contents
	 of anInstance. This is a convenience method not used by SRP."

	self copyFrom: anInstance into: self.! !

!SrpState methodsFor: 'converting' stamp: ''!
copyFrom: source into: destination
	"Makes like attributes and indexable contents of destination
	 the same as source. This is a convenience method not used by SRP."

	"Embellish indexed variables -- if both have some."
	1 
	to: (source size min: destination size)
	do: [:indexedVarIndex |
		destination 
			at: indexedVarIndex 
			put: (source at: indexedVarIndex).
	].
	"Embellish named variables -- if there are any."
	destination srpClass allInstVarNames do: [:varName |
		(source srpClass srpSlotNamed: varName) > 0 ifTrue: [
			destination srpAttribute: varName put: (source srpAttribute: varName).
		].
	].
	^destination! !

!SrpState methodsFor: 'customary behavior' stamp: ''!
doesNotUnderstand: aMessage
	"This is implemented just to be user friendly, it simulates
	 having basic accessor methods for the instance variables."

	| varName |
	"Handle named inst var getters."
	aMessage arguments size = 0 ifTrue: [ 
		varName := aMessage selector.
		(metastate srpSlotNamed: varName) = 0 ifFalse: [
			^self srpAttribute: varName
		].
	].
	"Handle named inst var setters."
	aMessage arguments size = 1 ifTrue: [ 
		varName := aMessage selector asString copyWithout: $:.
		(metastate allInstVarNames includes: varName) = 0 ifFalse: [
			^self srpAttribute: varName put: aMessage arguments first.
		].
	].
	^super doesNotUnderstand: aMessage.! !

!SrpState methodsFor: 'converting' stamp: ''!
embellish: anInstance
	"Add similarly named attributes and indexable contents
	 to anInstance. This is a convenience method not
	 used by SRP."

	self copyFrom: self into: anInstance.
	^anInstance.! !

!SrpState methodsFor: 'customary behavior' stamp: ''!
size
	"Answer an Integer which represents the number of
	 indexable fields in the receiver. This value is the
	 same as the largest legal subscript."

	^self srpIndexableSize! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpAttribute: varNameString
	"This can be sent by mapping rules. It is a way to access named
	 variable slots regardless of whether accessor methods are defined."

	| s |
	^(s := self srpClass srpSlotNamed: varNameString) = 0
		ifTrue: [nil]
		ifFalse: [self srpNamedValueAt: s]! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpAttribute: varNameString put: anObject
	"This can be sent by mapping rules. It is a way to access named
	 variable slots regardless of whether accessor methods are defined."

	| s |
	(s := self srpClass srpSlotNamed: varNameString) = 0
		ifFalse: [self srpNamedValueAt: s put: anObject].
	^anObject! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpClass
	"Answer the class of the receiver with respect to srp
	 code. The receiver doesn't have a normal class (otherwise
	 it wouldn't have been created); therefore, answer the
	 metastate of the receiver which implements some basic
	 Behavior behaviors."

	^metastate! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpExtendedAttributes
	"Answer an object that is to be saved along with the receiver.
	 The object will be sent as an argument to #srpExtendedAttributes:
	 when loading. For flexibility it is recommended that you answer
	 a dictionary with key value pairs."

	^extendedAttributes
! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpExtendedAttributes: anObject
	"anObject was saved along with the receiver as a holder of additional
	 information related to the receiver. This method is sent when loading to
	 reassociate the two objects. anObject is customarily a dictionary with
	 key value pairs."

	extendedAttributes := anObject.
! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpIndexableSize
	"Answer the basic size of the receiver. This can be sent by
	 mapping rules. It is preferable to #basicSize because it
	 works with special loader constructs."

	^self srpIndexableValues size! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpIndexableValueAt: index
	"Answer the object at the variable index position of the receiver."

	^self srpIndexableValues at: index! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpIndexableValueAt: index put: value
	"Store the Object argument value at the variable index position 
	 in the receiver."

	^self srpIndexableValues at: index put: value! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpIndexableValues

	^indexableValues! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpIndexableValues: anIndexableCollection

	indexableValues := anIndexableCollection! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpMetastate
	"Answer the metastate that describes the data of the
	 receiver. The metastate is acting as a class to provide
	 basic behaviors."

	^metastate! !

!SrpState methodsFor: 'initialization' stamp: ''!
srpMetastate: aMetastate
	"Private. Instances of SrpState are typically only created
	 when behaviors of state aren't defined in a loading image.
	 Minimal behavior is associated to the state by associating
	 instances with metastate which describes them. Association
	 with metastate was far less complicated than generating native
	 state classes (with behavior) and then creating instances."

	metastate := aMetastate.

! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpNamedValueAt: index
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt: because it
	 works with special loader constructs."

	^self srpNamedValues at: index! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpNamedValueAt: anInteger put: anObject
	"Answer an Object contained within the fixed variable
	 position of the receiver, indicated by the Integer
	 argument index. The numbering of the variables
	 corresponds to the enumeration of the receiver's
	 named instance variables. This can be sent by
	 mapping rules. It is preferable to #instVarAt:put: because
	 it works with special loader constructs."

	^self srpNamedValues at: anInteger put: anObject
! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpNamedValues

	^namedValues! !

!SrpState methodsFor: 'accessing' stamp: ''!
srpNamedValues: anIndexableCollection

	namedValues := anIndexableCollection! !

!SrpState methodsFor: 'loading' stamp: ''!
srpPostLoadActionForLoader: aMarshaler
	"Answer a command (directed message or block) that
	 will answer an object to take the place of the
	 receiver when sent #value. The command will be
	 sent #value after the root object has been completely
	 loaded. Answer nil if no further processing is required."

	^[ aMarshaler missingBehaviorOf: self ]
! !

!SrpState methodsFor: 'loading' stamp: ''!
value
	"This is probably no longer necessary. It used to be sent
	 when the receiver acted itself as a post load action when
	 a behavior was not found."! !

Object subclass: #SrpStreamEncoding
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

SrpStreamEncoding subclass: #SrpBase64
	instanceVariableNames: 'hasCachedWrite bufferValue bitsCached'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpBase64 classSide methodsFor: 'documentation' stamp: ''!
documentation
"
	This class is used to encode binary data as common 
	alpha-numeric text so data can be transmitted through the 
	Internet without being mangled on the way.

	This class is used to augment the behavior of a stream which
	it decorates. Bytes written to the stream through instances
	of this class will be converted into base64 characters of
	the MIME alphabet. The size of data written will be 33%
	larger than the original size. Data is written in 24 bit groups.
	Three 8 bit bytes (24 bits) are logically joined and split
	into four 6 bit values (24 bits). Each 6 bit value is used as
	an index into the MIME alphabet to determine the
	portable character that should be used to represent the
	base64 (6 bit) value.

	In conformance with MIME standards, the = (equals)
	character is a special character used to pad the end of
	data that is not an exact multiple of 24 bits.

	Bytes that are not part of the MIME alphabet are simply
	skipped when reading. This means any reformatting of the
	encoded data with whitespace will not compromise the
	integrity of the data.
"! !

!SrpBase64 classSide methodsFor: 'accessing' stamp: ''!
validCharacters
	"Answer all the valid characters of the MIME alphabet. Each
	 base64 value is an index into this collection."

	^'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
! !

!SrpBase64 methodsFor: 'private' stamp: ''!
base256ForBase64: int

	^SrpPortal base256ToBase64Table at: int + 1! !

!SrpBase64 methodsFor: 'private' stamp: ''!
base64ForBase256: byte

	^SrpPortal base64ToBase256Table at: byte + 1
! !

!SrpBase64 methodsFor: 'Stream writing' stamp: ''!
close
	"Close the stream that the receiver is accessing."

	self srpFinishByteGroupWrite.
	stream close.! !

!SrpBase64 methodsFor: 'Stream reading' stamp: ''!
contents
	"Answer the entire decoded contents of stream. If
	 there are any cached writes then this will flush
	 those writes but keep the current position and
	 doesn't assume this is the end of any written data."

	| holder pos copy |
	hasCachedWrite ifTrue: [self srpFlushInPlace].
	pos := stream position.
	stream reset.
	copy := self class on: stream.
	holder := WriteStream on: ByteArray new.
	[copy atEnd] whileFalse: [holder nextPut: copy next].
	stream position: pos.
	^holder contents
! !

!SrpBase64 methodsFor: 'Stream writing' stamp: ''!
flush
	"Write any unwritten information."

	self srpFlushInPlace.
	super flush.! !

!SrpBase64 methodsFor: 'Stream reading' stamp: ''!
next
	| base64 base256 |
	bitsCached = 0 ifTrue: [
		base256 := self nextBase64 bitShift: 2.
		base64 := self nextBase64.
		base256 := base256 + (base64 bitShift: -4).
		bufferValue := base64 bitAnd: 2r001111.
		bitsCached := 4.
		^base256.
	].
	hasCachedWrite ifTrue: [self srpFlushInPlace].
	bitsCached = 4 ifTrue: [
		base256 := bufferValue bitShift: 4.
		base64 := self nextBase64.
		base256 := base256 + (base64 bitShift: -2).
		bufferValue := base64 bitAnd: 2r000011.
		bitsCached := 2.
		^base256
	].
	bitsCached = 2 ifTrue: [
		base256 := bufferValue bitShift: 6.
		base64 := self nextBase64.
		base256 := base256 + base64.
		bufferValue := bitsCached := 0.
		^base256
	].
	self srpMarshaler configuration reportError: 'Unexpected cache length'.! !

!SrpBase64 methodsFor: 'Stream reading' stamp: ''!
next: int

	| answer |
	answer := ByteArray new: int.
	1 to: int do: [:i | answer at: i put: self next ].
	^answer
! !

!SrpBase64 methodsFor: 'private' stamp: ''!
nextBase64
	"Read the next proper byte from the stream and
	 answer the six bit value (0 to 63) that that character
	 represents. This will skip over any unexpected
	 characters read from the stream so that reformatting
	 the encoded text will not make the data unreadable."

	| int |
	[
		(int := self base64ForBase256: stream next) < 0 
	]
		whileTrue: [].
	^int
! !

!SrpBase64 methodsFor: 'Stream writing' stamp: ''!
nextPut: byte

	bitsCached = 0 ifTrue: [
		self nextPutBase64: (byte bitShift: -2).
		bufferValue := byte bitAnd: 2r00000011.
		bitsCached := 2.
		hasCachedWrite := true.
		^byte
	].
	bitsCached = 2 ifTrue: [
		self nextPutBase64: (bufferValue bitShift: 4) + (byte bitShift: -4).
		bufferValue := byte bitAnd: 2r00001111.
		bitsCached := 4.
		hasCachedWrite := true.
		^byte
	].
	bitsCached = 4 ifTrue: [
		self nextPutBase64: (bufferValue bitShift: 2) + (byte bitShift: -6).
		self nextPutBase64: (byte bitAnd: 2r00111111).
		bufferValue := bitsCached := 0.
		hasCachedWrite := false.
		^byte
	].
	self srpMarshaler configuration reportError: 'Unexpected cache length'.! !

!SrpBase64 methodsFor: 'Stream writing' stamp: ''!
nextPutAll: bytes

	bytes do: [:ea | self nextPut: ea ].
	^bytes
! !

!SrpBase64 methodsFor: 'private' stamp: ''!
nextPutBase64: value

	stream nextPut: (self base256ForBase64: value).
! !

!SrpBase64 methodsFor: 'Stream positioning' stamp: ''!
position: newPos
	"This can only be safely used with extreme care.
	 Base64 data is intended to be dynamic, so the position you
 	 are jumping to may no longer be a valid start point. Also,
	 only 1/4 of the stream positions are capable of being a
	 valid start position for Base64 encoded data. Base64
	 is not intended for random access."

	self flush.
	super position: newPos.
	self srpInitialize.
! !

!SrpBase64 methodsFor: 'Stream positioning' stamp: ''!
reset
	"Reposition to the start of the stream. If there are
	 any cached writes then this will flush those writes."

	self flush.
	self srpInitialize.
	super reset.! !

!SrpBase64 methodsFor: 'Stream positioning' stamp: ''!
setToEnd
	" Set the underlying stream to its end. "

	self flush.
	super setToEnd.
	self srpInitialize.
! !

!SrpBase64 methodsFor: 'Stream positioning' stamp: ''!
skip: int
	"This can only be safely used with extreme care.
	 Base64 data is intended to be dynamic, so the position you
 	 are jumping to may no longer be a valid start point. Also,
	 only 1/4 of the stream positions are capable of being a
	 valid start position for Base64 encoded data. Base64
	 is not intended for random access."

	self srpFlushInPlace.
	stream skip: int.
! !

!SrpBase64 methodsFor: 'SRP' stamp: ''!
srpFinishByteGroupRead
	"This is used to position the stream to the end of a partially
	 read byte group. This must be done in order to be properly
	 positioned for the next object in the stream."

	"bitsCached = 0 ifTrue: []."
		"Already at the start of a byte group. Two bytes
		 would be read for the first base256 encoded value."
	bitsCached = 4 ifTrue: [
		"Two of four bytes have been read. Just
		 read the next two. They should equal $=
		 if the stream is properly encoded."
		stream next; next.
	].
	bitsCached = 2 ifTrue: [
		"Three of four bytes have been read. Just
		 read the next one. It should equal $=
		 if the stream is properly encoded."
		stream next.
	].
	bitsCached := bufferValue := 0.
	super srpFinishByteGroupRead.! !

!SrpBase64 methodsFor: 'SRP' stamp: ''!
srpFinishByteGroupWrite
	"If there are any cached writes then write them to the
	 stream and fill unused bytes in the byte group with $=
	 to indicate end of data.."

	hasCachedWrite ifFalse: [^self].
	bitsCached = 2 ifTrue: [
		"One of four bytes has been written. Write the
		 cached bits in the second byte and fill the last
		 two with $= to signal end of data."
		self nextPutBase64: (bufferValue bitShift: 4).
		bufferValue := bitsCached := 0.
		stream nextPut: 61. "$= value"
		stream nextPut: 61. "$= value"
	].
	bitsCached = 4 ifTrue: [
		"Two of four bytes has been written. Write the
		 cached bits in the third byte and fill the fourth
		 with $= to signal end of data."
		self nextPutBase64: (bufferValue bitShift: 2).
		bufferValue := bitsCached := 0.
		stream nextPut: 61. "$= value"
	].
	hasCachedWrite := false.
	super srpFinishByteGroupWrite.
! !

!SrpBase64 methodsFor: 'SRP' stamp: ''!
srpFlushInPlace
	"This makes it possible to switch from writing to a stream
	 to reading from the stream. If any writes haven't been
	 written when a read is desired then this will write them
	 without destroying data that might exist in the next
	 stream position."

	| keptBufferValue keptBitsCached keptPosition 	existingBitValue |
	hasCachedWrite ifFalse: [^self].
	keptBufferValue := bufferValue.
	keptBitsCached := bitsCached.
	keptPosition := stream position.
	existingBitValue := 0.
	(stream atEnd or: [stream peek = $= asInteger]) ifFalse: [ 
		(existingBitValue := self nextBase64) = 0 
			ifTrue: [stream position: keptPosition]
			ifFalse: [stream skip: -1].
	].
	bitsCached = 2 ifTrue: [
		"One of four bytes has been written. Write the
		 cached bits in the second byte while preserving
		 the next four bits if they are defined."
		self nextPutBase64: (keptBufferValue bitShift: 4) + (existingBitValue bitAnd: 2r001111).
	].
	bitsCached = 4 ifTrue: [
		"Two of four bytes has been written. Write the
		 cached bits in the third byte while preserving
		 the next two bits if they are defined."
		self nextPutBase64: (bufferValue bitShift: 2) + (existingBitValue bitAnd: 2r000011).
	].
	stream position: keptPosition.
	bufferValue := keptBufferValue.
	bitsCached := keptBitsCached.
	hasCachedWrite := false.! !

!SrpBase64 methodsFor: 'SRP' stamp: ''!
srpInitialize

	super srpInitialize.
	hasCachedWrite := false.
	bufferValue := bitsCached := 0.! !

!SrpBase64 methodsFor: 'SRP' stamp: ''!
srpMaxSizeToEncodeBytesOfSize: numBytes
	"Answer the theoretical maximum number of physical bytes
	 that could be required to encode numBytes as an unsigned
	 integer stream."

	^(super srpMaxSizeToEncodeBytesOfSize: numBytes) * 4
! !

!SrpBase64 methodsFor: 'SRP' stamp: ''!
srpSetEncodingType: encodingType

	encodingType == #Text ifTrue: [^self].
	encodingType == #Binary ifTrue: [^stream srpSetEncodingType: encodingType].
	(encodingType == #WrappedText and: [(stream isKindOf: SrpWrappedBase64) not]) 
		ifTrue: [stream := SrpWrappedBase64 on: stream].
	^super srpSetEncodingType: encodingType
! !

SrpStreamEncoding subclass: #SrpNativeStreamWrapper
	instanceVariableNames: 'marshaler binary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpNativeStreamWrapper classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 This encoding is used to treat a stream containing a collection with
 text contents as if it were a binary stream. If you are writing/reading
 data to/from a string or a file stream in text mode then you need this
 as the last/first encoding. Without this, you would get an error when
 attempting to write bytes to a stream that is expecting characters.
 This encoding is used automatically when an object stream is set
 to #text mode.
"! !

!SrpNativeStreamWrapper classSide methodsFor: 'documentation' stamp: ''!
on: aStream for: aMarshaler

	^self new 
		srpMarshaler: aMarshaler;
		srpNativeStream: aStream;
		yourself! !

!SrpNativeStreamWrapper methodsFor: 'Stream writing' stamp: ''!
flush
	"Write any unwritten information. Some dialects
	 only implement #flush on file streams."

	(stream respondsTo: #flush)
		ifTrue: [stream flush].! !

!SrpNativeStreamWrapper methodsFor: 'Stream reading' stamp: ''!
next
	"Get the next character and return its value."

	^binary ifTrue: [stream next] ifFalse: [stream next asInteger]! !

!SrpNativeStreamWrapper methodsFor: 'Stream writing' stamp: ''!
nextPut: byte
	"The stream expects characters instead of bytes."
	"Note: GS/S doesn't understand #value: as most dialects
	 do, use #withValue: or #asCharacter for GS/S instead."

	^stream nextPut: (binary ifTrue: [byte] ifFalse: [SrpPortal unsignedToCharacter: byte])! !

!SrpNativeStreamWrapper methodsFor: 'Stream writing' stamp: ''!
nextPutAll: bytes

	binary 
		ifTrue: [stream nextPutAll: bytes]
		ifFalse: [bytes do: [:ea | self nextPut: ea ]].
	^bytes! !

!SrpNativeStreamWrapper methodsFor: 'Stream reading' stamp: ''!
peek
	"Get the next character and return its value."

	^binary ifTrue: [stream peek] ifFalse: [stream peek asInteger]
! !

!SrpNativeStreamWrapper methodsFor: 'Stream positioning' stamp: ''!
position
	" Answer the zero-based position of the underlying stream. "

	^stream position - SrpPortableBehavior portal streamBasePosition! !

!SrpNativeStreamWrapper methodsFor: 'Stream positioning' stamp: ''!
position: newPos
	" Set the zero based position of the underlying stream. "

	stream position: newPos + SrpPortableBehavior portal streamBasePosition! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpFinishByteGroupRead
	"This is used to position the stream to the end of a partially
	 read byte group. This must be done in order to be properly
	 positioned for the next object in the stream."

! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpFinishByteGroupWrite
	"This is used to position the stream to the end of a partially
	 read byte group. This must be done in order to be properly
	 positioned for the next object in the stream."

! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpInitialize
	"binary is true if the native stream contains bytes. binary is false
	 if the native stream contains characters."

	binary := true.! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpMarshaler

	^marshaler
! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpMarshaler: aMarshaler

	marshaler := aMarshaler
! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpNativeStream

	^stream! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpNativeStream: aStream

	stream := aStream.
	"[GemStone] GsFile>>next returns characters despite being opened for binary access."
	aStream class name == #GsFile ifTrue: [binary := false].

! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpSetEncodingType: encodingType

	encodingType == #Binary ifTrue: [binary := true].
	(encodingType == #Text or: [encodingType == #WrappedText]) ifTrue: [
		binary := false.
		^(SrpBase64 on: self)
			srpSetEncodingType: encodingType.
	].
	^super srpSetEncodingType: encodingType
! !

!SrpNativeStreamWrapper methodsFor: 'SRP' stamp: ''!
srpStream: aStream

	stream := aStream
! !

!SrpStreamEncoding classSide methodsFor: 'documentation' stamp: ''!
documentation
"
This is an abstract class that provides general stream
protocol. Subclasses are created to apply special
encodings to stream data as it is read and written.
An example of a useful encoding would be one that
encrypts bytes as they are written to a stream.
"! !

!SrpStreamEncoding classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new srpInitialize
! !

!SrpStreamEncoding classSide methodsFor: 'instance creation' stamp: ''!
on: aStream

	^self new srpStream: aStream! !

!SrpStreamEncoding methodsFor: 'Stream reading' stamp: ''!
atEnd

	^stream atEnd! !

!SrpStreamEncoding methodsFor: 'Stream writing' stamp: ''!
close
	"Close the stream that the receiver is accessing."

	self flush.
	stream close.! !

!SrpStreamEncoding methodsFor: 'Stream reading' stamp: ''!
contents

	^stream contents
! !

!SrpStreamEncoding methodsFor: 'Stream writing' stamp: ''!
cr
	"Answer the receiver after starting a new line
	 in the stream."

	stream cr! !

!SrpStreamEncoding methodsFor: 'Stream writing' stamp: ''!
flush
	"Write any unwritten information. Some dialects
	 only implement #flush on file streams."

	stream flush.! !

!SrpStreamEncoding methodsFor: 'Stream reading' stamp: ''!
next

	^stream next
! !

!SrpStreamEncoding methodsFor: 'Stream reading' stamp: ''!
next: int

	^stream next: int
! !

!SrpStreamEncoding methodsFor: 'Stream writing' stamp: ''!
nextPut: byte

	^stream nextPut: byte
! !

!SrpStreamEncoding methodsFor: 'Stream writing' stamp: ''!
nextPutAll: bytes

	^stream nextPutAll: bytes
! !

!SrpStreamEncoding methodsFor: 'Stream reading' stamp: ''!
peek

	^stream peek! !

!SrpStreamEncoding methodsFor: 'Stream positioning' stamp: ''!
position
	" Answer the position of the underlying stream. "

	^stream position! !

!SrpStreamEncoding methodsFor: 'Stream positioning' stamp: ''!
position: newPos
	" Set the position of the underlying stream. "

	stream position: newPos! !

!SrpStreamEncoding methodsFor: 'Stream positioning' stamp: ''!
reset

	stream reset
! !

!SrpStreamEncoding methodsFor: 'Stream positioning' stamp: ''!
setToEnd
	" Set the underlying stream to its end. "

	stream setToEnd! !

!SrpStreamEncoding methodsFor: 'Stream positioning' stamp: ''!
skip: int
	"This skips physical positions in the stream.
	 Some types of encodings may not support this."

	stream skip: int.
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpFinishByteGroupRead
	"This is used to position the stream to the end of a partially
	 read byte group. This must be done in order to be properly
	 positioned for the next object in the stream."

	stream srpFinishByteGroupRead.
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpFinishByteGroupWrite
	"This is used to position the stream to the end of a partially
	 read byte group. This must be done in order to be properly
	 positioned for the next object in the stream."

	stream srpFinishByteGroupWrite.
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpInitialize
	"Do nothing by default."! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpMarshaler

	^stream srpMarshaler
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpMaxSizeToEncodeBytesOfSize: numBytes
	"Answer the theoretical maximum number of physical bytes
	 that could be required to encode numBytes as an unsigned
	 integer stream."

	^numBytes! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpNativeStream

	^stream srpNativeStream
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpNativeStream: aStream

	stream srpNativeStream: aStream
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpSetEncodingType: encodingType

	stream := stream srpSetEncodingType: encodingType.
	^self
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpStream

	^stream
! !

!SrpStreamEncoding methodsFor: 'SRP' stamp: ''!
srpStream: aStream

	stream := (aStream isKindOf: SrpStreamEncoding)
		ifTrue: [aStream]
		ifFalse: [SrpNativeStreamWrapper on: aStream]
! !

!SrpStreamEncoding methodsFor: 'Stream writing' stamp: ''!
tab

	^stream tab! !

SrpStreamEncoding subclass: #SrpWrappedBase64
	instanceVariableNames: 'maxLength lineSeparator lastLfPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpWrappedBase64 classSide methodsFor: 'documentation' stamp: ''!
documentation
"
This is meant to be used to write binary data in text base64
encoding--wrapped to a given line length. It can be used for 
both reading and writing to the stream, but is really intended 
for just writing without exceeding line length standards.
"! !

!SrpWrappedBase64 methodsFor: 'Stream reading' stamp: ''!
cr
	"Answer the receiver after starting a new line
	 in the stream. It is assumed that #srpLineSeparator will
	 return a byte array that contains an LF character in it."

	super cr.
	lastLfPosition := stream position - 1.! !

!SrpWrappedBase64 methodsFor: 'Stream reading' stamp: ''!
next

	| byte |
	(byte := stream next) = 10 
		ifTrue: [ lastLfPosition := stream position - 1 ].
	^byte
! !

!SrpWrappedBase64 methodsFor: 'Stream reading' stamp: ''!
next: int
	"Answer the next int bytes. Maintain the lastLfPosition
	 while repositioning in the stream. This isn't used by SRP,
	 but is implemented for completeness."

	| bytes |
	bytes := stream next: int.
	bytes size to: 1 by: -1 do: [:i |
		(bytes at: i) = 10 ifTrue: [
			lastLfPosition := (stream position - (bytes size - i)) - 1.
			^ bytes
		].
	].
	^bytes
! !

!SrpWrappedBase64 methodsFor: 'Stream writing' stamp: ''!
nextPut: byte

	byte = 10 ifTrue: [
		stream nextPut: byte.
		lastLfPosition := stream position.
		^byte
	].
	self srpCurrentLineLength >= maxLength 
		ifTrue: [ self cr ].
	stream nextPut: byte.
	^byte! !

!SrpWrappedBase64 methodsFor: 'Stream writing' stamp: ''!
nextPutAll: bytes
	"Not the most efficient, but the simplest thing
	 that could possibly work. SRP uses #nextPut:
	 instead of this, so performance doesn't
	 matter here anyway."

	bytes do: [:byte | self nextPut: byte ].
	^bytes
! !

!SrpWrappedBase64 methodsFor: 'Stream positioning' stamp: ''!
position: int
	"This can only be safely used with extreme care.
	 Base64 data is intended to be dynamic, so the position you
 	 are jumping to may no longer be a valid start point. Also,
	 only 1/4 of the stream positions are capable of being a
	 valid start position for Base64 encoded data. Base64
	 is not intended for random access."

	super position: int.
	self srpSetLastLfPosition.
! !

!SrpWrappedBase64 methodsFor: 'Stream positioning' stamp: ''!
reset
	"Reposition to the start of the stream (position 0).
	 Set the lastLfPosition to the position before the
	 start position (0 - 1)."

	super reset.
	lastLfPosition := -1.! !

!SrpWrappedBase64 methodsFor: 'Stream positioning' stamp: ''!
setToEnd
	" Set the underlying stream to its end. "

	super setToEnd.
	self srpSetLastLfPosition.! !

!SrpWrappedBase64 methodsFor: 'Stream positioning' stamp: ''!
skip: anInteger
	"This can only be safely used with extreme care.
	 Base64 data is intended to be dynamic, so the position you
 	 are jumping to may no longer be a valid start point. Also,
	 only 1/4 of the stream positions are capable of being a
	 valid start position for Base64 encoded data. Base64
	 is not intended for random access."
	
	self position: self position + anInteger.
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpCurrentLineLength

	^stream position - (lastLfPosition max: 0)
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpDefaultMaxLength
	"The Internet standard defined by RFC1341 (MIME) specifies
	 that lines should be no more than 76 characters each. RFC1113
	 (base64) says lines should be exactly 64 characters." 

	^64! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpFindLastLfPosition
	"Answer the position of the last character of the
	 platform line delimiter from the current position
	 or -1 if an LF character isn't found (streams are 0 based)."

	| origPos lookFor |
	lookFor := SrpPortal characterToUnsigned: SrpPortal platformLineDelimiter last.
	origPos := stream position.
	origPos - 1 to: 0 by: -1 do: [:pos |
		stream position: pos.
		stream peek = lookFor ifTrue: [
			stream position: origPos.
			^pos
		].
	].
	stream position: origPos.
	^-1
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpInitialize

	super srpInitialize.
	maxLength := self srpDefaultMaxLength.! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpMaxLength

	^maxLength
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpMaxLength: int

	maxLength := int.
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpMaxSizeToEncodeBytesOfSize: numBytes
	"Answer the theoretical maximum number of physical bytes
	 that could be required to encode numBytes as an unsigned
	 integer stream."

	| size |
	size := (super srpMaxSizeToEncodeBytesOfSize: numBytes) * 4.
	^size + (size // self srpMaxLength * 2)
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpSetEncodingType: encodingType

	encodingType == #WrappedText ifTrue: [^self].
	encodingType == #Binary ifTrue: [^stream srpSetEncodingType: encodingType].
	^super srpSetEncodingType: encodingType
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpSetLastLfPosition

	lastLfPosition := self srpFindLastLfPosition.
! !

!SrpWrappedBase64 methodsFor: 'SRP' stamp: ''!
srpStream: aStream

	super srpStream: aStream.
	self srpSetLastLfPosition.
! !

SrpStreamEncoding subclass: #SrtNaturalBitOrder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtNaturalBitOrder classSide methodsFor: 'documentation' stamp: ''!
documentation
"
This encoding arranges bits in order of least significant bit to most
significant bit instead of the traditional most significant bit/digit
to least significant bit/digit.

To explain the purpose of this encoding would delve into number
theory. Suffice to say that it is more natural for numbers to grow
in magnitude in the order they are read than for the magnitude to
be determined after reaching the last bit/digit read. In other
words, We've become accustomed to reading/writing digits
of numbers in reverse order.

Text is read from left to right yet numbers are mentally interpreted
from right to left. When you look at a number, say 499248120, you
first visualize the whole number to get a feel for the size and to find
the end of the number. To comprehend the magnitude of the number
you start at the right most digit and mentally build the value working
towards the left.

Why? I doubt anyone could explain. Could be for the same reason that
keys in the QWERTY keyboard were purposely arranged in an
inefficient order. We are so accustomed to backwards digits that
nobody realizes how odd and wasteful it is. It would be too confusing
to change our ways now, but I would expect most alien civilizations
to write numbers (in whatever base) with least significant digits
coming before most significant digits--in whatever order aliens read
their written word.

Use of backward numbers has a negative effect on how we think.
When we show the binary form of the decimal 3, we show it like
...000011 (I show the elipsus because there could be any number
of binary digits to the left). Because we write numbers backards,
there is a natural inclination to declare some arbitrary (but sufficient) 
data structure size to hold the largest expected number. We end 
up thinking of numbers only by the space allocated for them 
(octet-8 bits, unsigned short-16 bits, unsigned long-32 bits, 
unsigned long long-64 bits).

If people put the least significant digit first then computers might
be much different. The bits of streamed data units, encoded in an
SRP like fashion, would logically join very naturally with bits in
the additive data units with only a one bit (the carry bit) data shift.
In other words, computers would have the potential to operate
faster for the same reason that humans could comprehend the
magnitude of numbers faster if written the other way.

Is this class necessary for SRP? Absolutely not; in fact, this
implementation doesn't have any noticeable advantage because
it only supports 8 bit data unit sizes and pays a performance cost
for double reversal (instead of a performance improvement with a
simple and efficient shift by the microprocessor). The only thing
this class is good for is as an example encoding for SRP. Oh, and
perhaps to communicate with aliens. <grin>

Performance comparisons of other bit widths has shown that
8 bits is currently the most efficient size for today's data
needs (probably as a carry-over from computers being born
using a 7 bit character set).
"! !

!SrtNaturalBitOrder methodsFor: 'Stream reading' stamp: ''!
next

	^self srpReverseBitsOfByte: stream next
! !

!SrtNaturalBitOrder methodsFor: 'Stream reading' stamp: ''!
next: int

	^(stream next: int)
		collect: [:ea | self srpReverseBitsOfByte: ea ]
! !

!SrtNaturalBitOrder methodsFor: 'Stream writing' stamp: ''!
nextPut: byte

	^stream nextPut: (self srpReverseBitsOfByte: byte)
! !

!SrtNaturalBitOrder methodsFor: 'Stream writing' stamp: ''!
nextPutAll: bytes

	^stream nextPutAll: (bytes collect: [:ea | self srpReverseBitsOfByte: ea ])
! !

!SrtNaturalBitOrder methodsFor: 'SRP' stamp: ''!
srpReverseBitsOfByte: aByte
	"Private. Answer the value of aByte if its bit representation was reversed.

	((self printStringRadix: 2 padTo: 8) reverse asRtcNumberRadix: 2).

	Fail if the value of the receiver is greater than 255."

	aByte == 0 ifTrue: [^0].
	^#(128 64 192 32 160 96 224 16 144 80 208 48 176 112 240 8 136 72 200 40 168 104 232 24 152 88 216 56 184 120 248 4 132 68 196 36 164 100 228 20 148 84 212 52 180 116 244 12 140 76 204 44 172 108 236 28 156 92 220 60 188 124 252 2 130 66 194 34 162 98 226 18 146 82 210 50 178 114 242 10 138 74 202 42 170 106 234 26 154 90 218 58 186 122 250 6 134 70 198 38 166 102 230 22 150 86 214 54 182 118 246 14 142 78 206 46 174 110 238 30 158 94 222 62 190 126 254 1 129 65 193 33 161 97 225 17 145 81 209 49 177 113 241 9 137 73 201 41 169 105 233 25 153 89 217 57 185 121 249 5 133 69 197 37 165 101 229 21 149 85 213 53 181 117 245 13 141 77 205 45 173 109 237 29 157 93 221 61 189 125 253 3 131 67 195 35 163 99 227 19 147 83 211 51 179 115 243 11 139 75 203 43 171 107 235 27 155 91 219 59 187 123 251 7 135 71 199 39 167 103 231 23 151 87 215 55 183 119 247 15 143 79 207 47 175 111 239 31 159 95 223 63 191 127 255)
		at: aByte
! !

Object subclass: #SrpSubstituteObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

SrpSubstituteObject subclass: #PmrIndexable
	instanceVariableNames: 'indexableValues'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

PmrIndexable subclass: #PmrArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrArray classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Array)! !

!PmrArray methodsFor: 'Not categorized' stamp: ''!
asNative

	| answer |
	answer := self nativeClass new: self srpIndexableValues size.
	1 to: self srpIndexableValues size do: [:index |
		answer at: index put: (self srpIndexableValueAt: index)
	].
	^answer! !

!PmrArray methodsFor: 'Not categorized' stamp: ''!
nativeClass

	^Array! !

PmrIndexable subclass: #PmrDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrDictionary classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Dictionary)! !

!PmrDictionary methodsFor: 'Not categorized' stamp: ''!
addAll: aCollection

	| index |
	index := 0.
	SrpPortal
		keyedCollection: aCollection 
		associationsDo: [ :assoc | 
			index := index + 1.
			self srpIndexableValueAt: index put: assoc
		].
	" ^self "
! !

!PmrDictionary methodsFor: 'Not categorized' stamp: ''!
nativeClass

	^Dictionary! !

PmrIndexable subclass: #PmrIdentityDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrIdentityDictionary classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(IdentityDictionary)! !

!PmrIdentityDictionary methodsFor: 'Not categorized' stamp: ''!
addAll: aCollection

	| index |
	index := 0.
	SrpPortal
		keyedCollection: aCollection 
		associationsDo: [ :assoc | 
			index := index + 1.
			self srpIndexableValueAt: index put: assoc
		].
	" ^self "
! !

!PmrIdentityDictionary methodsFor: 'Not categorized' stamp: ''!
nativeClass

	^IdentityDictionary! !

PmrIndexable subclass: #PmrIdentitySet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrIdentitySet classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(IdentitySet EsIdentitySet)! !

!PmrIdentitySet methodsFor: 'As yet unclassified' stamp: ''!
nativeClass

	^self portal identitySetClass! !

!PmrIndexable classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	"^anObject convertToClass: self"
	^(self srpNew: anObject size)
		addAll: anObject;
		yourself.
! !

!PmrIndexable classSide methodsFor: 'Not categorized' stamp: ''!
documentation "

	SRP will treat instances of this class as if it had indexable slots. 
	This class exists for the sake of environments that don't
	allow user defined indexable/variable classes. ObjectStudio
	is the only known Smalltalk dialect with this limitation, but
	this will be useful for porting to other languages too.

	The indexableValues variable is a transient holder of 
	indexable values that might otherwise go into indexable slots.

"! !

!PmrIndexable classSide methodsFor: 'Not categorized' stamp: ''!
srpDataTypeMap
	"ObjectStudio doesn't allow user defined indexable classes.
	 The indexableValues variable is a transient holder of these values."

	^#((indexableValues nil))
! !

!PmrIndexable classSide methodsFor: 'Not categorized' stamp: ''!
srpHasIndexableValues

	^true! !

!PmrIndexable classSide methodsFor: 'Not categorized' stamp: ''!
srpNew: intSize

	^self new 
		srpIndexableValues: (Array new: intSize);
		yourself
! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
addAll: aCollection

	| index |
	index := 0.
	aCollection do: [ :assoc | 
		index := index + 1.
		self srpIndexableValueAt: index put: assoc
	].
	" ^self "
! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
asNative

	| answer |
	answer := self nativeClass new: self srpIndexableValues size.
	self srpIndexableValues do: [:item | answer add: item ].
	^answer! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
srpIndexableSize
	"Answer the basic size of the receiver. This can be sent by
	 mapping rules. It is preferable to #basicSize because it
	 works with special loader constructs."

	^self srpIndexableValues size! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
srpIndexableValueAt: index
	"Answer the object at the variable index position of the receiver."

	^self srpIndexableValues at: index! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
srpIndexableValueAt: index put: value
	"Store the Object argument value at the variable index position 
	 in the receiver."

	^self srpIndexableValues at: index put: value! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
srpIndexableValues

	^indexableValues! !

!PmrIndexable methodsFor: 'Not categorized' stamp: ''!
srpIndexableValues: anIndexableCollection

	indexableValues := anIndexableCollection! !

PmrIndexable subclass: #PmrOrderedCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrOrderedCollection classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(OrderedCollection)! !

!PmrOrderedCollection methodsFor: 'Not categorized' stamp: ''!
nativeClass

	^OrderedCollection! !

PmrIndexable subclass: #PmrSet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrSet classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Set)! !

!PmrSet methodsFor: 'Not categorized' stamp: ''!
nativeClass

	^Set! !

SrpSubstituteObject subclass: #PmrInterval
	instanceVariableNames: 'from to by'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrInterval classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Interval)! !

!PmrInterval classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"VisualAge uses from,to,by without getters."
	"ObjectStudio uses start,stop,increment with getters."
	"VisualSmalltalk uses beginning,end,increment without getters."
	"VisualWorks uses start,stop,step without getters."

	"At least they all use the same order, so..."
	^self
		from: (anObject srpNamedValueAt: 1)
		to: (anObject srpNamedValueAt: 2)
		by: (anObject srpNamedValueAt: 3)

	"	A less generic form:
		^self
			from: (anObject srpAttribute: 'from')
			to: (anObject srpAttribute: 'to')
			by: (anObject srpAttribute: 'by')
	"! !

!PmrInterval classSide methodsFor: 'As yet unclassified' stamp: ''!
from: fromValue to: toValue by: byValue

	^self new
		from: fromValue 
		to: toValue 
		by: byValue
! !

!PmrInterval methodsFor: 'Not categorized' stamp: ''!
asNative

	^from to: to by: by! !

!PmrInterval methodsFor: 'Not categorized' stamp: ''!
from: fromValue to: toValue by: byValue

	from := fromValue.
	to := toValue.
	by := byValue.
! !

SrpSubstituteObject subclass: #PmrMagnitude
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

PmrMagnitude subclass: #PmrChronometric
	instanceVariableNames: ''
	classVariableNames: 'TimeZonesFromGMT'
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrChronometric classSide methodsFor: 'Not categorized' stamp: ''!
isUtcTime: aTime

	^(aTime respondsTo: #isUtc) 
		ifTrue: [ aTime isUtc ] 
		ifFalse: [ false ].
! !

!PmrChronometric classSide methodsFor: 'Not categorized' stamp: ''!
secondsFromUtc

	^self portal secondsFromUtc! !

!PmrChronometric classSide methodsFor: 'Not categorized' stamp: ''!
timeZonesFromGMT: anInteger
	"Set the number of time zones from GMT that the user
	 is geographically located. CAUTION: This should be set
	 before any instances are created and/or migrated in
	 the image."

	anInteger > 24 | (anInteger < -24) 
		ifTrue: [ ^self error: 'Invalid number of time zones.' ].
	TimeZonesFromGMT := anInteger! !

!PmrChronometric methodsFor: 'Not categorized' stamp: ''!
isUtc
	"Answer boolean whether the receiver represents
	 a time relative to Universal Coordinated Time."

	^false
! !

PmrChronometric subclass: #PmrDate
	instanceVariableNames: 'year month day'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrDate classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Date)! !

!PmrDate classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self
		year: anObject year
		month: (self portal dateMonth: anObject)
		day: (self portal dateDayOfMonth: anObject)! !

!PmrDate classSide methodsFor: 'As yet unclassified' stamp: ''!
year: year month: month day: day

	^self new
		year: year
		month: month
		day: day
! !

!PmrDate methodsFor: 'Not categorized' stamp: ''!
asNative

	^self portal
		dateClassNewDay: day
		monthIndex: month
		year: year! !

!PmrDate methodsFor: 'Not categorized' stamp: ''!
year: newYear month: newMonth day: newDay

	year := newYear.
	month := newMonth.
	day := newDay.
! !

PmrChronometric subclass: #PmrDateTime
	instanceVariableNames: 'date time'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrDateTime classSide methodsFor: 'Not categorized' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(AbtTimestamp)! !

!PmrDateTime classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self newFromTimestamp: anObject
! !

!PmrDateTime classSide methodsFor: 'Not categorized' stamp: ''!
newFromTimestamp: aTimestamp

	^self 
		date: aTimestamp date
		time: aTimestamp time! !

!PmrDateTime methodsFor: 'Not categorized' stamp: ''!
asNative

	^self portal
		dateTimeClassDate: self date
		time: self time! !

!PmrDateTime methodsFor: 'Not categorized' stamp: ''!
date

	^date
! !

!PmrDateTime methodsFor: 'Not categorized' stamp: ''!
date: aDate time: aTime

	date :=  aDate.
	time := aTime.
! !

!PmrDateTime methodsFor: 'Not categorized' stamp: ''!
time

	^time
! !

PmrChronometric subclass: #PmrElapsedSeconds
	instanceVariableNames: 'elapsedSeconds'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrElapsedSeconds classSide methodsFor: 'Not categorized' stamp: ''!
fromSeconds: secondCount
	"Answer an instance which represents secondCount
	 number of seconds since a known time."

	 ^self new setFromSeconds: secondCount! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
asLocalSeconds

	^self isUtc
		ifTrue:  [ elapsedSeconds - self class secondsFromUtc ]
		ifFalse: [ elapsedSeconds ]

		

! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
asSeconds

	^elapsedSeconds
! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
asUtcSeconds

	^self isUtc
		ifTrue:  [ elapsedSeconds ]
		ifFalse: [ elapsedSeconds + self class secondsFromUtc]

		

! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
secondsPerDay  
	"The number of seconds in a typical day."

	^86400
! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
setFromLocalSeconds: seconds

	| adjustedSeconds |
	adjustedSeconds := self isUtc
		ifTrue:  [ seconds + self class secondsFromUtc ]
		ifFalse: [ seconds ].
	self setFromSeconds: adjustedSeconds! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
setFromSeconds: seconds

	elapsedSeconds := seconds.

! !

!PmrElapsedSeconds methodsFor: 'Not categorized' stamp: ''!
setFromUtcSeconds: seconds

	| adjustedSeconds |
	adjustedSeconds := self isUtc
		ifTrue:  [ seconds ]
		ifFalse: [ seconds - self class secondsFromUtc].
	self setFromSeconds: adjustedSeconds! !

PmrElapsedSeconds subclass: #PmrTime
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrTime classSide methodsFor: 'Not categorized' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Time)! !

!PmrTime classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self newFromTime: anObject
! !

!PmrTime classSide methodsFor: 'Not categorized' stamp: ''!
fromSeconds: secondCount
	"Answer an instance of Time which represents
	 secondCount number of seconds since midnight.
	 Fail if secondCount is not an Integer."

	 ^self new setFromSeconds: secondCount \\ 86400 "seconds per day"! !

!PmrTime classSide methodsFor: 'Not categorized' stamp: ''!
hours: hrs minutes: mins seconds: secs
	"Answer the receiver with the time set to that specified
	 by hrs:mins:secs."

	^self new
		hours: hrs 
		minutes: mins 
		seconds: secs
! !

!PmrTime classSide methodsFor: 'Not categorized' stamp: ''!
newFromTime: aTime

	| originalSeconds |
	originalSeconds := self portal secondsFromTime: aTime.
	^(self isUtcTime: aTime)
		ifTrue: [self new setFromUtcSeconds: originalSeconds ]
		ifFalse: [self new setFromLocalSeconds: originalSeconds ]! !

!PmrTime methodsFor: 'Not categorized' stamp: ''!
asNative

	^self portal timeClassFromSeconds: self asLocalSeconds! !

!PmrTime methodsFor: 'Not categorized' stamp: ''!
hours

	^elapsedSeconds // 3600.! !

!PmrTime methodsFor: 'Not categorized' stamp: ''!
hours: hrs minutes: mins seconds: secs
	"Answer the receiver with the time set to that specified
	 by hrs:mins:secs."

	elapsedSeconds := hrs * 3600 + (mins * 60) + secs.
! !

!PmrTime methodsFor: 'Not categorized' stamp: ''!
minutes

	^elapsedSeconds - (self hours * 3600) // 60.! !

!PmrTime methodsFor: 'Not categorized' stamp: ''!
seconds

	^elapsedSeconds - (self hours * 3600) - (self minutes * 60).! !

PmrMagnitude subclass: #PmrNumber
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

PmrNumber subclass: #PmrDecimal
	instanceVariableNames: 'integral scale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrDecimal classSide methodsFor: 'Not categorized' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(
		Decimal 
		ScaledDecimal 	"VA"
	 )
! !

!PmrDecimal classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self fromDecimal: anObject.
! !

!PmrDecimal classSide methodsFor: 'Not categorized' stamp: ''!
fromDecimal: aDecimal

	| scale |
	scale := aDecimal scale.
	^self new
		integral: (aDecimal * (10 raisedTo: scale)) asInteger
		scale: scale! !

!PmrDecimal methodsFor: 'Not categorized' stamp: ''!
asNative

	^self portal decimalClassFromString: self asString! !

!PmrDecimal methodsFor: 'Not categorized' stamp: ''!
asString

	^self printString
! !

!PmrDecimal methodsFor: 'Not categorized' stamp: ''!
integral: int scale: sc
	"Private. Never try to change just the integral or the
	 scale. Both are related to each other."

	integral := int.
	scale := sc.
! !

!PmrDecimal methodsFor: 'Not categorized' stamp: ''!
printOn: aStream

	| pInt shifter |
	pInt := integral < 0 
		ifTrue:  [ aStream nextPut: $-. integral negated ]
		ifFalse: [ integral ].
	shifter := 10 raisedTo: scale.
	(pInt // shifter) printOn: aStream.
	scale > 0 ifTrue: [
		aStream nextPut: $..
		(pInt \\ shifter) printOn: aStream.
	].
! !

PmrNumber subclass: #PmrFloat
	instanceVariableNames: 'numerator denominator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrFloat classSide methodsFor: 'Not categorized' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Float)! !

!PmrFloat classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self fromFloat: anObject.
! !

!PmrFloat classSide methodsFor: 'Not categorized' stamp: ''!
fromFloat: aFloat
	"Floats are stored like a fraction so as to avoid making assumptions
	 about the structure the float will be loaded into. This aids portability,
	 but some floats can convert to very large fractions:
		1234.567 asFraction --> (15432087500000000091/12500000000000000)
	 Loading such integers could be a problem for any programming language with
	 a maximum supported integer size."

	| array |
	array := self portal floatAsNumeratorAndDenominator: aFloat.
	^self new
		numerator: (array at: 1);
		denominator: (array at: 2);
		yourself! !

!PmrFloat methodsFor: 'As yet unclassified' stamp: ''!
asNative

	^(self numerator / self denominator) asFloat! !

!PmrFloat methodsFor: 'As yet unclassified' stamp: ''!
denominator
	"A sender of this method will not be expecting
	 a placeholder so make sure one is not given."

	^denominator srpWithoutPlaceholder
! !

!PmrFloat methodsFor: 'As yet unclassified' stamp: ''!
denominator: den

	denominator := den.
! !

!PmrFloat methodsFor: 'As yet unclassified' stamp: ''!
numerator
	"A sender of this method will not be expecting
	 a placeholder so make sure one is not given."

	^numerator srpWithoutPlaceholder
! !

!PmrFloat methodsFor: 'As yet unclassified' stamp: ''!
numerator: num

	numerator := num.
! !

SrpSubstituteObject subclass: #PmrMessage
	instanceVariableNames: 'selector arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

PmrMessage subclass: #PmrDirectedMessage
	instanceVariableNames: 'receiver'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrDirectedMessage classSide methodsFor: 'accessors' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Message DirectedMessage 
		Action	"ObjectStudio")! !

!PmrDirectedMessage classSide methodsFor: 'accessors' stamp: ''!
applicationPriority
	"Answer a value that will be used to sequence mapping 
	 rules in case there are multiple rules defined for the same
	 class. Higher values have priority over lower values."

	^550
! !

!PmrDirectedMessage classSide methodsFor: 'applying' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	(anObject respondsTo: #receiver) ifFalse: [^self].
	^self 
		selector: anObject selector
		arguments: anObject arguments
		receiver: anObject receiver
! !

!PmrDirectedMessage classSide methodsFor: 'applying' stamp: ''!
selector: aSymbol arguments: anArray receiver: anObject

	^self new
		receiver: anObject;
		selector: aSymbol;
		arguments: anArray! !

!PmrDirectedMessage methodsFor: 'load conversion' stamp: ''!
asNative

	^self portal
		directedMessageClassSelector: selector
		arguments: arguments
		receiver: receiver! !

!PmrDirectedMessage methodsFor: 'printing' stamp: ''!
printOn: aStream
	"Append to the argument aStream, a sequence of characters that
	 describes the receiver."

	aStream nextPutAll: self class name. 
	aStream space; nextPut: $(.
	self receiver printOn: aStream.	
	aStream nextPut: $,; space.	
	self selector printOn: aStream.	
	aStream nextPut: $,; space.	
	self arguments printOn: aStream.
	aStream nextPut: $).	 ! !

!PmrDirectedMessage methodsFor: 'accessors' stamp: ''!
receiver

	"Answer the receiver instance variable of the receiver."

	^receiver! !

!PmrDirectedMessage methodsFor: 'accessors' stamp: ''!
receiver: anObject

	"Set the receiver instance variable of the receiver."

	receiver := anObject! !

!PmrDirectedMessage methodsFor: 'utility' stamp: ''!
send
	"A PmrDirectedMessage is usually mapped to a native message
	 class on load; therefore, this method may only ever be
	 used by a dialect that lacks a native message class. Don't
	 send this message until after a load traversal has finished."

	^receiver perform: selector withArguments: arguments! !

!PmrMessage classSide methodsFor: 'accessors' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Message 
		SmalltalkMessage "ObjectStudio")! !

!PmrMessage classSide methodsFor: 'applying' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self 
		selector: anObject selector
		arguments: anObject arguments
! !

!PmrMessage classSide methodsFor: 'applying' stamp: ''!
new

	^super new initialize! !

!PmrMessage classSide methodsFor: 'applying' stamp: ''!
selector: aSymbol arguments: anArray

	^self new
		selector: aSymbol;
		arguments: anArray! !

!PmrMessage methodsFor: 'accessors' stamp: ''!
arguments

	"Answer the arguments of the receiver."

	^arguments! !

!PmrMessage methodsFor: 'accessors' stamp: ''!
arguments: anArray

	"Set the arguments of the receiver."

	arguments := anArray! !

!PmrMessage methodsFor: 'load conversion' stamp: ''!
asNative

	^self portal
		messageClassSelector: selector
		arguments: arguments! !

!PmrMessage methodsFor: 'initialization' stamp: ''!
initialize

	arguments := #().! !

!PmrMessage methodsFor: 'printing' stamp: ''!
printOn: aStream
	"Append to the argument aStream, a sequence of characters that
	 describes the receiver."

	aStream nextPutAll: self class name. 
	aStream space; nextPut: $(.
	self selector printOn: aStream.	
	aStream nextPut: $,; space.	
	self arguments printOn: aStream.
	aStream nextPut: $).	 ! !

!PmrMessage methodsFor: 'accessors' stamp: ''!
selector

	"Answer the selector of the receiver."

	^selector! !

!PmrMessage methodsFor: 'accessors' stamp: ''!
selector: aSymbol

	"Set the selector of the receiver."

	selector := aSymbol! !

!PmrMessage methodsFor: 'utility' stamp: ''!
sendTo: anObject
	"A PmrMessage is usually mapped to a native message
	 class on load; therefore, this method may only ever be
	 used by a dialect that lacks native message class. Don't
	 send this message until after a load traversal has finished."

	^anObject 
		perform: selector 
		withArguments: arguments! !

SrpSubstituteObject subclass: #PmrNamespaceReference
	instanceVariableNames: 'namespace'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

PmrNamespaceReference subclass: #PmrClassReference
	instanceVariableNames: 'classSymbol'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrClassReference classSide methodsFor: 'Not categorized' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Metaclass MetaClass)

	"VisualSmalltalk uses #MetaClass."! !

!PmrClassReference classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	(self portal behaviorIsMetaclass: anObject) ifFalse: [^self forClass: anObject].
	^anObject! !

!PmrClassReference classSide methodsFor: 'Not categorized' stamp: ''!
forClass: aClass

	^self new classSymbol: (self portal shortNameForClass: aClass)! !

!PmrClassReference methodsFor: 'As yet unclassified' stamp: ''!
asNative

	^self portal
		namespace: self namespace
		classAt: self classSymbol! !

!PmrClassReference methodsFor: 'As yet unclassified' stamp: ''!
classSymbol
	"A sender of this method will not be expecting
	 a placeholder so make sure one is not given."

	^classSymbol srpWithoutPlaceholder
! !

!PmrClassReference methodsFor: 'As yet unclassified' stamp: ''!
classSymbol: aSymbol

	classSymbol := aSymbol
! !

PmrClassReference subclass: #PmrMetaclassReference
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrMetaclassReference classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	(self portal behaviorIsMetaclass: anObject) ifTrue: [^self forClass: anObject].
	^anObject! !

!PmrMetaclassReference classSide methodsFor: 'Not categorized' stamp: ''!
forClass: aClass

	^self new classSymbol: (self portal shortNameForClass: aClass)! !

!PmrMetaclassReference methodsFor: 'As yet unclassified' stamp: ''!
asNative

	| aClass |
	^(aClass := super asNative) isNil 
		ifFalse: [aClass class]
		ifTrue:  [nil]! !

PmrNamespaceReference subclass: #PmrGlobalReference
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrGlobalReference classSide methodsFor: 'Not categorized' stamp: ''!
name: aSymbol

	^self new name: aSymbol
! !

!PmrGlobalReference methodsFor: 'As yet unclassified' stamp: ''!
asNative

	^self portal
		namespace: self namespace
		ensureBindingFor: self name! !

!PmrGlobalReference methodsFor: 'As yet unclassified' stamp: ''!
name
	"A sender of this method will not be expecting
	 a placeholder so make sure one is not given."

	^name srpWithoutPlaceholder
! !

!PmrGlobalReference methodsFor: 'As yet unclassified' stamp: ''!
name: aSymbol

	name := aSymbol
! !

PmrNamespaceReference subclass: #PmrMethodBinding
	instanceVariableNames: 'method key'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrMethodBinding classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(
		Association
		VariableBinding
	  )
! !

!PmrMethodBinding classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	(context originalReference isKindOf: CompiledMethod)
		ifFalse: [^anObject].

	^self
		method: context originalReference
		key: anObject key
! !

!PmrMethodBinding classSide methodsFor: 'As yet unclassified' stamp: ''!
method: aCM key: key

	^self new
		method: aCM;
		key: key	! !

!PmrMethodBinding methodsFor: 'As yet unclassified' stamp: ''!
asNative
	"Answer an object that is to represent the receiver.
	 Keep in mind that each slot of the receiver may
	 have a placeholder around it. Also be aware that
	 when an object references the receiver, and the receiver
	 also references that object, that the slots for that
	 object may not be fully loaded yet."

	^self portal
		namespace: self namespace
		ensureBindingFor: self key! !

!PmrMethodBinding methodsFor: 'As yet unclassified' stamp: ''!
key

	^key! !

!PmrMethodBinding methodsFor: 'As yet unclassified' stamp: ''!
key: aKey

	key := aKey! !

!PmrMethodBinding methodsFor: 'As yet unclassified' stamp: ''!
method

	^method! !

!PmrMethodBinding methodsFor: 'As yet unclassified' stamp: ''!
method: aCM

	method := aCM! !

!PmrMethodBinding methodsFor: 'As yet unclassified' stamp: ''!
methodClass

	^self method methodClass
! !

!PmrNamespaceReference classSide methodsFor: 'Not categorized' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^Array 
		with: SrpPortal rootNamespace class name
		with: #EsSmalltalkDictionary	"VA only"
		with: #EsSmalltalkNamespace	"VA only"! !

!PmrNamespaceReference classSide methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	anObject == SrpPortal rootNamespace ifFalse: [^anObject].
	^self new
! !

!PmrNamespaceReference methodsFor: 'Not categorized' stamp: ''!
asNative

	^self namespace
! !

!PmrNamespaceReference methodsFor: 'Not categorized' stamp: ''!
defaultNamespace

	^SrpPortal rootNamespace
! !

!PmrNamespaceReference methodsFor: 'Not categorized' stamp: ''!
namespace

	namespace isNil ifTrue: [^self defaultNamespace].
	^namespace! !

!PmrNamespaceReference methodsFor: 'Not categorized' stamp: ''!
namespace: myNamespace
	"myNamespace is either nil to indicate to default
	 namespace or is some form of collection."

	myNamespace == self defaultNamespace
		ifTrue: [ namespace := nil ]
		ifFalse: [ namespace := myNamespace ].
! !

SrpSubstituteObject subclass: #PmrString
	instanceVariableNames: 'string'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrString classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(
		String 
		DBString	"VA"
		ByteString TwoByteString	"VW"
	 )! !

!PmrString classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self newFromString: anObject
! !

!PmrString classSide methodsFor: 'As yet unclassified' stamp: ''!
newFromString: aString

	^self new fromString: aString.
! !

!PmrString classSide methodsFor: 'As yet unclassified' stamp: ''!
srpDataTypeMap

	^super srpDataTypeMap, #(
		(string String)
	)! !

!PmrString methodsFor: 'As yet unclassified' stamp: ''!
asNative

	^self asString! !

!PmrString methodsFor: 'As yet unclassified' stamp: ''!
asString
	"Before SRP saves the type #String it sends #asString to
	 the object to be saved. This allows us to simply answer the
	 original string without needing to copy the contents first."

	^string! !

!PmrString methodsFor: 'As yet unclassified' stamp: ''!
fromString: aString
	"Before SRP saves the type #String it sends #asString to
	 the object to be saved. This allows us to simply answer the
	 original string without needing to copy the contents first."

	string := aString.
! !

!PmrString methodsFor: 'As yet unclassified' stamp: ''!
printOn: aStream

	super printOn: aStream.
	aStream nextPut: $ .
	self asNative printOn: aStream.
! !

PmrString subclass: #PmrSymbol
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-PortableObjectRepresentations'!

!PmrSymbol classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(
		Symbol 
		ByteSymbol "VW"
	 )! !

!PmrSymbol classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self newFromSymbol: anObject
! !

!PmrSymbol classSide methodsFor: 'As yet unclassified' stamp: ''!
newFromSymbol: aSymbol
	"We don't need to convert the symbol to a
	 real string because symbols have protocol
	 compatible with strings."
	
	^self new fromString: aSymbol asString
! !

!PmrSymbol classSide methodsFor: 'As yet unclassified' stamp: ''!
srpDataTypeMap

	^super srpDataTypeMap, #(
		(string String)
	)! !

!PmrSymbol methodsFor: 'As yet unclassified' stamp: ''!
asNative

	^self portal symbolClassIntern: self asString! !

SrpSubstituteObject subclass: #SrpConfigurableRule
	instanceVariableNames: 'applyWhen priority applicableClassNames command'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpConfigurableRule classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 This class is a instance-based exchange rule that allows load and/or
 save rules to be defined without creating class definitions. 
"! !

!SrpConfigurableRule classSide methodsFor: 'instance creation' stamp: ''!
newForLoading
	"Answer a new mapping rule that is to be applied when loading."

	^self new initialize setAsLoadRule! !

!SrpConfigurableRule classSide methodsFor: 'instance creation' stamp: ''!
newForSaving
	"Answer a new mapping rule that is to be applied when saving."

	^self new initialize setAsSaveRule! !

!SrpConfigurableRule methodsFor: 'rule collection protocol' stamp: ''!
, exchangeRule
	"Add an exchange rule while keeping rules sorted by priority. Higher
	 priority is before lower priority. Duplicates are ignored."

	self == exchangeRule ifTrue: [^self].
	^SrpSubstitutions new , exchangeRule , self! !

!SrpConfigurableRule methodsFor: 'accessing' stamp: ''!
applicableClassNames
	"Answer the receiver's instance variable applicableClassNames."

	^applicableClassNames! !

!SrpConfigurableRule methodsFor: 'accessing' stamp: ''!
applicableClassNames: anObject
	"Set the receiver's instance variable applicableClassNames to anObject."

	applicableClassNames := anObject! !

!SrpConfigurableRule methodsFor: 'accessing' stamp: ''!
applicationPriority
	"Answer a value that will be used to sequence mapping 
	 rules in case there are multiple rules defined for the same
	 class. Higher values have priority over lower values."

	^priority! !

!SrpConfigurableRule methodsFor: 'accessing' stamp: ''!
applicationPriority: aMagnitude
	"Set a value that will be used to sequence mapping 
	 rules in case there are multiple rules defined for the same
	 class. Higher values have priority over lower values."

	priority := aMagnitude.! !

!SrpConfigurableRule methodsFor: 'applying' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self command value: anObject value: context! !

!SrpConfigurableRule methodsFor: 'accessing' stamp: ''!
command
	"Answer an object that responds to #value:value: that will
	 be passed an object being mapped as the first argument,
	 and the context in which that object is mapped as the
	 second argument."

	^command! !

!SrpConfigurableRule methodsFor: 'accessing' stamp: ''!
command: aValueValueThingy
	"Set an object that responds to #value:value: that will
	 be passed an object being mapped as the first argument,
	 and the context in which that object is mapped as the
	 second argument."

	command := aValueValueThingy.! !

!SrpConfigurableRule methodsFor: 'initialization' stamp: ''!
defaultCommandBlock

	^[:object :context | object].! !

!SrpConfigurableRule methodsFor: 'rule collection protocol' stamp: ''!
includes: aMappingRule
	"Receiver is not a rule collection, so just return
	 boolean whether the receiver is the rule."

	^self == aMappingRule
! !

!SrpConfigurableRule methodsFor: 'initialization' stamp: ''!
initialize
	applicableClassNames := self class applicableClassNames.
	priority := self class applicationPriority.
	command := self defaultCommandBlock.! !

!SrpConfigurableRule methodsFor: 'testing' stamp: ''!
isDefaultRule
	"Answer true for dynamic mapping rules, because whether
	 they will be used depends on if they are visible to a configuration."

	^true! !

!SrpConfigurableRule methodsFor: 'initialization' stamp: ''!
setAsLoadRule
	"This rule is to be applied when loading."

	applyWhen := #Loading.! !

!SrpConfigurableRule methodsFor: 'initialization' stamp: ''!
setAsSaveRule
	"This rule is to be applied when loading."

	applyWhen := #Saving.! !

!SrpConfigurableRule methodsFor: 'testing' stamp: ''!
useWhenLoading
	"Answer boolean whether this rule is used when loading."

	^applyWhen == #Loading
! !

!SrpConfigurableRule methodsFor: 'testing' stamp: ''!
useWhenSaving
	"Answer boolean whether this rule is used when saving."

	^applyWhen == #Saving
! !

!SrpConfigurableRule methodsFor: 'rule collection protocol' stamp: ''!
without: aMappingRule
	"Receiver is not a rule collection, so just return
	 nil if the receiver was the rule to be removed."

	^self = aMappingRule
		ifTrue: [nil]
		ifFalse: [self]
! !

!SrpSubstituteObject classSide methodsFor: 'adding/removing' stamp: ''!
, exchangeRule 
	"Add an exchange rule while keeping rules sorted by priority. Higher
	 priority is before lower priority. Duplicates are ignored."

	self == exchangeRule ifTrue: [^self].
	^SrpSubstitutions new , exchangeRule , self! !

!SrpSubstituteObject classSide methodsFor: 'accessors' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#()! !

!SrpSubstituteObject classSide methodsFor: 'accessors' stamp: ''!
applicationPriority
	"Answer a value that will be used to sequence mapping 
	 rules in case there are multiple rules defined for the same
	 class. Higher values have priority over lower values."

	^500
! !

!SrpSubstituteObject classSide methodsFor: 'applying' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^anObject
! !

!SrpSubstituteObject classSide methodsFor: 'printing' stamp: ''!
debugPrintOn: aStream

	self printOn: aStream.
	aStream cr; tab; nextPutAll: 'priority: '.
	self applicationPriority printOn: aStream.
	aStream cr; tab; nextPutAll: 'applies to any of: '.
	self applicableClassNames printOn: aStream.
! !

!SrpSubstituteObject classSide methodsFor: 'documentation' stamp: ''!
documentation
"
 This is the abstract superclass of classes that represent
 class-based mapping rules that can be applied to instances
 of multiple specific classes as instances are saved or
 loaded.

 An example would be to map all instances of Association
 and VariableBinding when saving using the structure and rules 
 defined by PmrBinding. The idea is to map instances of certain
 classes (not their subclasses) to a portable form when saving,
 and then convert them back to a native form when loading.

 Multiple rules can be applied to the same instance. When
 that is the case, the mapping rules will be applied in order
 according to the value returned by #applicationPriority.

"! !

!SrpSubstituteObject classSide methodsFor: 'testing' stamp: ''!
includes: aMappingRule
	"Receiver is not a rule collection, so just return
	 boolean whether the receiver is the rule."

	^self == aMappingRule
! !

!SrpSubstituteObject classSide methodsFor: 'testing' stamp: ''!
isDefaultRule
	"Answer boolean whether this rule should added to
	 the collection of active rules by default. By default,
	 all mapping rules will be added. You must override
	 this method to answer false if you want to add rules
	 only under certain conditions--which you also specify."

	^true

! !

!SrpSubstituteObject classSide methodsFor: 'accessors' stamp: ''!
portal
	"This accessor is defined so that the portal can be accessed by subclasses
	 without introducing namespace visibility problems in VW."

	^SrpPortal! !

!SrpSubstituteObject classSide methodsFor: 'testing' stamp: ''!
useWhenLoading
	"Answer boolean whether this rule can be used
	 when loading."

	^self useWhenSaving not! !

!SrpSubstituteObject classSide methodsFor: 'testing' stamp: ''!
useWhenSaving
	"Answer boolean whether this rule can be used
	 when saving."

	^true! !

!SrpSubstituteObject classSide methodsFor: 'adding/removing' stamp: ''!
without: aMappingRule
	"Receiver is not a rule collection, so just return
	 nil if the receiver was the rule to be removed."

	^self == aMappingRule
		ifTrue: [nil]
		ifFalse: [self]
! !

!SrpSubstituteObject methodsFor: 'load conversion' stamp: ''!
asNative
	"Answer an object that is to represent the receiver.
	 Keep in mind that each slot of the receiver may
	 have a placeholder around it. Also be aware that
	 when an object references the receiver, and the receiver
	 also references that object, that the slots for that
	 object may not be fully loaded yet."

	^self subclassResponsibility! !

!SrpSubstituteObject methodsFor: 'load conversion' stamp: ''!
portal
	"This accessor is defined so that the portal can be accessed by subclasses
	 without introducing namespace visibility problems in VW."

	^self class portal! !

!SrpSubstituteObject methodsFor: 'load conversion' stamp: ''!
srpLoadPostMap: aContext
	"Answer an object that is to represent the receiver when
	 loaded. Keep in mind that slots of the receiver could 
	 still contain placeholders around an object that may
	 yet be mapped. Also be aware that when an object
	 references the receiver, and the receiver also references
	 that object, that the slots for that object may not be fully
	 loaded yet."

	^self asNative! !

SrpSubstituteObject subclass: #SrpSubstitutions
	instanceVariableNames: 'rules'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-StateReplicationProtocol'!

!SrpSubstitutions classSide methodsFor: 'Not categorized' stamp: ''!
documentation	"

	This is used to apply more than one substitution to a given type of object.

"! !

!SrpSubstitutions classSide methodsFor: 'Not categorized' stamp: ''!
new

	^super new initialize! !

!SrpSubstitutions methodsFor: 'Not categorized' stamp: ''!
, exchangeRule
	"Add an exchange rule while keeping rules sorted by priority. Higher
	 priority is before lower priority. Between rules of the same priority,
	 the last added has higher priority."

	rules do: [:ea |
		ea = exchangeRule
			ifTrue: [^self].
		ea applicationPriority <= exchangeRule applicationPriority
			ifTrue: [ rules add: exchangeRule before: ea. ^self ].
	].
	rules add: exchangeRule.! !

!SrpSubstitutions methodsFor: 'Not categorized' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply exchange
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject. Give another
	 rule a chance so long as the result is the same class."

	| answer prior |
	answer := anObject.
	rules do: [:exchangeRule |
		prior := answer.
		answer := exchangeRule applyTo: answer context: context.
		context mappedCurrent: answer.
		answer class == prior class ifFalse: [^answer].
	].
	^answer
! !

!SrpSubstitutions methodsFor: 'Not categorized' stamp: ''!
debugPrintOn: aStream

	| rule |
	self printOn: aStream.
	1 to: rules size do: [:i |
		rule := rules at: i.
		aStream cr; tab.
		i printOn: aStream.
		aStream nextPutAll: ': '.
		rule printOn: aStream.
		aStream nextPutAll: ' ('.
		rule applicationPriority printOn: aStream.
		aStream nextPutAll: ')'.
	].! !

!SrpSubstitutions methodsFor: 'Not categorized' stamp: ''!
includes: exchangeRule
	"Receiver is not a rule collection, so just return
	 boolean whether the receiver is the rule."

	^rules includes: exchangeRule
! !

!SrpSubstitutions methodsFor: 'Not categorized' stamp: ''!
initialize

	"Initialize the receiver."

	rules := OrderedCollection new.! !

!SrpSubstitutions methodsFor: 'Not categorized' stamp: ''!
without: exchangeRule
	"Remove a mapping rule."

	rules remove: exchangeRule ifAbsent: [].
	rules isEmpty ifTrue: [^nil].
	^rules size > 1
		ifTrue: [self]
		ifFalse: [rules first]
! !

SrpSubstituteObject subclass: #SrtSaveExchangeRule
	instanceVariableNames: 'string actual hash'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtSaveExchangeRule classSide methodsFor: 'As yet unclassified' stamp: ''!
applicableClassNames
	"Answer a collection of all class names whose instances
	 are affected by the receiver. The receiver will be applied
	 to all instances of classes named in the list."

	^#(Date)! !

!SrtSaveExchangeRule classSide methodsFor: 'As yet unclassified' stamp: ''!
applicationPriority
	"Answer a value that will be used to sequence mapping 
	 rules in case there are multiple rules defined for the same
	 class. Higher values have priority over lower values."

	^1500
! !

!SrtSaveExchangeRule classSide methodsFor: 'As yet unclassified' stamp: ''!
applyTo: anObject context: context
	"The receiver is given the opportunity to apply mapping
	 rules to anObject. Answer anObject or another object
	 that is to take the place of anObject."

	^self new object: anObject
! !

!SrtSaveExchangeRule classSide methodsFor: 'As yet unclassified' stamp: ''!
isDefaultRule
	"Answer boolean whether this rule should added to
	 the collection of active rules by default. By default,
	 all mapping rules will be added. You must override
	 this method to answer false if you want to add rules
	 only under certain conditions--which you also specify."

	^false

! !

!SrtSaveExchangeRule methodsFor: 'As yet unclassified' stamp: ''!
asNative
	"Answer an object that is to represent the receiver.
	 Keep in mind that each slot of the receiver may
	 have a placeholder around it. Also be aware that
	 when an object references the receiver, and the receiver
	 also references that object, that the slots for that
	 object may not be fully loaded yet."

	^actual! !

!SrtSaveExchangeRule methodsFor: 'As yet unclassified' stamp: ''!
object: anObject

	string := anObject printString.
	actual := anObject.
	hash := anObject hash.! !

!SrtSaveExchangeRule methodsFor: 'As yet unclassified' stamp: ''!
srpRecursiveReferencesShouldReferenceMapped
	"An original object has been mapped to the receiver, and while saving the receiver there
	 was a reference to the original object. Answer true (the default) if a reference to the
	 mapped object should be saved, or false if the original object should be serialized
	 without remapping. This method should only be overridden to answer false for instances
	 that are only meant for mapping operations--not normal kernel instances."

	^false
! !

Object subclass: #SrtDistinctTestObject
	instanceVariableNames: 'key value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtDistinctTestObject classSide methodsFor: 'Not categorized' stamp: ''!
key: key value: value

	^self new 
		key: key; 
		value: value; 
		yourself! !

!SrtDistinctTestObject classSide methodsFor: 'Not categorized' stamp: ''!
srpDataTypeMap

	^#(
			(self Distinct)
		)


! !

!SrtDistinctTestObject methodsFor: 'Not categorized' stamp: ''!
key
	^key! !

!SrtDistinctTestObject methodsFor: 'Not categorized' stamp: ''!
key: aKey

	key := aKey! !

!SrtDistinctTestObject methodsFor: 'Not categorized' stamp: ''!
value
	^value! !

!SrtDistinctTestObject methodsFor: 'Not categorized' stamp: ''!
value: aValue
	value := aValue! !

Object subclass: #SrtProxyTestObject
	instanceVariableNames: 'key value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SRP-Tests'!

!SrtProxyTestObject classSide methodsFor: 'Not categorized' stamp: ''!
key: key value: value

	^self new 
		key: key; 
		value: value; 
		yourself! !

!SrtProxyTestObject classSide methodsFor: 'Not categorized' stamp: ''!
srpDataTypeMap
	"We want 'value' to be turned into a proxy."

	^#(
			(value Proxy)
		)


! !

!SrtProxyTestObject methodsFor: 'Not categorized' stamp: ''!
key
	^key! !

!SrtProxyTestObject methodsFor: 'Not categorized' stamp: ''!
key: aKey

	key := aKey! !

!SrtProxyTestObject methodsFor: 'Not categorized' stamp: ''!
value
	^value! !

!SrtProxyTestObject methodsFor: 'Not categorized' stamp: ''!
value: aValue
	value := aValue! !
SrpPortableBehavior initialize!
